using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using AOT;
using UnityEngine;

namespace ZEGO
{
    public class ZIMBridge
    {
        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_log_uploaded_callback))]
        public static void zim_on_log_uploaded_callback(System.IntPtr handle, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMLogUploadedCallback callback = null;

            lock (ZIMImpl.logUploadedCallbackLock)
            {
                zim.logUploadedCallback.TryGetValue(sequence, out callback);
                zim.logUploadedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_error_event))]
        public static void zim_on_error_event(IntPtr handle, zim_error error_info)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;
            if (zim.onError != null)
                zim.onError.BeginInvoke(zim, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_logged_in_callback))]
        public static void zim_on_logged_in_callback(System.IntPtr handle, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;
            ZIMLoggedInCallback callback = null;
            lock (ZIMImpl.loggedInCallbackLock)
            {
                zim.loggedInCallback.TryGetValue(sequence, out callback);
                zim.loggedInCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_connection_state_changed_event))]
        public static void zim_on_connection_state_changed_event(System.IntPtr handle,
            System.Int32 state, System.Int32 connection_event,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string extended_data)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            Dictionary<string, string> extendedData = ZIMUtil.DeserializeStringToDictionary(extended_data);

            if (zim.onConnectionStateChanged != null)
                zim.onConnectionStateChanged.BeginInvoke(zim, (ZIMConnectionState)state,
                    (ZIMConnectionEvent)connection_event, extendedData, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_state_changed_event))]
        public static void zim_on_room_state_changed_event(System.IntPtr handle,
            System.Int32 state, System.Int32 room_event,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string extended_data,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            Dictionary<string, string> extendedData = ZIMUtil.DeserializeStringToDictionary(extended_data);

            if (zim.onRoomStateChanged != null)
                zim.onRoomStateChanged.BeginInvoke(zim, (ZIMRoomState)state,
                    (ZIMRoomEvent)room_event, extendedData, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_token_will_expire_event))]
        public static void zim_on_token_will_expire_event(System.IntPtr handle, uint second)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            if (zim.onTokenWillExpire != null)
                zim.onTokenWillExpire.BeginInvoke(zim, second, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_token_renewed_callback))]
        public static void zim_on_token_renewed_callback(System.IntPtr handle, [In()][MarshalAsAttribute(UnmanagedType.LPStr)] string token, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMTokenRenewedCallback callback = null;

            lock (ZIMImpl.tokenRenewedCallbackLock)
            {
                zim.tokenRenewedCallback.TryGetValue(sequence, out callback);
                zim.tokenRenewedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(token, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_user_name_updated_callback))]
        public static void zim_on_user_name_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string user_name,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMUserNameUpdatedCallback callback = null;
            lock (ZIMImpl.userNameUpdatedCallbackLock)
            {
                zim.userNameUpdatedCallback.TryGetValue(sequence, out callback);
                zim.userNameUpdatedCallback.Remove(sequence);
            }
            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(user_name, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_user_avatar_url_updated_callback))]
        public static void zim_on_user_avatar_url_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string user_avatar_url, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMUserAvatarUrlUpdatedCallback callback = null;
            lock (ZIMImpl.userAvatarUrlUpdatedCallbackLock)
            {
                zim.userAvatarUrlUpdatedCallback.TryGetValue(sequence, out callback);
                zim.userAvatarUrlUpdatedCallback.Remove(sequence);
            }
            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(user_avatar_url, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_user_extended_data_updated_callback))]
        public static void zim_on_user_extended_data_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string user_extended_data,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMUserExtendedDataUpdatedCallback callback = null;
            lock (ZIMImpl.userExtendedDataUpdatedCallbackLock)
            {
                zim.userExtendedDataUpdatedCallback.TryGetValue(sequence, out callback);
                zim.userExtendedDataUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(user_extended_data, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_users_info_queried_callback))]
        public static void zim_on_users_info_queried_callback(System.IntPtr handle, System.IntPtr user_list,
            uint user_list_length, System.IntPtr error_user_list_ptr, uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMUsersInfoQueriedCallback callback = null;
            lock (ZIMImpl.usersInfoQueriedCallbackLock)
            {
                zim.usersInfoQueriedCallback.TryGetValue(sequence, out callback);
                zim.usersInfoQueriedCallback.Remove(sequence);
            }
            if (callback == null) return;

            List<ZIMUserFullInfo> userFullInfoList = new List<ZIMUserFullInfo>();
            zim_user_full_info[] user_full_infos = new zim_user_full_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_user_full_info>(ref user_full_infos, user_list, user_list_length);
            for (int i = 0; i < user_list_length; i++)
            {
                var userFullInfo = ZIMConverter.ChangeZIMUserFullInfoStructToClass(user_full_infos[i]);
                userFullInfoList.Add(userFullInfo);
            }

            zim_error_user_info[] error_user_infos_list = new zim_error_user_info[error_user_list_length];
            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos_list, error_user_list_ptr, error_user_list_length);

            List<ZIMErrorUserInfo> errorUserInfoList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(userFullInfoList, errorUserInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_attributes_updated_event))]
        public static void zim_on_room_attributes_updated_event(System.IntPtr handle, [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            zim_room_attributes_update_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string extended_data)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomAttributesUpdateInfo roomAttributesUpdateInfo = new ZIMRoomAttributesUpdateInfo();
            roomAttributesUpdateInfo.action = (ZIMRoomAttributesUpdateAction)info.room_attributes_action;
            roomAttributesUpdateInfo.roomAttributes = ZIMConverter.ChangeZIMRoomAttributesPtrToDictionary(info.room_attributes, info.room_attributes_length);

            if (zim.onRoomAttributesUpdated != null)
                zim.onRoomAttributesUpdated.BeginInvoke(zim, roomAttributesUpdateInfo, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_attributes_batch_updated_event))]
        public static void zim_on_room_attributes_batch_updated_event(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr infos_ptr, uint infos_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string extended_data)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMRoomAttributesUpdateInfo> roomAttributesUpdateInfoList = new List<ZIMRoomAttributesUpdateInfo>();
            zim_room_attributes_update_info[] room_attributes_update_infos = new zim_room_attributes_update_info[infos_length];

            ZIMUtil.GetStructListByPtr<zim_room_attributes_update_info>(ref room_attributes_update_infos, infos_ptr, infos_length);
            for (int i = 0; i < infos_length; i++)
            {
                var roomAttributesUpdateInfo = new ZIMRoomAttributesUpdateInfo();
                roomAttributesUpdateInfo.roomAttributes = ZIMConverter.ChangeZIMRoomAttributesPtrToDictionary(room_attributes_update_infos[i].room_attributes,
                    room_attributes_update_infos[i].room_attributes_length);
                roomAttributesUpdateInfo.action = (ZIMRoomAttributesUpdateAction)room_attributes_update_infos[i].room_attributes_action;
                roomAttributesUpdateInfoList.Add(roomAttributesUpdateInfo);
            }

            if (zim.onRoomAttributesBatchUpdated != null)
                zim.onRoomAttributesBatchUpdated.BeginInvoke(zim, roomAttributesUpdateInfoList, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_member_attributes_updated_event))]
        public static void zim_on_room_member_attributes_updated_event(System.IntPtr handle,
            System.IntPtr infos_ptr, uint infos_length, zim_room_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMRoomMemberAttributesUpdateInfo> roomMemberAttributesUpdateInfoList = new List<ZIMRoomMemberAttributesUpdateInfo>();
            zim_room_member_attributes_update_info[] room_member_attributes_update_infos = new zim_room_member_attributes_update_info[infos_length];
            ZIMUtil.GetStructListByPtr<zim_room_member_attributes_update_info>(ref room_member_attributes_update_infos, infos_ptr, infos_length);
            for (int i = 0; i < infos_length; i++)
            {
                var room_member_attributes_info = room_member_attributes_update_infos[i].attributes_info;
                var roomMemberAttributesUpdateInfo = new ZIMRoomMemberAttributesUpdateInfo();
                roomMemberAttributesUpdateInfo.attributesInfo.userID =
                room_member_attributes_info.user_id;
                roomMemberAttributesUpdateInfo.attributesInfo.attributes =
                    ZIMConverter.ChangeZIMRoomMemberAttributesPtrToDictionary(room_member_attributes_info.attributes,
                    room_member_attributes_info.attributes_length);
            }
            ZIMRoomOperatedInfo operatedInfo = new ZIMRoomOperatedInfo();
            operatedInfo.userID = operated_info.user_id;

            if (zim.onRoomMemberAttributesUpdated != null)
                zim.onRoomMemberAttributesUpdated.BeginInvoke(zim, roomMemberAttributesUpdateInfoList, operatedInfo, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_sent_callback))]
        public static void zim_on_message_sent_callback(System.IntPtr handle,
            zim_message message, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageSentCallback callback = null;
            ZIMMessage cMessage = null;
            lock (ZIMImpl.messageSentCallbackLock)
            {
                zim.messageSentCallback.TryGetValue(sequence, out callback);
                if (zim.messageObjMap.ContainsKey(sequence))
                {
                    zim.messageObjMap.TryGetValue(sequence, out cMessage);
                    zim.messageObjMap.Remove(sequence);
                }
                zim.messageSentCallback.Remove(sequence);
                zim.messageSendNotificationCallback.Remove(sequence);
                zim.mediaMessageSendNotificationCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;
            if (cMessage == null)
            {
                cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);
            }
            else
            {
                ZIMConverter.ChangeZIMMessageStructToClass(ref cMessage, message);
            }

            callback.BeginInvoke(cMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_attached_callback))]
        public static void zim_on_message_attached_callback(System.IntPtr handle,
            zim_message message, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageSendNotification messageNotification = null;
            ZIMMediaMessageSendNotification mediaMessageNotification = null;

            lock (ZIMImpl.messageSendNotificationCallbackLock)
            {

                zim.messageSendNotificationCallback.TryGetValue(sequence, out messageNotification);

            }

            lock (ZIMImpl.mediaMessageSendNotificationCallbackLock)
            {
                zim.mediaMessageSendNotificationCallback.TryGetValue(sequence, out mediaMessageNotification);
            }

            ZIMMessage cMessage = null;

            lock (ZIMImpl.messageObjMapLock)
            {
                if (zim.messageObjMap.ContainsKey(sequence))
                {
                    zim.messageObjMap.TryGetValue(sequence, out cMessage);
                    zim.messageObjMap.Remove(sequence);
                }
            }

            if (cMessage == null)
            {
                cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);
            }

            if (messageNotification != null)
            {
                if (messageNotification.onMessageAttached != null)
                {
                    messageNotification.onMessageAttached.BeginInvoke(cMessage, null, null);
                }
            }

            if (mediaMessageNotification != null)
            {
                if (mediaMessageNotification.onMessageAttached != null)
                {
                    mediaMessageNotification.onMessageAttached.BeginInvoke(cMessage, null, null);
                }
            }
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_receive_peer_message_event))]
        public static void zim_on_receive_peer_message_event(System.IntPtr handle,
            System.IntPtr message_list_ptr, uint message_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string from_user_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMMessage> messageList = new List<ZIMMessage>();
            zim_message[] zim_messages_list = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref zim_messages_list, message_list_ptr, message_list_length);
            for (uint i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(zim_messages_list[i]));
            }

            if (zim.onReceivePeerMessage != null)
                zim.onReceivePeerMessage.BeginInvoke(zim, messageList, from_user_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_receive_room_message_event))]
        public static void zim_on_receive_room_message_event(System.IntPtr handle,
            System.IntPtr message_list_ptr, uint message_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string from_room_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMMessage> messageList = new List<ZIMMessage>();
            zim_message[] zim_messages_list = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref zim_messages_list, message_list_ptr, message_list_length);
            for (uint i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(zim_messages_list[i]));
            }

            if (zim.onReceiveRoomMessage != null)
                zim.onReceiveRoomMessage.BeginInvoke(zim, messageList, from_room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_receive_group_message_event))]
        public static void zim_on_receive_group_message_event(System.IntPtr handle,
            System.IntPtr message_list_ptr, uint message_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string from_group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMMessage> messageList = new List<ZIMMessage>();
            zim_message[] zim_messages_list = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref zim_messages_list, message_list_ptr, message_list_length);
            for (uint i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(zim_messages_list[i]));
            }

            if (zim.onReceiveGroupMessage != null)
                zim.onReceiveGroupMessage.BeginInvoke(zim, messageList, from_group_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_queried_callback))]
        public static void zim_on_message_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id,
            System.Int32 conversation_type, System.IntPtr message_list_ptr,
            uint message_list_length, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageQueriedCallback callback = null;

            lock (ZIMImpl.messageQueriedCallbackLock)
            {
                zim.messageQueriedCallback.TryGetValue(sequence, out callback);
                zim.messageQueriedCallback.Remove(sequence);
            }


            if (callback == null) return;

            List<ZIMMessage> messageList = new List<ZIMMessage>();
            zim_message[] zim_messages_list = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref zim_messages_list, message_list_ptr, message_list_length);
            for (uint i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(zim_messages_list[i]));
            }

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, messageList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_deleted_callback))]
        public static void zim_on_message_deleted_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id,
            System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageDeletedCallback callback = null;

            lock (ZIMImpl.messageDeletedCallbackLock)
            {
                zim.messageDeletedCallback.TryGetValue(sequence, out callback);
                zim.messageDeletedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_revoked_callback))]
        public static void zim_on_message_revoked_callback(System.IntPtr handle, zim_message message,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageRevokedCallback callback = null;

            lock (ZIMImpl.messageRevokedCallbackLock)
            {
                zim.messageRevokedCallback.TryGetValue(sequence, out callback);
                zim.messageRevokedCallback.Remove(sequence);
            }

            if (callback == null) return;

            var cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(cMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_receipts_read_sent_callback))]
        public static void zim_on_message_receipts_read_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id,
            System.Int32 conversation_type,
            System.IntPtr error_message_id_list_ptr, uint error_message_id_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageReceiptsReadSentCallback callback = null;

            lock (ZIMImpl.messageReceiptsReadSentCallbackLock)
            {
                zim.messageReceiptsReadSentCallback.TryGetValue(sequence, out callback);
                zim.messageReceiptsReadSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;
            long[] error_message_ids = new long[error_message_id_list_length];
            ZIMUtil.GetStructListByPtr<long>(ref error_message_ids, error_message_id_list_ptr, error_message_id_list_length);

            List<long> errorMessageIDs = new List<long>(error_message_ids);

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, errorMessageIDs, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_receipts_info_queried_callback))]
        public static void zim_on_message_receipts_info_queried_callback(System.IntPtr handle,
            System.IntPtr receipt_infos_list_ptr,
            uint receipt_info_list_length,
            IntPtr error_message_id_list_ptr,
            uint error_message_id_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageReceiptsInfoQueriedCallback callback = null;
            lock (ZIMImpl.messageReceiptsInfoQueriedCallbackLock)
            {
                zim.messageReceiptsInfoQueriedCallback.TryGetValue(sequence, out callback);
                zim.messageReceiptsInfoQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_message_receipt_info[] message_receipt_infos_list = new zim_message_receipt_info[receipt_info_list_length];
            ZIMUtil.GetStructListByPtr<zim_message_receipt_info>(ref message_receipt_infos_list, receipt_infos_list_ptr, receipt_info_list_length);
            List<ZIMMessageReceiptInfo> infoList = ZIMConverter.ChangeZIMMessageReceiptInfoStructListToClassList(message_receipt_infos_list);

            long[] error_message_ids = new long[error_message_id_list_length];
            ZIMUtil.GetStructListByPtr<long>(ref error_message_ids, error_message_id_list_ptr, error_message_id_list_length);

            List<long> errorMessageIDs = new List<long>(error_message_ids);

            callback.BeginInvoke(infoList, errorMessageIDs, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_queried_callback))]
        public static void zim_on_conversation_queried_callback(System.IntPtr handle,
    zim_conversation conversation, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationQueriedCallback callback = null;

            lock (ZIMImpl.conversationQueriedCallbackLock)
            {
                zim.conversationQueriedCallback.TryGetValue(sequence, out callback);
                zim.conversationQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            var ZimConversation = ZIMConverter.ChangeZIMConversationStructToClass(conversation);

            callback.BeginInvoke(ZimConversation, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_list_queried_callback))]
        public static void zim_on_conversation_list_queried_callback(System.IntPtr handle,
            System.IntPtr conversations_list_ptr, uint conversation_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationListQueriedCallback callback = null;

            lock (ZIMImpl.conversationListQueriedCallbackLock)
            {
                zim.conversationListQueriedCallback.TryGetValue(sequence, out callback);
                zim.conversationListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_conversation[] zim_conversations_list = new zim_conversation[conversation_list_length];
            ZIMUtil.GetStructListByPtr<zim_conversation>(ref zim_conversations_list, conversations_list_ptr, conversation_list_length);

            List<ZIMConversation> conversationList = ZIMConverter.ChangeZIMConversationStructListToClassList(zim_conversations_list);

            callback.BeginInvoke(conversationList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_pinned_list_queried_callback))]
        public static void zim_on_conversation_pinned_list_queried_callback(System.IntPtr handle,
            System.IntPtr conversations_list_ptr, uint conversation_list_length, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationPinnedListQueriedCallback callback = null;

            lock (ZIMImpl.conversationPinnedListQueriedCallbackLock)
            {
                zim.conversationPinnedListQueriedCallback.TryGetValue(sequence, out callback);
                zim.conversationPinnedListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_conversation[] zim_conversations_list = new zim_conversation[conversation_list_length];
            ZIMUtil.GetStructListByPtr<zim_conversation>(ref zim_conversations_list, conversations_list_ptr, conversation_list_length);

            List<ZIMConversation> conversationList = ZIMConverter.ChangeZIMConversationStructListToClassList(zim_conversations_list);

            callback.BeginInvoke(conversationList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_pinned_state_updated_callback))]
        public static void zim_on_conversation_pinned_state_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id, System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationPinnedStateUpdatedCallback callback = null;

            lock (ZIMImpl.conversationPinnedStateUpdatedCallbackLock)
            {
                zim.conversationPinnedStateUpdatedCallback.TryGetValue(sequence, out callback);
                zim.conversationPinnedStateUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_deleted_callback))]
        public static void zim_on_conversation_deleted_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id,
            System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationDeletedCallback callback = null;

            lock (ZIMImpl.conversationDeletedCallbackLock)
            {
                zim.conversationDeletedCallback.TryGetValue(sequence, out callback);
                zim.conversationDeletedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_local_extended_data_updated_callback))]
        public static void zim_on_message_local_extended_data_updated_callback(System.IntPtr handle,
        zim_message message,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageLocalExtendedDataUpdatedCallback callback = null;

            lock (ZIMImpl.messageLocalExtendedDataUpdatedCallbackLock)
            {
                zim.messageLocalExtendedDataUpdatedCallback.TryGetValue(sequence, out callback);
                zim.messageLocalExtendedDataUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMMessage cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);

            callback.BeginInvoke(cMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_messages_searched_callback))]
        public static void zim_on_messages_searched_callback(
        System.IntPtr handle, [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id, System.Int32 conversation_type,
        System.IntPtr message_list_ptr, uint message_list_length,
         ref zim_message next_message, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessagesSearchedCallback callback = null;

            lock (ZIMImpl.messagesSearchedCallbackLock)
            {
                zim.messagesSearchedCallback.TryGetValue(sequence, out callback);
                zim.messagesSearchedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<ZIMMessage> messageList = new List<ZIMMessage>();

            zim_message[] messages = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref messages, message_list_ptr, message_list_length);
            for (int i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(messages[i]));
            }

            ZIMMessage nextMessage = null;

            if(next_message.type != 0)
            {
                nextMessage = ZIMConverter.ChangeZIMMessageStructToClass(next_message);
            }

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, messageList, nextMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_messages_global_searched_callback))]
        public static void zim_on_messages_global_searched_callback(
        System.IntPtr handle, System.IntPtr message_list_ptr, uint message_list_length,
         ref zim_message next_message, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessagesGlobalSearchedCallback callback = null;

            lock (ZIMImpl.messagesGlobalSearchedCallbackLock)
            {
                zim.messagesGlobalSearchedCallback.TryGetValue(sequence, out callback);
                zim.messagesGlobalSearchedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<ZIMMessage> messageList = new List<ZIMMessage>();

            zim_message[] messages = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref messages, message_list_ptr, message_list_length);
            for (int i = 0; i < message_list_length; i++)
            {
                messageList.Add(ZIMConverter.ChangeZIMMessageStructToClass(messages[i]));
            }

            ZIMMessage nextMessage = null;

            if (next_message.type != 0)
            {
                nextMessage = ZIMConverter.ChangeZIMMessageStructToClass(next_message);
            }

            callback.BeginInvoke(messageList, nextMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversations_searched_callback))]
        public static void zim_on_conversations_searched_callback(
        System.IntPtr handle, System.IntPtr conversation_search_info_list_ptr, uint conversation_search_info_list_length, uint next_flag,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationsSearchedCallback callback = null;

            lock (ZIMImpl.conversationsSearchedCallbackLock)
            {
                zim.conversationsSearchedCallback.TryGetValue(sequence, out callback);
                zim.conversationsSearchedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<ZIMConversationSearchInfo> globalInfoList = new List<ZIMConversationSearchInfo>();

            zim_conversation_search_info[] conversation_search_infos = new zim_conversation_search_info[conversation_search_info_list_length];
            ZIMUtil.GetStructListByPtr<zim_conversation_search_info>(ref conversation_search_infos, conversation_search_info_list_ptr, conversation_search_info_list_length);
            for (int i = 0; i < conversation_search_info_list_length; i++)
            {
                globalInfoList.Add(ZIMConverter.ChangeZIMConversationSearchInfoStructToClass(conversation_search_infos[i]));
            }

            callback.BeginInvoke(globalInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.on_message_reaction_added_callback))]
        public static void on_message_reaction_added_callback(
        System.IntPtr handle, zim_message_reaction reaction,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageReactionAddedCallback callback = null;

            lock (ZIMImpl.messageReactionAddedCallbackLock)
            {
                zim.messageReactionAddedCallback.TryGetValue(sequence, out callback);
                zim.messageReactionAddedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMMessageReaction cReaction = ZIMConverter.ChangeZIMMessageReactionStructToClass(reaction);

            callback.BeginInvoke(cReaction, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.on_message_reaction_deleted_callback))]
        public static void on_message_reaction_deleted_callback(
        System.IntPtr handle, zim_message_reaction reaction,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageReactionDeletedCallback callback = null;

            lock (ZIMImpl.messageReactionDeletedCallbackLock)
            {
                zim.messageReactionDeletedCallback.TryGetValue(sequence, out callback);
                zim.messageReactionDeletedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMMessageReaction cReaction = ZIMConverter.ChangeZIMMessageReactionStructToClass(reaction);

            callback.BeginInvoke(cReaction, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.on_message_reaction_users_queried_callback))]
        public static void on_message_reaction_users_queried_callback(
        System.IntPtr handle, zim_message message, [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string reaction_type,
        System.IntPtr user_infos_ptr, uint user_info_length, long next_flag, uint total_count,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageReactionUserListQueriedCallback callback = null;

            lock (ZIMImpl.messageReactionUserListQueriedCallbackLock)
            {
                zim.messageReactionUserListQueriedCallback.TryGetValue(sequence, out callback);
                zim.messageReactionUserListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_message_reaction_user_info[] user_infos = new zim_message_reaction_user_info[user_info_length];
            ZIMUtil.GetStructListByPtr<zim_message_reaction_user_info>(ref user_infos, user_infos_ptr, user_info_length);

            List<ZIMMessageReactionUserInfo> userList = ZIMConverter.ChangeZIMMessageReactionUserInfoStructListToClassList(user_infos);

            ZIMMessage cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);

            callback.BeginInvoke(cMessage, userList, reaction_type, next_flag, total_count, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_reactions_changed_event))]
        public static void zim_on_message_reactions_changed_event(System.IntPtr handle,
        System.IntPtr reactions_ptr, uint reaction_length)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_message_reaction[] reactions = new zim_message_reaction[reaction_length];
            ZIMUtil.GetStructListByPtr<zim_message_reaction>(ref reactions, reactions_ptr, reaction_length);

            List<ZIMMessageReaction> messageReactionList = ZIMConverter.ChangeZIMMessageReactionStructListToClassList(reactions);

            if (zim.onMessageReactionsChanged != null)
                zim.onMessageReactionsChanged.BeginInvoke(zim, messageReactionList, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_groups_searched_callback))]
        public static void zim_on_groups_searched_callback(
        System.IntPtr handle, System.IntPtr group_search_infos_ptr, uint group_search_info_list_length, uint next_flag,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupsSearchedCallback callback = null;

            lock (ZIMImpl.groupsSearchedCallbackLock)
            {
                zim.groupsSearchedCallback.TryGetValue(sequence, out callback);
                zim.groupsSearchedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_group_search_info[] group_search_infos = new zim_group_search_info[group_search_info_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_search_info>(ref group_search_infos, group_search_infos_ptr, group_search_info_list_length);

            List<ZIMGroupSearchInfo> groupSearchInfoList = ZIMConverter.ChangeZIMGroupSearchInfoStructListToClassList(group_search_infos);

            callback.BeginInvoke(groupSearchInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_members_searched_callback))]
        public static void zim_on_group_members_searched_callback(
        System.IntPtr handle, [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
        System.IntPtr user_list_ptr, uint user_list_length, uint next_flag,
        zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMembersSearchedCallback callback = null;

            lock (ZIMImpl.groupMembersSearchedCallbackLock)
            {
                zim.groupMembersSearchedCallback.TryGetValue(sequence, out callback);
                zim.groupMembersSearchedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_group_member_info[] group_member_infos = new zim_group_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos, user_list_ptr, user_list_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos);

            callback.BeginInvoke(group_id, groupMemberInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_calling_invitation_sent_callback))]
        public static void zim_on_calling_invitation_sent_callback(System.IntPtr handle,
                    [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
                    zim_calling_invitation_sent_info info,
                    zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallingInvitationSentCallback callback = null;

            lock (ZIMImpl.callingInvitationSentCallbackLock)
            {
                zim.callingInvitationSentCallback.TryGetValue(sequence, out callback);
                zim.callingInvitationSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMCallingInvitationSentInfo callingInvitationSentInfo = new ZIMCallingInvitationSentInfo();
            uint error_user_list_length = info.error_invitees_length;
            zim_error_user_info[] error_user_infos = new zim_error_user_info[error_user_list_length];
            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos, info.error_invitees, error_user_list_length);

            callingInvitationSentInfo.errorUserList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos);

            callback.BeginInvoke(call_id, callingInvitationSentInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_quit_sent_callback))]
        public static void zim_on_call_quit_sent_callback(System.IntPtr handle,
                    [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
                    zim_call_quit_sent_info info,
                    zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallQuitSentCallback callback = null;

            lock (ZIMImpl.callQuitSentCallbackLock)
            {
                zim.callQuitSentCallback.TryGetValue(sequence, out callback);
                zim.callQuitSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMCallQuitSentInfo callQuitSentInfo = ZIMConverter.ChangeZIMCallQuitSentInfoStructToClass(info);

            callback.BeginInvoke(call_id, callQuitSentInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_end_sent_callback))]
        public static void zim_on_call_end_sent_callback(System.IntPtr handle,
                    [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
                    zim_call_end_sent_info info,
                    zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallEndSentCallback callback = null;

            lock (ZIMImpl.callEndSentCallbackLock)
            {
                zim.callEndSentCallback.TryGetValue(sequence, out callback);
                zim.callEndSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            ZIMCallEndedSentInfo callEndSentInfo = ZIMConverter.ChangeZIMCallEndedSentInfoStructToClass(info);

            callback.BeginInvoke(call_id, callEndSentInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_list_queried_callback))]
        public static void zim_on_call_list_queried_callback(System.IntPtr handle,
                    System.IntPtr call_list_ptr, uint call_list_length, long next_flag,
                    zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationListQueriedCallback callback = null;

            lock (ZIMImpl.callInvitationListQueriedCallbackLock)
            {
                zim.callInvitationListQueriedCallback.TryGetValue(sequence, out callback);
                zim.callInvitationListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_call_info[] call_Infos = new zim_call_info[call_list_length];
            ZIMUtil.GetStructListByPtr<zim_call_info>(ref call_Infos, call_list_ptr, call_list_length);

            List<ZIMCallInfo> callInfoList = ZIMConverter.ChangeZIMCallInfoStructListToClassList(call_Infos);

            callback.BeginInvoke(callInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_broadcast_message_received_event))]
        public static void zim_on_broadcast_message_received_event(System.IntPtr handle,
        zim_message message)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessage cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);

            if (zim.onBroadcastMessageReceived != null)
                zim.onBroadcastMessageReceived.BeginInvoke(zim, cMessage, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_ended_event))]
        public static void zim_on_call_invitation_ended_event(System.IntPtr handle,
            zim_call_invitation_ended_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            var Info = ZIMConverter.ChangeZIMCallInvitationEndedInfoStructToClass(info);

            if (zim.onCallInvitationEnded != null)
                zim.onCallInvitationEnded.BeginInvoke(zim, Info, call_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_user_state_changed_event))]
        public static void zim_on_call_user_state_changed_event(System.IntPtr handle,
            zim_event_call_user_state_changed_info info)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            var cInfo = ZIMConverter.ChangeZIMCallUserStateChangeInfoStructToClass(info);

            if (zim.onCallUserStateChanged != null)
                zim.onCallUserStateChanged.BeginInvoke(zim, cInfo, info.call_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_unread_message_count_cleared_callback))]
        public static void zim_on_conversation_unread_message_count_cleared_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id, System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationUnreadMessageCountClearedCallback callback = null;

            lock (ZIMImpl.conversationUnreadMessageCountClearedCallbackLock)
            {
                zim.conversationUnreadMessageCountClearedCallback.TryGetValue(sequence, out callback);
                zim.conversationUnreadMessageCountClearedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_message_receipt_read_sent_callback))]
        public static void zim_on_conversation_message_receipt_read_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string conversation_id, System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationMessageReceiptReadSentCallback callback = null;

            lock (ZIMImpl.conversationMessageReceiptReadSentCallbackLock)
            {
                zim.conversationMessageReceiptReadSentCallback.TryGetValue(sequence, out callback);
                zim.conversationMessageReceiptReadSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(conversation_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_changed_event))]
        public static void zim_on_conversation_changed_event(System.IntPtr handle,
            System.IntPtr infos_ptr, uint infos_length)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_conversation_change_info[] zim_conversation_change_infos_list = new zim_conversation_change_info[infos_length];
            ZIMUtil.GetStructListByPtr<zim_conversation_change_info>(ref zim_conversation_change_infos_list, infos_ptr, infos_length);

            List<ZIMConversationChangeInfo> conversationChangeInfoList = ZIMConverter.ChangeZIMConversationChangeInfoStructListToClassList(zim_conversation_change_infos_list);

            if (zim.onConversationChanged != null)
                zim.onConversationChanged.BeginInvoke(zim, conversationChangeInfoList, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_total_unread_message_count_updated_event))]
        public static void zim_on_conversation_total_unread_message_count_updated_event(System.IntPtr handle,
            uint total_unread_message_count)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            if (zim.onConversationTotalUnreadMessageCountUpdated != null)
                zim.onConversationTotalUnreadMessageCountUpdated(zim, total_unread_message_count);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_created_callback))]
        public static void zim_on_room_created_callback(System.IntPtr handle,
            zim_room_full_info room_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomCreatedCallback callback = null;

            lock (ZIMImpl.roomCreatedCallbackLock)
            {
                zim.roomCreatedCallback.TryGetValue(sequence, out callback);
                zim.roomCreatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            var roomFullInfo = ZIMConverter.ChangeZIMRoomFullInfoStructToClass(room_info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(roomFullInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_joined_callback))]
        public static void zim_on_room_joined_callback(System.IntPtr handle,
            zim_room_full_info room_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomJoinedCallback callback = null;

            lock (ZIMImpl.roomJoinedCallbackLock)
            {
                zim.roomJoinedCallback.TryGetValue(sequence, out callback);
                zim.roomJoinedCallback.Remove(sequence);
            }

            var roomFullInfo = ZIMConverter.ChangeZIMRoomFullInfoStructToClass(room_info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(roomFullInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_entered_callback))]
        public static void zim_on_room_entered_callback(System.IntPtr handle,
            zim_room_full_info room_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomEnteredCallback callback = null;

            lock (ZIMImpl.roomEnteredCallbackLock)
            {
                zim.roomEnteredCallback.TryGetValue(sequence, out callback);
                zim.roomEnteredCallback.Remove(sequence);
            }

            if (callback == null) return;

            var roomFullInfo = ZIMConverter.ChangeZIMRoomFullInfoStructToClass(room_info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(roomFullInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_left_callback))]
        public static void zim_on_room_left_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomLeftCallback callback = null;

            lock (ZIMImpl.roomLeftCallbackLock)
            {
                zim.roomLeftCallback.TryGetValue(sequence, out callback);
                zim.roomLeftCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_member_queried_callback))]
        public static void zim_on_room_member_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr user_list_ptr, uint user_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string next_flag,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomMemberQueriedCallback callback = null;

            lock (ZIMImpl.roomMemberQueriedCallbackLock)
            {
                zim.roomMemberQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomMemberQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_user_info[] user_infos_list = new zim_user_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_user_info>(ref user_infos_list, user_list_ptr, user_list_length);

            List<ZIMUserInfo> memberList = ZIMConverter.ChangeZIMUserInfoStructListToClassList(user_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, memberList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_members_queried))]
        public static void zim_on_room_members_queried(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr user_list_ptr, uint user_list_length,
            System.IntPtr error_user_list_ptr, uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomMembersQueriedCallback callback = null;

            lock (ZIMImpl.roomMembersQueriedCallbackLock)
            {
                zim.roomMembersQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomMembersQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;


            zim_room_member_info[] user_infos_list = new zim_room_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_room_member_info>(ref user_infos_list, user_list_ptr, user_list_length);

            List<ZIMRoomMemberInfo> memberList = ZIMConverter.ChangeZIMRoomMemberInfoStructListToClassList(user_infos_list);


            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            zim_error_user_info[] error_user_infos_list = new zim_error_user_info[error_user_list_length];

            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos_list, error_user_list_ptr, error_user_list_length);

            List<ZIMErrorUserInfo> errorUserInfoList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos_list);

            callback.BeginInvoke(room_id, memberList, errorUserInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_online_member_count_queried_callback))]
        public static void zim_on_room_online_member_count_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            uint count, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomOnlineMemberCountQueriedCallback callback = null;
            lock (ZIMImpl.roomOnlineMemberCountQueriedCallbackLock)
            {
                zim.roomOnlineMemberCountQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomOnlineMemberCountQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, count, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_member_joined_event))]
        public static void zim_on_room_member_joined_event(System.IntPtr handle,
            System.IntPtr member_list_ptr, uint member_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_user_info[] user_infos_list = new zim_user_info[member_list_length];
            ZIMUtil.GetStructListByPtr<zim_user_info>(ref user_infos_list, member_list_ptr, member_list_length);

            List<ZIMUserInfo> memberList = ZIMConverter.ChangeZIMUserInfoStructListToClassList(user_infos_list);

            if (zim.onRoomMemberJoined != null)
                zim.onRoomMemberJoined.BeginInvoke(zim, memberList, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_member_left_event))]
        public static void zim_on_room_member_left_event(System.IntPtr handle,
            System.IntPtr member_list_ptr, uint member_list_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_user_info[] user_infos_list = new zim_user_info[member_list_length];
            ZIMUtil.GetStructListByPtr<zim_user_info>(ref user_infos_list, member_list_ptr, member_list_length);

            List<ZIMUserInfo> memberList = ZIMConverter.ChangeZIMUserInfoStructListToClassList(user_infos_list);

            if (zim.onRoomMemberLeft != null)
                zim.onRoomMemberLeft.BeginInvoke(zim, memberList, room_id, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_attributes_operated_callback))]
        public static void zim_on_room_attributes_operated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr error_key_list_ptr, uint error_key_list_length, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomAttributesOperatedCallback callback = null;

            lock (ZIMImpl.roomAttributesOperatedCallbackLock)
            {
                zim.roomAttributesOperatedCallback.TryGetValue(sequence, out callback);
                zim.roomAttributesOperatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<string> errorKeyList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[error_key_list_length];
            if (error_key_list_length != 0)
                Marshal.Copy(error_key_list_ptr, ptrArray, 0, (int)error_key_list_length);
            for (int i = 0; i < error_key_list_length; i++)
            {
                errorKeyList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            callback.BeginInvoke(room_id, errorKeyList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_attributes_queried_callback))]
        public static void zim_on_room_attributes_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr room_attributes,
            uint room_attributes_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomAttributesQueriedCallback callback = null;
            lock (ZIMImpl.roomAttributesQueriedCallbackLock)
            {
                zim.roomAttributesQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomAttributesQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            var roomAttributes = ZIMConverter.ChangeZIMRoomAttributesPtrToDictionary(room_attributes, room_attributes_length);

            callback.BeginInvoke(room_id, roomAttributes, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_attributes_batch_operated_callback))]
        public static void zim_on_room_attributes_batch_operated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomAttributesBatchOperatedCallback callback = null;

            lock (ZIMImpl.roomAttributesBatchOperatedCallbackLock)
            {
                zim.roomAttributesBatchOperatedCallback.TryGetValue(sequence, out callback);
                zim.roomAttributesBatchOperatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_members_attributes_operated_callback))]
        public static void zim_on_room_members_attributes_operated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr infos_ptr, uint infos_length,
            System.IntPtr error_user_list_ptr, uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomMembersAttributesOperatedCallback callback = null;
            lock (ZIMImpl.roomMembersAttributesOperatedCallbackLock)
            {
                zim.roomMembersAttributesOperatedCallback.TryGetValue(sequence, out callback);
                zim.roomMembersAttributesOperatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_room_member_attributes_operated_info[] room_member_attributes_operated_infos_list =
                new zim_room_member_attributes_operated_info[infos_length];

            ZIMUtil.GetStructListByPtr<zim_room_member_attributes_operated_info>(ref room_member_attributes_operated_infos_list, infos_ptr, infos_length);


            List<ZIMRoomMemberAttributesOperatedInfo> roomMemberAttributesUpdateInfoList = ZIMConverter.ChangeZIMRoomMemberAttributesOperatedInfoStructListToClassList(room_member_attributes_operated_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<string> errorUserList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[error_user_list_length];
            if (error_user_list_length != 0)
                Marshal.Copy(error_user_list_ptr, ptrArray, 0, (int)error_user_list_length);
            for (int i = 0; i < error_user_list_length; i++)
            {
                errorUserList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            callback.BeginInvoke(room_id, roomMemberAttributesUpdateInfoList, errorUserList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_members_attributes_queried_callback))]
        public static void zim_on_room_members_attributes_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr infos_ptr, uint infos_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomMembersAttributesQueriedCallback callback = null;
            lock (ZIMImpl.roomMembersAttributesQueriedCallbackLock)
            {
                zim.roomMembersAttributesQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomMembersAttributesQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            List<ZIMRoomMemberAttributesInfo> roomMemberAttributesInfoList = new List<ZIMRoomMemberAttributesInfo>();
            zim_room_member_attributes_info[] room_member_attributes_infos_list = new zim_room_member_attributes_info[infos_length];
            ZIMUtil.GetStructListByPtr<zim_room_member_attributes_info>(ref room_member_attributes_infos_list, infos_ptr, infos_length);

            for (int i = 0; i < infos_length; i++)
            {
                var roomMemberAttributesInfo = new ZIMRoomMemberAttributesInfo();
                roomMemberAttributesInfo.userID = room_member_attributes_infos_list[i].user_id;
                roomMemberAttributesInfo.attributes = ZIMConverter.ChangeZIMRoomMemberAttributesPtrToDictionary(room_member_attributes_infos_list[i].attributes,
                    room_member_attributes_infos_list[i].attributes_length);
                roomMemberAttributesInfoList.Add(roomMemberAttributesInfo);
            }

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, roomMemberAttributesInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_room_member_attributes_list_queried_callback))]
        public static void zim_on_room_member_attributes_list_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string room_id,
            System.IntPtr infos_ptr, uint infos_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string next_flag,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMRoomMemberAttributesListQueriedCallback callback = null;

            lock (ZIMImpl.roomMemberAttributesListQueriedCallbackLock)
            {
                zim.roomMemberAttributesListQueriedCallback.TryGetValue(sequence, out callback);
                zim.roomMemberAttributesListQueriedCallback.Remove(sequence);
            }

            List<ZIMRoomMemberAttributesInfo> roomMemberAttributesInfoList = new List<ZIMRoomMemberAttributesInfo>();
            zim_room_member_attributes_info[] room_member_attributes_infos = new zim_room_member_attributes_info[infos_length];
            ZIMUtil.GetStructListByPtr<zim_room_member_attributes_info>(ref room_member_attributes_infos, infos_ptr, infos_length);
            for (int i = 0; i < infos_length; i++)
            {
                var roomMemberAttributesInfo = new ZIMRoomMemberAttributesInfo();
                roomMemberAttributesInfo.userID = room_member_attributes_infos[i].user_id;
                roomMemberAttributesInfo.attributes = ZIMConverter.ChangeZIMRoomMemberAttributesPtrToDictionary(room_member_attributes_infos[i].attributes,
                    room_member_attributes_infos[i].attributes_length);
                roomMemberAttributesInfoList.Add(roomMemberAttributesInfo);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(room_id, roomMemberAttributesInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_created_callback))]
        public static void zim_on_group_created_callback(System.IntPtr handle,
            zim_group_full_info group_info, IntPtr user_list_ptr,
            uint user_list_length, IntPtr error_user_list_ptr,
            uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupCreatedCallback callback = null;

            lock (ZIMImpl.groupCreatedCallbackLock)
            {
                zim.groupCreatedCallback.TryGetValue(sequence, out callback);
                zim.groupCreatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMGroupFullInfo groupInfo = ZIMConverter.ChangeZIMGroupFullInfoStructToClass(group_info);

            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_list_length];

            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_list_ptr, user_list_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            zim_error_user_info[] error_user_infos_list = new zim_error_user_info[error_user_list_length];

            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos_list, error_user_list_ptr, error_user_list_length);

            List<ZIMErrorUserInfo> errorUserInfoList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(groupInfo, groupMemberInfoList, errorUserInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_dismissed_callback))]
        public static void zim_on_group_dismissed_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupDismissedCallback callback = null;
            lock (ZIMImpl.groupDismissedCallbackLock)
            {
                zim.groupDismissedCallback.TryGetValue(sequence, out callback);
                zim.groupDismissedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_joined_callback))]
        public static void zim_on_group_joined_callback(System.IntPtr handle,
            zim_group_full_info group_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupJoinedCallback callback = null;

            lock (ZIMImpl.groupJoinedCallbackLock)
            {
                zim.groupJoinedCallback.TryGetValue(sequence, out callback);
                zim.groupJoinedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMGroupFullInfo groupInfo = ZIMConverter.ChangeZIMGroupFullInfoStructToClass(group_info);


            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(groupInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_left_callback))]
        public static void zim_on_group_left_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupLeftCallback callback = null;

            lock (ZIMImpl.groupLeftCallbackLock)
            {
                zim.groupLeftCallback.TryGetValue(sequence, out callback);
                zim.groupLeftCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_users_invited_callback))]
        public static void zim_on_group_users_invited_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr user_list_ptr, uint user_list_length,
            System.IntPtr error_user_list_ptr, uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupUsersInvitedCallback callback = null;

            lock (ZIMImpl.groupUsersInvitedCallbackLock)
            {
                zim.groupUsersInvitedCallback.TryGetValue(sequence, out callback);
                zim.groupUsersInvitedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_list_ptr, user_list_length);
            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            zim_error_user_info[] error_user_infos_list = new zim_error_user_info[error_user_list_length];

            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos_list, error_user_list_ptr, error_user_list_length);

            List<ZIMErrorUserInfo> errorUserInfoList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, groupMemberInfoList, errorUserInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_kicked_callback))]
        public static void zim_on_group_member_kicked_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr kicked_user_id_list_ptr, uint kicked_user_id_list_length,
            System.IntPtr error_user_list_ptr, uint error_user_list_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberKickedCallback callback = null;

            lock (ZIMImpl.groupMemberKickedoutCallbackLock)
            {
                zim.groupMemberKickedoutCallback.TryGetValue(sequence, out callback);
                zim.groupMemberKickedoutCallback.Remove(sequence);
            }

            if (callback == null) return;

            List<string> kickedUserIDList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[kicked_user_id_list_length];
            if (kicked_user_id_list_length != 0)
                Marshal.Copy(kicked_user_id_list_ptr, ptrArray, 0, (int)kicked_user_id_list_length);
            for (int i = 0; i < kicked_user_id_list_length; i++)
            {
                kickedUserIDList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            zim_error_user_info[] error_user_infos_list = new zim_error_user_info[error_user_list_length];
            ZIMUtil.GetStructListByPtr<zim_error_user_info>(ref error_user_infos_list, error_user_list_ptr, error_user_list_length);
            List<ZIMErrorUserInfo> errorUserInfoList = ZIMConverter.ChangeZIMErrorUserInfoStructListToClassList(error_user_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, kickedUserIDList, errorUserInfoList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_owner_transferred_callback))]
        public static void zim_on_group_owner_transferred_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string to_user_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOwnerTransferredCallback callback = null;
            lock (ZIMImpl.groupOwnerTransferredCallbackLock)
            {
                zim.groupOwnerTransferredCallback.TryGetValue(sequence, out callback);
                zim.groupOwnerTransferredCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, to_user_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_name_updated_callback))]
        public static void zim_on_group_name_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_name,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupNameUpdatedCallback callback = null;

            lock (ZIMImpl.groupNameUpdatedCallbackLock)
            {
                zim.groupNameUpdatedCallback.TryGetValue(sequence, out callback);
                zim.groupNameUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, group_name, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_avatar_url_updated_callback))]
        public static void zim_on_group_avatar_url_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_avatar_url,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupAvatarUrlUpdatedCallback callback = null;
            lock (ZIMImpl.groupAvatarUrlUpdatedCallbackLock)
            {
                zim.groupAvatarUrlUpdatedCallback.TryGetValue(sequence, out callback);
                zim.groupAvatarUrlUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, group_avatar_url, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_notice_updated_callback))]
        public static void zim_on_group_notice_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_notice,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupNoticeUpdatedCallback callback = null;

            lock (ZIMImpl.groupNoticeUpdatedCallbackLock)
            {
                zim.groupNoticeUpdatedCallback.TryGetValue(sequence, out callback);
                zim.groupNoticeUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, group_notice, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_info_queried_callback))]
        public static void zim_on_group_info_queried_callback(System.IntPtr handle,
            zim_group_full_info group_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupInfoQueriedCallback callback = null;

            lock (ZIMImpl.groupInfoQueriedCallbackLock)
            {
                zim.groupInfoQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupInfoQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            var groupInfo = ZIMConverter.ChangeZIMGroupFullInfoStructToClass(group_info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(groupInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_attributes_operated_callback))]
        public static void zim_on_group_attributes_operated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr error_keys_ptr, uint error_keys_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupAttributesOperatedCallback callback = null;

            lock (ZIMImpl.groupAttributesOperatedCallbackLock)
            {
                zim.groupAttributesOperatedCallback.TryGetValue(sequence, out callback);
                zim.groupAttributesOperatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            List<string> errorKeyList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[error_keys_length];
            if (error_keys_length != 0)
                Marshal.Copy(error_keys_ptr, ptrArray, 0, (int)error_keys_length);
            for (int i = 0; i < error_keys_length; i++)
            {
                errorKeyList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;


            callback.BeginInvoke(group_id, errorKeyList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_attributes_queried_callback))]
        public static void zim_on_group_attributes_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr group_attributes_ptr, uint group_attributes_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupAttributesQueriedCallback callback = null;

            lock (ZIMImpl.groupAttributesQueriedCallbackLock)
            {
                zim.groupAttributesQueriedCallback.TryGetValue(sequence, out callback);
            }

            if (callback == null) return;

            var groupAttributes = ZIMConverter.ChangeZIMGroupAttributesStructListPtrToDictionary(group_attributes_ptr, group_attributes_length);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, groupAttributes, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_nickname_updated_callback))]
        public static void zim_on_group_member_nickname_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string for_user_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string nickname,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberNicknameUpdatedCallback callback = null;
            lock (ZIMImpl.groupMemberNicknameUpdatedCallbackLock)
            {
                zim.groupMemberNicknameUpdatedCallback.TryGetValue(sequence, out callback);
                zim.groupMemberNicknameUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, for_user_id, nickname, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_role_updated_callback))]
        public static void zim_on_group_member_role_updated_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string for_user_id,
            int role,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberRoleUpdatedCallback callback = null;

            lock (ZIMImpl.groupMemberRoleUpdatedCallbackLock)
            {
                zim.groupMemberRoleUpdatedCallback.TryGetValue(sequence, out callback);
                zim.groupMemberRoleUpdatedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, for_user_id, role, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_info_queried_callback))]
        public static void zim_on_group_member_info_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            zim_group_member_info user_info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberInfoQueriedCallback callback = null;
            lock (ZIMImpl.groupMemberInfoQueriedCallbackLock)
            {
                zim.groupMemberInfoQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupMemberInfoQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            var groupMemberInfo = ZIMConverter.ChangeZIMGroupMemberInfoStructToClass(user_info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, groupMemberInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_list_queried_callback))]
        public static void zim_on_group_list_queried_callback(System.IntPtr handle,
            System.IntPtr group_list_ptr,
            uint group_list_length, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupListQueriedCallback callback = null;

            lock (ZIMImpl.groupListQueriedCallbackLock)
            {
                zim.groupListQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_group[] groups_list = new zim_group[group_list_length];
            ZIMUtil.GetStructListByPtr<zim_group>(ref groups_list, group_list_ptr, group_list_length);
            List<ZIMGroup> groupList = ZIMConverter.ChangeZIMGroupStructListToClassList(groups_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(groupList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_list_queried_callback))]
        public static void zim_on_group_member_list_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr user_list_ptr,
            uint user_list_length, uint next_flag,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberListQueriedCallback callback = null;

            lock (ZIMImpl.groupMemberListQueriedCallbackLock)
            {
                zim.groupMemberListQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupMemberListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_list_ptr, user_list_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, groupMemberInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_count_queried_callback))]
        public static void zim_on_group_member_count_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            uint count, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMemberCountQueriedCallback callback = null;
            lock (ZIMImpl.groupMemberCountQueriedCallbackLock)
            {
                zim.groupMemberCountQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupMemberCountQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, count, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_message_receipt_member_list_queried_callback))]
        public static void zim_on_group_message_receipt_member_list_queried_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id,
            System.IntPtr user_list_ptr, uint user_list_length, uint next_flag,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupMessageReceiptMemberListQueriedCallback callback = null;

            lock (ZIMImpl.groupMessageReceiptMemberListQueriedCallbackLock)
            {
                zim.groupMessageReceiptMemberListQueriedCallback.TryGetValue(sequence, out callback);
                zim.groupMessageReceiptMemberListQueriedCallback.Remove(sequence);
            }

            if (callback == null) return;

            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_list_ptr, user_list_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(group_id, groupMemberInfoList, next_flag, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_conversation_notification_status_set_callback))]
        public static void zim_on_conversation_notification_status_set_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string converstaion_id, System.Int32 conversation_type,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMConversationNotificationStatusSetCallback callback = null;

            lock (ZIMImpl.conversationNotificationStatusSetCallbackLock)
            {
                zim.conversationNotificationStatusSetCallback.TryGetValue(sequence, out callback);
                zim.conversationNotificationStatusSetCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(converstaion_id, (ZIMConversationType)conversation_type, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_state_changed_event))]
        public static void zim_on_group_state_changed_event(System.IntPtr handle,
            System.Int32 state, System.Int32 group_event,
            zim_group_operated_info operated_info,
            zim_group_full_info group_info)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            var groupFullInfo = ZIMConverter.ChangeZIMGroupFullInfoStructToClass(group_info);

            var groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            if (zim.onGroupStateChanged != null)
                zim.onGroupStateChanged(zim, (ZIMGroupState)state,
                    (ZIMGroupEvent)group_event, groupOperatedInfo, groupFullInfo);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_state_changed_event))]
        public static void zim_on_group_member_state_changed_event(System.IntPtr handle,
            System.Int32 state, System.Int32 group_member_event,
            System.IntPtr user_list_ptr, uint user_list_length,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_list_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_list_ptr, user_list_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            if (zim.onGroupMemberStateChanged != null)
                zim.onGroupMemberStateChanged(zim, (ZIMGroupMemberState)state, (ZIMGroupMemberEvent)group_member_event,
                   groupMemberInfoList, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_name_updated_event))]
        public static void zim_on_group_name_updated_event(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_name,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            if (zim.onGroupNameUpdated != null)
                zim.onGroupNameUpdated(zim, group_name, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_avatar_url_updated_event))]
        public static void zim_on_group_avatar_url_updated_event(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_avatar_url,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            if (zim.onGroupAvatarUrlUpdated != null)
                zim.onGroupAvatarUrlUpdated(zim, group_avatar_url, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_notice_updated_event))]
        public static void zim_on_group_notice_updated_event(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_notice,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            if (zim.onGroupNoticeUpdated != null)
                zim.onGroupNoticeUpdated(zim, group_notice, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_attributes_updated_event))]
        public static void zim_on_group_attributes_updated_event(System.IntPtr handle,
            System.IntPtr infos_ptr, uint infos_length,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMGroupAttributesUpdateInfo> groupAttributesUpdateInfoList = new List<ZIMGroupAttributesUpdateInfo>();
            zim_group_attributes_update_info[] group_attributes_update_infos_list = new zim_group_attributes_update_info[infos_length];
            ZIMUtil.GetStructListByPtr<zim_group_attributes_update_info>(ref group_attributes_update_infos_list, infos_ptr, infos_length);

            for (int i = 0; i < infos_length; i++)
            {
                var size = group_attributes_update_infos_list[i].group_attributes_length;
                var groupAttributesUpdateInfo = new ZIMGroupAttributesUpdateInfo();
                groupAttributesUpdateInfo.action = (ZIMGroupAttributesUpdateAction)group_attributes_update_infos_list[i].action;
                groupAttributesUpdateInfo.groupAttributes = ZIMConverter.ChangeZIMGroupAttributesStructListPtrToDictionary(group_attributes_update_infos_list[i].group_attributes, size);
                groupAttributesUpdateInfoList.Add(groupAttributesUpdateInfo);
            }

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);

            if (zim.onGroupAttributesUpdated != null)
                zim.onGroupAttributesUpdated(zim, groupAttributesUpdateInfoList, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_group_member_info_updated_event))]
        public static void zim_on_group_member_info_updated_event(System.IntPtr handle,
            System.IntPtr user_infos_ptr, uint user_infos_length,
            zim_group_operated_info operated_info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string group_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMGroupOperatedInfo groupOperatedInfo = ZIMConverter.ChangeZIMGroupOperatedInfoStructToClass(operated_info);
            zim_group_member_info[] group_member_infos_list = new zim_group_member_info[user_infos_length];
            ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos_list, user_infos_ptr, user_infos_length);

            List<ZIMGroupMemberInfo> groupMemberInfoList = ZIMConverter.ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos_list);

            if (zim.onGroupMemberInfoUpdated != null)
                zim.onGroupMemberInfoUpdated(zim, groupMemberInfoList, groupOperatedInfo, group_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_received_event))]
        public static void zim_on_call_invitation_received_event(System.IntPtr handle,
            zim_call_invitation_received_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationReceivedInfo callInvitationReceivedInfo = ZIMConverter.
                ChangeZIMCallInvitationReceivedInfoStructToClass(info);

            if (zim.onCallInvitationReceived != null)
                zim.onCallInvitationReceived(zim, callInvitationReceivedInfo, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_cancelled_event))]
        public static void zim_on_call_invitation_cancelled_event(System.IntPtr handle,
            zim_call_invitation_cancelled_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationCancelledInfo callInvitationCancelledInfo = ZIMConverter.
                ChangeZIMCallInvitationCancelledInfoStructToClass(info);

            if (zim.onCallInvitationCancelled != null)
                zim.onCallInvitationCancelled(zim, callInvitationCancelledInfo, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_accepted_event))]
        public static void zim_on_call_invitation_accepted_event(System.IntPtr handle,
            zim_call_invitation_accepted_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationAcceptedInfo callInvitationAcceptedInfo = ZIMConverter.
                ChangeZIMCallInvitationAcceptedInfoStructToClass(info);
            if (zim.onCallInvitationAccepted != null)
                zim.onCallInvitationAccepted(zim, callInvitationAcceptedInfo, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_rejected_event))]
        public static void zim_on_call_invitation_rejected_event(System.IntPtr handle,
            zim_call_invitation_rejected_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationRejectedInfo callInvitationRejectedInfo = ZIMConverter.
                ChangeZIMCallInvitationRejectedInfoStructToClass(info);

            if (zim.onCallInvitationRejected != null)
                zim.onCallInvitationRejected(zim, callInvitationRejectedInfo, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_timeout_event))]
        public static void zim_on_call_invitation_timeout_event(System.IntPtr handle,
            zim_call_invitation_timeout_info info,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationTimeoutInfo callInvitationTimeoutInfo = ZIMConverter.
                ChangeZIMCallInvitationTimeoutInfoStructToClass(info);

            if (zim.onCallInvitationTimeout != null)
                zim.onCallInvitationTimeout(zim, callInvitationTimeoutInfo, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitees_answered_timeout_event))]
        public static void zim_on_call_invitees_answered_timeout_event(System.IntPtr handle,
            System.IntPtr invitees_ptr, uint invitees_length,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<string> inviteeList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[invitees_length];
            if (invitees_length != 0)
                Marshal.Copy(invitees_ptr, ptrArray, 0, (int)invitees_length);
            for (int i = 0; i < invitees_length; i++)
            {
                inviteeList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            if (zim.onCallInviteesAnsweredTimeout != null)
                zim.onCallInviteesAnsweredTimeout(zim, inviteeList, call_id);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_invitation_sent_callback))]
        public static void zim_on_call_invitation_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
            zim_call_invitation_sent_info info,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallInvitationSentCallback callback = null;

            lock (ZIMImpl.callInvitationSentCallbackLock)
            {
                zim.callInvitationSentCallback.TryGetValue(sequence, out callback);
                zim.callInvitationSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMCallInvitationSentInfo callInvitationSentInfo = ZIMConverter.
                ZIMCallInvitationSentInfoStructToClass(info);

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(call_id, callInvitationSentInfo, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_cancel_sent_callback))]
        public static void zim_on_call_cancel_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
            System.IntPtr error_invitees_ptr, uint error_invitees_length,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallCancelSentCallback callback = null;

            lock (ZIMImpl.callCancelSentCallbackLock)
            {
                zim.callCancelSentCallback.TryGetValue(sequence, out callback);
                zim.callCancelSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            List<string> errorInviteeList = new List<string>();
            IntPtr[] ptrArray = new IntPtr[error_invitees_length];
            if (error_invitees_length != 0)
                Marshal.Copy(error_invitees_ptr, ptrArray, 0, (int)error_invitees_length);
            for (int i = 0; i < error_invitees_length; i++)
            {
                errorInviteeList.Add(ZIMUtil.PtrToStringUTF8(ptrArray[i]));
            }

            callback.BeginInvoke(call_id, errorInviteeList, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_acceptance_sent_callback))]
        public static void zim_on_call_acceptance_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallAcceptanceSentCallback callback = null;

            lock (ZIMImpl.callAcceptanceSentCallbackLock)
            {
                zim.callAcceptanceSentCallback.TryGetValue(sequence, out callback);
                zim.callAcceptanceSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(call_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_call_rejection_sent_callback))]
        public static void zim_on_call_rejection_sent_callback(System.IntPtr handle,
            [In()][MarshalAs(UnmanagedType.CustomMarshaler, MarshalTypeRef = typeof(ZIMUtil.UTF8StringMarshaler))] string call_id,
            zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMCallRejectionSentCallback callback = null;

            lock (ZIMImpl.callRejectionSentCallbackLock)
            {
                zim.callRejectionSentCallback.TryGetValue(sequence, out callback);
                zim.callRejectionSentCallback.Remove(sequence);
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            callback.BeginInvoke(call_id, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_media_uploading_progress_callback))]
        public static void zim_on_media_uploading_progress_callback(System.IntPtr handle, zim_message message, long current_file_size,
            long total_file_size, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMediaUploadingProgress callback = null;

            lock (ZIMImpl.mediaMessageSendNotificationCallbackLock)
            {
                if (zim.mediaMessageSendNotificationCallback.ContainsKey(sequence))
                {
                    callback = zim.mediaMessageSendNotificationCallback[sequence].onMediaUploadingProgress;
                }
            }

            ZIMMediaMessageSendNotification notification = null;

            lock (ZIMImpl.mediaMessageSendNotificationCallbackLock)
            {
                if (zim.mediaMessageSendNotificationCallback.ContainsKey(sequence))
                {
                    zim.mediaMessageSendNotificationCallback.TryGetValue(sequence, out notification);
                }
            }

            ZIMMessage cMessage = null;

            lock (ZIMImpl.messageObjMapLock)
            {
                if (zim.messageObjMap.ContainsKey(sequence))
                {
                    zim.messageObjMap.TryGetValue(sequence, out cMessage);
                    zim.messageObjMap.Remove(sequence);
                }
            }

            if (cMessage == null)
            {
                cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);
            }

            if (callback != null)
            {
                callback.BeginInvoke(cMessage, (ulong)current_file_size, (ulong)total_file_size, null, null);
            }
            else
            {
                if (notification != null)
                {
                    if (notification.onMediaUploadingProgress != null)
                    {
                        notification.onMediaUploadingProgress.BeginInvoke(cMessage, (ulong)current_file_size, (ulong)total_file_size, null, null);
                    }
                }
            }

        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_media_downloading_progress_callback))]
        public static void zim_on_media_downloading_progress_callback(System.IntPtr handle, zim_message message,
            long current_file_size, long total_file_size, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMediaDownloadingProgress callback = null;

            lock (ZIMImpl.mediaDownloadingProgressCallbackLock)
            {
                zim.mediaDownloadingProgressCallback.TryGetValue(sequence, out callback);
            }

            if (callback == null) return;

            ZIMMediaMessage cMessage = (ZIMMediaMessage)ZIMConverter.ChangeZIMMessageStructToClass(message);
            callback.BeginInvoke(cMessage, (ulong)current_file_size, (ulong)total_file_size, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_media_downloaded_callback))]
        public static void zim_on_media_downloaded_callback(System.IntPtr handle, zim_message message, zim_error error_info,
                                                 int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMediaDownloadedCallback callback = null;

            ZIMMessage cMessage = null;

            lock (ZIMImpl.mediaDownloadedCallbackLock)
            {
                zim.mediaDownloadedCallback.TryGetValue(sequence, out callback);
                zim.mediaDownloadedCallback.Remove(sequence);
                zim.mediaDownloadingProgressCallback.Remove(sequence);

                lock (ZIMImpl.messageObjMapLock)
                {
                    if (zim.messageObjMap.ContainsKey(sequence))
                    {
                        zim.messageObjMap.TryGetValue(sequence, out cMessage);
                        zim.messageObjMap.Remove(sequence);
                    }
                }
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            if (cMessage == null)
            {
                cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);
            }
            else
            {
                ZIMConverter.ChangeZIMMessageStructToClass(ref cMessage, message);
            }

            callback.BeginInvoke((ZIMMediaMessage)cMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_inserted_callback))]
        public static void zim_on_message_inserted_callback(System.IntPtr handle,
            zim_message message, zim_error error_info, int sequence)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            ZIMMessageInsertedCallback callback = null;
            ZIMMessage cMessage = null;

            lock (ZIMImpl.messageInsertedCallbackLock)
            {
                zim.messageInsertedCallback.TryGetValue(sequence, out callback);
                zim.messageInsertedCallback.Remove(sequence);

                lock (ZIMImpl.messageObjMapLock)
                {
                    zim.messageObjMap.TryGetValue(sequence, out cMessage);
                    zim.messageObjMap.Remove(sequence);
                }
            }

            if (callback == null) return;

            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)error_info.code;
            error.message = error_info.message;

            if (cMessage == null)
            {
                cMessage = ZIMConverter.ChangeZIMMessageStructToClass(message);
            }
            else
            {
                ZIMConverter.ChangeZIMMessageStructToClass(ref cMessage, message);
            }

            callback.BeginInvoke(cMessage, error, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_receipt_changed_event))]
        public static void zim_on_message_receipt_changed_event(System.IntPtr handle,
            bool is_conv_read, System.IntPtr infos_ptr, uint infos_length)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_message_receipt_info[] message_receipt_info_list = new zim_message_receipt_info[infos_length];

            ZIMUtil.GetStructListByPtr<zim_message_receipt_info>(ref message_receipt_info_list, infos_ptr, infos_length);

            List<ZIMMessageReceiptInfo> messageReceiptInfoList = ZIMConverter.ChangeZIMMessageReceiptInfoStructListToClassList(message_receipt_info_list);

            if (is_conv_read)
            {
                if (zim.onConversationMessageReceiptChanged != null)
                    zim.onConversationMessageReceiptChanged.BeginInvoke(zim, messageReceiptInfoList, null, null);
            }
            else
            {
                if (zim.onMessageReceiptChanged != null)
                    zim.onMessageReceiptChanged.BeginInvoke(zim, messageReceiptInfoList, null, null);
            }
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_revoke_received_event))]
        public static void zim_on_message_revoke_received_event(System.IntPtr handle,
            System.IntPtr message_list_ptr, uint message_list_length)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            List<ZIMRevokeMessage> revokeMessageList = new List<ZIMRevokeMessage>();

            zim_message[] messages_list = new zim_message[message_list_length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref messages_list, message_list_ptr, message_list_length);
            for (int i = 0; i < message_list_length; i++)
            {
                revokeMessageList.Add((ZIMRevokeMessage)ZIMConverter.ChangeZIMMessageStructToClass(messages_list[i]));
            }


            if (zim.onMessageRevokeReceived != null)
                zim.onMessageRevokeReceived.BeginInvoke(zim, revokeMessageList, null, null);
        }

        [MonoPInvokeCallback(typeof(IZIMInternal.zim_on_message_sent_status_changed_event))]
        public static void zim_on_message_sent_status_changed_event(System.IntPtr handle,
            System.IntPtr infos_list_ptr, uint infos_list_length)
        {
            ZIMImpl zim = ZIMImpl.GetZIMImplInstance(handle);

            if (zim == null) return;

            zim_message_sent_status_change_info[] message_sent_status_change_infos_list = new zim_message_sent_status_change_info[infos_list_length];

            ZIMUtil.GetStructListByPtr<zim_message_sent_status_change_info>(ref message_sent_status_change_infos_list, infos_list_ptr, infos_list_length);

            List<ZIMMessageSentStatusChangeInfo> messageSentStatusChangeInfoList = ZIMConverter.ChangeZIMMessageSentStatusChangeInfoStructListToClassList(message_sent_status_change_infos_list);

            if (zim.onMessageSentStatusChanged != null)
                zim.onMessageSentStatusChanged.BeginInvoke(zim, messageSentStatusChangeInfoList, null, null);
        }
    }
}
