using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using UnityEngine;
using static ZEGO.ZIMConstans;

namespace ZEGO
{
    class ZIMUtil
    {

        public static Dictionary<string, string> DeserializeStringToDictionary(string jsonStr)
        {
            if (string.IsNullOrEmpty(jsonStr))
                return new Dictionary<string, string>();

            try
            {
                Dictionary<string, string> dict = JsonUtility.FromJson<Dictionary<string, string>>(jsonStr);
                return dict;
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return new Dictionary<string, string>();
            }
        }

        public static void GetExceptionInfoToZIMLog(Exception e)
        {
            string error = "Exception: ";
            error += e.Message;
            error += " StackTrace:";
            error += e.StackTrace;
            ZEGO.IZIMInternal.zim_write_custom_log(error, "Unity");
        }

        public static string PtrToString(IntPtr p)
        {
            if (p == IntPtr.Zero)
                return "";
            try
            {
                return Marshal.PtrToStringAnsi(p);
            }
            catch(Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return "";
            }
        }
        // Convert from IntPtr to UTF-8 string
        public static string PtrToStringUTF8(IntPtr p)
        {
            if (p == IntPtr.Zero)
            {
                return "";
            }
            try
            {
                List<byte> bytes = new List<byte>();
                for (int offset = 0; ; offset++)
                {
                    byte b = Marshal.ReadByte(p, offset);
                    if (b == 0) break;
                    else bytes.Add(b);
                }
                return Encoding.UTF8.GetString(bytes.ToArray(), 0, bytes.Count);
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return "";
            }
        }

        public static IntPtr ArrayToPtr(byte[] array)
        {
            try
            {
                return System.Runtime.InteropServices.Marshal.UnsafeAddrOfPinnedArrayElement(array, 0);
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return IntPtr.Zero;
            }
        }

        public static byte[] PtrToArray(IntPtr ptr, uint length)
        {
            byte[] b = new byte[length];
            try
            {
                Marshal.Copy(ptr, b, 0, (int)length);
            } catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
            }
            return b;
        }

        public static void GetStructListByPtr<StructType>(ref StructType[] list, IntPtr ptr, uint count)
        {
            if (ptr == IntPtr.Zero) return;
            for (int i = 0; i < count; ++i)
            {
                try
                {
                    IntPtr item_ptr = new IntPtr(ptr.ToInt64() + Marshal.SizeOf(typeof(StructType)) * i);

                    if ((list != null) && (item_ptr != null))
                    {
                        list[i] = (StructType)Marshal.PtrToStructure(item_ptr, typeof(StructType));
                    }
                }
                catch (Exception e)
                {
                    GetExceptionInfoToZIMLog(e);
                }
            }
        }

        public static IntPtr AllocGetIntPtrFromStructList<StructType>(StructType[] array)
        {
            int structSize;
            IntPtr structPtr;
            try
            {
                structSize = Marshal.SizeOf(typeof(StructType));
                structPtr = Marshal.AllocHGlobal(array.Length * structSize);
            } catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return IntPtr.Zero;
            }
            IntPtr tempPtr = structPtr;
            for (int count = 0; count < array.Length; count++)
            {
                try
                {
                    Marshal.StructureToPtr(array[count], tempPtr, false);
                    tempPtr += structSize;
                }
                catch (Exception e)
                {
                    GetExceptionInfoToZIMLog(e);
                }
            }
            return structPtr;
        }

        public static IntPtr AllocGetIntPtrFromStructList<StructType>(List<StructType> array)
        {
            int structSize;
            IntPtr structPtr;
            try
            {
                structSize = Marshal.SizeOf(typeof(StructType));
                structPtr = Marshal.AllocHGlobal(array.Count * structSize);
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return IntPtr.Zero;
            }
            IntPtr tempPtr = structPtr;
            for (int count = 0; count < array.Count; count++)
            {
                try
                {
                    Marshal.StructureToPtr(array[count], tempPtr, false);
                    tempPtr += structSize;
                }
                catch (Exception e)
                {
                    GetExceptionInfoToZIMLog(e);
                }
            }
            return structPtr;
        }

        public static (IntPtr, IntPtr[]) AllocGetIntPtrFromStringList(List<string> stringList)
        {

            try
            {

                IntPtr[] stringPointers = new IntPtr[stringList.Count];

                for (int i = 0; i < stringList.Count; i++)
                {
                    stringPointers[i] = Marshal.StringToHGlobalAnsi(stringList[i]);
                }

                IntPtr arrayPointer = Marshal.AllocHGlobal(IntPtr.Size * stringPointers.Length);

                Marshal.Copy(stringPointers, 0, arrayPointer, stringPointers.Length);
                return (arrayPointer, stringPointers);
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return default;
            }
        }

        public static IntPtr AllocCoTaskMemGetIntPtrFromIntArray(int[] src_array)
        {
            IntPtr intArrayPtr = IntPtr.Zero;

            try
            {
                intArrayPtr = Marshal.AllocCoTaskMem(src_array.Length * sizeof(int));
                Marshal.Copy(src_array, 0, intArrayPtr, src_array.Length);

            } catch(Exception e)
            {
                GetExceptionInfoToZIMLog(e);
            }

            return intArrayPtr;
        }

        public static StructType GetStructByPtr<StructType>(IntPtr ptr)
        {
            try
            {
                return (StructType)Marshal.PtrToStructure(ptr, typeof(StructType));
            } catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return default;
            }
        }

        public static IntPtr StructToIntPtr<T>(T info)
        {
            try
            {
                int size = Marshal.SizeOf(info);
                IntPtr intPtr = Marshal.AllocHGlobal(size);
                Marshal.StructureToPtr(info, intPtr, false);
                return intPtr;
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return IntPtr.Zero;
            }
        }

        public static void ModifyFieldsValue<T>(T instance, Type newType, string membername, object Singlevalue = null)
        {
            try
            {
                var pInfo = newType.GetRuntimeProperties().FirstOrDefault(p => p.Name == membername);
                if (pInfo != null && pInfo.CanWrite)
                {
                    pInfo.SetValue(instance, Singlevalue);
                }
                else if (pInfo != null)
                {
                    var backFildInfo = newType.GetRuntimeFields().FirstOrDefault(f => f.Name.Contains($"<{membername}>") && f.Name.Contains("BackingField"));
                    if (backFildInfo != null)
                        backFildInfo.SetValue(instance, Singlevalue);
                }
                else
                {
                    return;
                }
            } catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return;
            }
        }

        public static void ModifyFieldsValue<T>(T instance, string membername, object Singlevalue = null)
        {
            try
            {
                var newType = instance.GetType();
                var pInfo = newType.GetRuntimeProperties().FirstOrDefault(p => p.Name == membername);
                if (pInfo != null && pInfo.CanWrite)
                {
                    //(S)Convert.ChangeType(Singlevalue, typeof(S))
                    pInfo.SetValue(instance, Singlevalue);
                }
                else if (pInfo != null)
                {
                    var backFildInfo = newType.GetRuntimeFields().FirstOrDefault(f => f.Name.Contains($"<{membername}>") && f.Name.Contains("BackingField"));

                    backFildInfo.SetValue(instance, Singlevalue);
                }
                else
                {
                    return;
                }
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
                return;
            }

        }
        public static void ReleaseStructPointer(IntPtr ptr)
        {
            try
            {
                Marshal.FreeHGlobal(ptr);
            } catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
            }
        }

        public static void ReleaseStructPointerArray(IntPtr[] ptrArray)
        {
            try
            {
                foreach(IntPtr ptr in ptrArray)
                {
                    Marshal.FreeHGlobal(ptr);
                }
            }
            catch (Exception e)
            {
                GetExceptionInfoToZIMLog(e);
            }
        }

        public static void ReleaseCoTaskMemPointer(IntPtr stringArrayPtr)
        {
            if (stringArrayPtr != IntPtr.Zero)
            {
                Marshal.FreeCoTaskMem(stringArrayPtr); 
            }
        }

        public class UTF8StringMarshaler : ICustomMarshaler
        {
            public void CleanUpManagedData(object ManagedObj)
            {
            }

            public void CleanUpNativeData(IntPtr pNativeData)
            {
                Marshal.FreeHGlobal(pNativeData);
            }

            public int GetNativeDataSize()
            {
                return -1;
            }

            public IntPtr MarshalManagedToNative(object ManagedObj)
            {
                if (object.ReferenceEquals(ManagedObj, null)) return IntPtr.Zero;
                if (!(ManagedObj is string)) throw new InvalidOperationException();
                byte[] utf8bytes = Encoding.UTF8.GetBytes(ManagedObj as string);
                IntPtr ptr = Marshal.AllocHGlobal(utf8bytes.Length + 1);
                Marshal.Copy(utf8bytes, 0, ptr, utf8bytes.Length);
                Marshal.WriteByte(ptr, utf8bytes.Length, 0);
                return ptr;
            }

            public object MarshalNativeToManaged(IntPtr pNativeData)
            {
                if (pNativeData == IntPtr.Zero) return null;
                List<byte> bytes = new List<byte>();
                for (int offset = 0; ; offset++)
                {
                    byte b = Marshal.ReadByte(pNativeData, offset);
                    if (b == 0) break;
                    else bytes.Add(b);
                }
                return Encoding.UTF8.GetString(bytes.ToArray(), 0, bytes.Count);
            }

            static UTF8StringMarshaler instance = new UTF8StringMarshaler();
            public static ICustomMarshaler GetInstance(string cookie)
            {
                return instance;
            }

        }
        public static string GetUTF8String(byte[] data)
        {
            string result = null;
            for (int i = 0; i < data.Length; i++)
            {
                if (data[i] == 0)
                {
                    result = Encoding.UTF8.GetString(data, 0, i);
                    break;
                }
            }
            return result;
        }

        public static void ZIMCopy(string src, ref byte[] dest, int length)
        {
            if (src != null)
            {
                dest = new byte[length];
                var bytes = Encoding.UTF8.GetBytes(src);
                int count = Math.Min(bytes.Length, length);
                Buffer.BlockCopy(bytes, 0, dest, 0, count);
            }
        }

    }

#pragma warning disable CS0618
    class ZIMConverter
    {
        public static ZIMMessage ChangeZIMMessageStructToClass(zim_message sMessage)
        {
            ZIMMessage cMessage;
            Type baseType;

            if (sMessage.type == (System.Int32)ZIMMessageType.Text)
            {
                string message = sMessage.message;
                cMessage = new ZIMTextMessage(message);
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Command)
            {
                cMessage = new ZIMCommandMessage(ZIMUtil.PtrToArray(sMessage.command_message, sMessage.command_message_length));
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Barrage)
            {
                string message = sMessage.message;
                cMessage = new ZIMBarrageMessage(message);
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Image)
            {
                cMessage = new ZIMImageMessage("");
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var imageMessage = cMessage as ZIMImageMessage;
                imageMessage.largeImageDownloadUrl = sMessage.large_image_download_url;
                imageMessage.thumbnailDownloadUrl = sMessage.thumbnail_download_url;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileSize", sMessage.file_size);
                imageMessage.fileDownloadUrl = sMessage.file_download_url;
                imageMessage.fileLocalPath = sMessage.file_local_path;
                string largeImageDownloadUrl = sMessage.large_image_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageDownloadUrl", largeImageDownloadUrl);
                string largeImageLocalPath = sMessage.large_image_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageLocalPath", largeImageLocalPath);
                string thumbnailDownloadUrl = sMessage.thumbnail_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailDownloadUrl", thumbnailDownloadUrl);

                string thumbnailLocalPath = sMessage.thumbnail_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailLocalPath", thumbnailLocalPath);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "originalImageWidth", sMessage.original_image_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "originalImageHeight", sMessage.original_image_height);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageWidth", sMessage.large_image_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageHeight", sMessage.large_image_height);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailWidth", sMessage.thumbnail_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailHeight", sMessage.thumbnail_height);
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.File)
            {
                cMessage = new ZIMFileMessage("");
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var fileMessage = cMessage as ZIMFileMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileUID", fileUID);

                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileSize", sMessage.file_size);
                fileMessage.fileDownloadUrl = sMessage.file_download_url;
                fileMessage.fileLocalPath = sMessage.file_local_path;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Audio)
            {
                cMessage = new ZIMAudioMessage("", 0);
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var audioMessage = cMessage as ZIMAudioMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileSize", sMessage.file_size);
                audioMessage.fileDownloadUrl = sMessage.file_download_url;
                audioMessage.fileLocalPath = sMessage.file_local_path;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Video)
            {
                cMessage = new ZIMVideoMessage("", 0);
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var videoMessage = cMessage as ZIMVideoMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileSize", sMessage.file_size);

                videoMessage.fileDownloadUrl = sMessage.file_download_url;
                videoMessage.fileLocalPath = sMessage.file_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoDuration", sMessage.media_duration);
                string videoFirstFrameLocalPath = sMessage.video_first_frame_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameLocalPath", videoFirstFrameLocalPath);
                string videoFirstFrameDownloadUrl = sMessage.video_first_frame_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameDownloadUrl", videoFirstFrameDownloadUrl);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameWidth", sMessage.video_first_frame_width);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameHeight", sMessage.video_first_frame_height);
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.System)
            {
                string message = sMessage.message;
                cMessage = new ZIMSystemMessage(message);
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Custom)
            {
                string message = sMessage.message;
                uint subType = sMessage.sub_type;
                cMessage = new ZIMCustomMessage(message, subType);
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Revoke)
            {
                cMessage = new ZIMRevokeMessage();
                baseType = cMessage.GetType().BaseType;
                var revokeMessage = cMessage as ZIMRevokeMessage;
                var revokeType = (ZIMRevokeType)sMessage.revoke_type;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeType", revokeType);
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeTimestamp", sMessage.revoke_timestamp);
                string operatedUserID = sMessage.operated_userid;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "operatedUserID", operatedUserID);
                var originalMessageType = (ZIMMessageType)sMessage.original_message_type;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "originalMessageType", originalMessageType);
                string originalTextMessageContent = sMessage.original_text_message;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "originalTextMessageContent", originalTextMessageContent);
                string revokeExtendedData = sMessage.revoke_extended_data;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeExtendedData", revokeExtendedData);
            }
            else
            {
                cMessage = new ZIMMessage();
                baseType = cMessage.GetType();
            }
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "messageID", sMessage.message_id);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "localMessageID", sMessage.local_message_id);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "timestamp", sMessage.timestamp);
            string senderUserID = sMessage.sender_user_id;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "senderUserID", senderUserID);
            string conversationID = sMessage.conversation_id;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationID", conversationID);
            var conversationType = (ZIMConversationType)sMessage.conversation_type;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationType", conversationType);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationSeq", sMessage.conversation_seq);
            var direction = (ZIMMessageDirection)sMessage.direction;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "direction", direction);
            var sentStatus = (ZIMMessageSentStatus)sMessage.sent_status;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "sentStatus", sentStatus);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "orderKey", sMessage.order_key);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "userInserted", sMessage.is_user_inserted);
            var receiptStatus = (ZIMMessageReceiptStatus)sMessage.receipt_status;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "receiptStatus", receiptStatus);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "isBroadcastMessage", sMessage.is_broadcast_message);

            cMessage.extendedData = sMessage.extended_data;
            cMessage.localExtendedData = sMessage.local_extended_data;

            uint reactions_length = sMessage.reaction_length;
            zim_message_reaction[] reactions = new zim_message_reaction[reactions_length];
            ZIMUtil.GetStructListByPtr<zim_message_reaction>(ref reactions, sMessage.reactions, reactions_length);
            cMessage.reactions = ChangeZIMMessageReactionStructListToClassList(reactions);

            return cMessage;
        }

        public static void ChangeZIMMessageStructToClass(ref ZIMMessage cMessage, zim_message sMessage)
        {
            Type baseType;
            if (cMessage.type != (ZIMMessageType)sMessage.type) return;

            if (sMessage.type == (System.Int32)ZIMMessageType.Text)
            {
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Command)
            {
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Barrage)
            {
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Image)
            {
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var imageMessage = cMessage as ZIMImageMessage;
                imageMessage.largeImageDownloadUrl = sMessage.large_image_download_url;
                imageMessage.thumbnailDownloadUrl = sMessage.thumbnail_download_url;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, mediaType, "fileSize", sMessage.file_size);
                imageMessage.fileDownloadUrl = sMessage.file_download_url;
                imageMessage.fileLocalPath = sMessage.file_local_path;
                string largeImageDownloadUrl = sMessage.large_image_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageDownloadUrl", largeImageDownloadUrl);
                string largeImageLocalPath = sMessage.large_image_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageLocalPath", largeImageLocalPath);
                string thumbnailDownloadUrl = sMessage.thumbnail_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailDownloadUrl", thumbnailDownloadUrl);

                string thumbnailLocalPath = sMessage.thumbnail_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailLocalPath", thumbnailLocalPath);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "originalImageWidth", sMessage.original_image_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "originalImageHeight", sMessage.original_image_height);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageWidth", sMessage.large_image_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "largeImageHeight", sMessage.large_image_height);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailWidth", sMessage.thumbnail_width);
                ZIMUtil.ModifyFieldsValue<ZIMImageMessage>(imageMessage, "thumbnailHeight", sMessage.thumbnail_height);
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.File)
            {
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var fileMessage = cMessage as ZIMFileMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileUID", fileUID);

                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMFileMessage>(fileMessage, mediaType, "fileSize", sMessage.file_size);
                fileMessage.fileDownloadUrl = sMessage.file_download_url;
                fileMessage.fileLocalPath = sMessage.file_local_path;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Audio)
            {
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var audioMessage = cMessage as ZIMAudioMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMAudioMessage>(audioMessage, mediaType, "fileSize", sMessage.file_size);
                audioMessage.fileDownloadUrl = sMessage.file_download_url;
                audioMessage.fileLocalPath = sMessage.file_local_path;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Video)
            {
                baseType = cMessage.GetType().BaseType.BaseType;
                var mediaType = cMessage.GetType().BaseType;
                var videoMessage = cMessage as ZIMVideoMessage;
                string fileUID = sMessage.file_uid;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileUID", fileUID);
                string fileName = sMessage.file_name;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileName", fileName);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, mediaType, "fileSize", sMessage.file_size);

                videoMessage.fileDownloadUrl = sMessage.file_download_url;
                videoMessage.fileLocalPath = sMessage.file_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoDuration", sMessage.media_duration);
                string videoFirstFrameLocalPath = sMessage.video_first_frame_local_path;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameLocalPath", videoFirstFrameLocalPath);
                string videoFirstFrameDownloadUrl = sMessage.video_first_frame_download_url;
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameDownloadUrl", videoFirstFrameDownloadUrl);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameWidth", sMessage.video_first_frame_width);
                ZIMUtil.ModifyFieldsValue<ZIMVideoMessage>(videoMessage, "videoFirstFrameHeight", sMessage.video_first_frame_height);
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.System)
            {
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Custom)
            {
                baseType = cMessage.GetType().BaseType;
            }
            else if (sMessage.type == (System.Int32)ZIMMessageType.Revoke)
            {
                baseType = cMessage.GetType().BaseType;
                var revokeMessage = cMessage as ZIMRevokeMessage;
                var revokeType = (ZIMRevokeType)sMessage.revoke_type;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeType", revokeType);
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeTimestamp", sMessage.revoke_timestamp);
                string operatedUserID = sMessage.operated_userid;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "operatedUserID", operatedUserID);
                var originalMessageType = (ZIMMessageType)sMessage.original_message_type;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "originalMessageType", originalMessageType);
                string originalTextMessageContent = sMessage.original_text_message;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "originalTextMessageContent", originalTextMessageContent);
                string revokeExtendedData = sMessage.revoke_extended_data;
                ZIMUtil.ModifyFieldsValue<ZIMRevokeMessage>(revokeMessage, "revokeExtendedData", revokeExtendedData);
            }
            else
            {
                baseType = cMessage.GetType();
            }
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "messageID", sMessage.message_id);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "localMessageID", sMessage.local_message_id);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "timestamp", sMessage.timestamp);
            string senderUserID = sMessage.sender_user_id;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "senderUserID", senderUserID);
            string conversationID = sMessage.conversation_id;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationID", conversationID);
            var conversationType = (ZIMConversationType)sMessage.conversation_type;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationType", conversationType);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "conversationSeq", sMessage.conversation_seq);
            var direction = (ZIMMessageDirection)sMessage.direction;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "direction", direction);
            var sentStatus = (ZIMMessageSentStatus)sMessage.sent_status;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "sentStatus", sentStatus);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "orderKey", sMessage.order_key);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "userInserted", sMessage.is_user_inserted);
            var receiptStatus = (ZIMMessageReceiptStatus)sMessage.receipt_status;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "receiptStatus", receiptStatus);
            string extendedData = sMessage.extended_data;
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "extendedData", extendedData);
            ZIMUtil.ModifyFieldsValue<ZIMMessage>(cMessage, baseType, "isBroadcastMessage", sMessage.is_broadcast_message);

            cMessage.extendedData = sMessage.extended_data;
            cMessage.localExtendedData = sMessage.local_extended_data;

            uint reactions_length = sMessage.reaction_length;
            zim_message_reaction[] reactions = new zim_message_reaction[reactions_length];
            ZIMUtil.GetStructListByPtr<zim_message_reaction>(ref reactions, sMessage.reactions, reactions_length);
            cMessage.reactions = ChangeZIMMessageReactionStructListToClassList(reactions);
        }

        public static IntPtr AllocChangeZIMMessageListToPtrList(List<ZIMMessage> messageList)
        {
            var size = messageList.Count;

            zim_message[] messages_list = new zim_message[size];

            for (int i = 0; i < size; i++)
            {
                messages_list[i] = ChangeZIMMessageClassToStruct(messageList[i]);
            }

            IntPtr ptr = ZIMUtil.AllocGetIntPtrFromStructList<zim_message>(messages_list);

            return ptr;
        }
        public static zim_message ChangeZIMMessageClassToStruct(ZIMMessage zimMessage)
        {
            zim_message sMessage = new zim_message();
            if (zimMessage == null) return sMessage;
            if (zimMessage.type == ZIMMessageType.Text)
            {
                var textMessage = zimMessage as ZIMTextMessage;
                sMessage.message = textMessage.message;
                sMessage.type = (System.Int32)ZIMMessageType.Text;
            }
            else if (zimMessage.type == ZIMMessageType.Command)
            {
                var commandMessage = zimMessage as ZIMCommandMessage;

                sMessage.command_message = ZIMUtil.ArrayToPtr(commandMessage.message);
                sMessage.command_message_length = (uint)commandMessage.message.Length;
                sMessage.type = (System.Int32)ZIMMessageType.Command;
            }
            else if (zimMessage.type == ZIMMessageType.Barrage)
            {
                var barrageMessage = zimMessage as ZIMBarrageMessage;
                sMessage.message = barrageMessage.message;
                sMessage.type = (System.Int32)ZIMMessageType.Barrage;
            }
            else if (zimMessage.type == ZIMMessageType.Image)
            {
                var imageMessage = zimMessage as ZIMImageMessage;
                sMessage.file_local_path = imageMessage.fileLocalPath;
                sMessage.file_uid = imageMessage.fileUID;
                sMessage.file_name = imageMessage.fileName;
                sMessage.file_size = imageMessage.fileSize;
                sMessage.file_download_url = imageMessage.fileDownloadUrl;
                sMessage.large_image_download_url =
                    imageMessage.largeImageDownloadUrl;
                sMessage.large_image_local_path =
                    imageMessage.largeImageLocalPath;
                sMessage.thumbnail_download_url =
                    imageMessage.thumbnailDownloadUrl;
                sMessage.thumbnail_local_path =
                    imageMessage.thumbnailLocalPath;
                sMessage.type = (System.Int32)ZIMMessageType.Image;
                sMessage.original_image_width = imageMessage.originalImageWidth;
                sMessage.original_image_height = imageMessage.originalImageHeight;
                sMessage.large_image_width = imageMessage.largeImageWidth;
                sMessage.large_image_height = imageMessage.largeImageHeight;
                sMessage.thumbnail_width = imageMessage.thumbnailWidth;
                sMessage.thumbnail_height = imageMessage.thumbnailHeight;
            }
            else if (zimMessage.type == ZIMMessageType.File)
            {
                var fileMessage = zimMessage as ZIMFileMessage;
                sMessage.file_local_path = fileMessage.fileLocalPath;
                sMessage.file_uid = fileMessage.fileUID;
                sMessage.file_name = fileMessage.fileName;
                sMessage.file_size = fileMessage.fileSize;
                sMessage.file_download_url = fileMessage.fileDownloadUrl;
                sMessage.type = (System.Int32)ZIMMessageType.File;
            }
            else if (zimMessage.type == ZIMMessageType.Audio)
            {
                var audioMessage = zimMessage as ZIMAudioMessage;
                sMessage.file_local_path = audioMessage.fileLocalPath;
                sMessage.file_uid = audioMessage.fileUID;
                sMessage.file_name = audioMessage.fileName;
                sMessage.file_size = audioMessage.fileSize;
                sMessage.file_download_url = audioMessage.fileDownloadUrl;
                sMessage.media_duration = audioMessage.audioDuration;
                sMessage.type = (System.Int32)ZIMMessageType.Audio;
            }
            else if (zimMessage.type == ZIMMessageType.Video)
            {
                var videoMessage = zimMessage as ZIMVideoMessage;
                sMessage.file_local_path = videoMessage.fileLocalPath;
                sMessage.file_uid = videoMessage.fileUID;
                sMessage.file_name = videoMessage.fileName;
                sMessage.file_size = videoMessage.fileSize;
                sMessage.file_download_url = videoMessage.fileDownloadUrl;
                sMessage.media_duration = videoMessage.videoDuration;
                sMessage.video_first_frame_local_path =
                    videoMessage.videoFirstFrameLocalPath;
                sMessage.video_first_frame_download_url =
                    videoMessage.videoFirstFrameDownloadUrl;
                sMessage.video_first_frame_width = videoMessage.videoFirstFrameWidth;
                sMessage.video_first_frame_height = videoMessage.videoFirstFrameHeight;
                sMessage.type = (System.Int32)ZIMMessageType.Video;
            }
            else if (zimMessage.type == ZIMMessageType.System)
            {
                var systemMessage = zimMessage as ZIMSystemMessage;
                sMessage.message = systemMessage.message;
                sMessage.type = (System.Int32)ZIMMessageType.System;
            }
            else if (zimMessage.type == ZIMMessageType.Custom)
            {
                var customMessage = zimMessage as ZIMCustomMessage;
                sMessage.message = customMessage.message;
                sMessage.sub_type = customMessage.subType;
                sMessage.searched_content = customMessage.searchedContent;
                sMessage.type = (System.Int32)ZIMMessageType.Custom;
            }
            else if (zimMessage.type == ZIMMessageType.Revoke)
            {
                var revokeMessage = zimMessage as ZIMRevokeMessage;
                sMessage.revoke_type = (System.Int32)revokeMessage.revokeType;
                sMessage.revoke_timestamp = revokeMessage.revokeTimestamp;
                sMessage.operated_userid =
                    revokeMessage.operatedUserID;
                sMessage.original_message_type =
                    (System.Int32)revokeMessage.originalMessageType;
                sMessage.original_text_message =
                    revokeMessage.originalTextMessageContent;
                sMessage.revoke_extended_data =
                    revokeMessage.revokeExtendedData;
                sMessage.revoke_status = (System.Int32)revokeMessage.revokeStatus;
            }
            else
            {
                sMessage.type = (System.Int32)ZIMMessageType.Unknown;
            }
            sMessage.conversation_seq = zimMessage.conversationSeq;
            sMessage.message_id = zimMessage.messageID;
            sMessage.local_message_id = zimMessage.localMessageID;
            sMessage.direction = (System.Int32)zimMessage.direction;
            sMessage.order_key = zimMessage.orderKey;
            sMessage.conversation_type = (System.Int32)zimMessage.conversationType;
            sMessage.sender_user_id = zimMessage.senderUserID;
            sMessage.conversation_id = zimMessage.conversationID;
            sMessage.timestamp = zimMessage.timestamp;
            sMessage.sent_status = (System.Int32)zimMessage.sentStatus;
            sMessage.receipt_status = (System.Int32)zimMessage.receiptStatus;
            sMessage.is_user_inserted = zimMessage.isUserInserted;
            sMessage.extended_data = zimMessage.extendedData;
            sMessage.is_broadcast_message = zimMessage.isBroadcastMessage;
            sMessage.local_extended_data = zimMessage.localExtendedData;
            sMessage.reaction_length = (uint)zimMessage.reactions.Count;
            sMessage.reactions = AllocChangeZIMMessageReactionListToPtrList(zimMessage.reactions);
            return sMessage;
        }

        public static IntPtr AllocChangeZIMMessageClassToStructPtr(ZIMMessage zimMessage)
        {
            zim_message message = ChangeZIMMessageClassToStruct(zimMessage);
            return ZIMUtil.StructToIntPtr<zim_message>(message);
        }

        public static List<ZIMMessageReaction> ChangeZIMMessageReactionStructListToClassList(zim_message_reaction[] reactions)
        {
            int array_length = reactions.Length;
            List<ZIMMessageReaction> reactionList = new List<ZIMMessageReaction>();
            for (int i = 0; i < array_length; i++)
            {
                var reaction = new ZIMMessageReaction();
                reaction.conversationID = reactions[i].conversation_id;
                reaction.conversationType = (ZIMConversationType)reactions[i].conversation_type;
                reaction.isSelfIncluded = reactions[i].self_included;
                reaction.messageID = reactions[i].message_id;
                reaction.reactionType = reactions[i].reaction_type;
                reaction.totalCount = reactions[i].total_count;
                uint users_length = reactions[i].user_info_length;
                zim_message_reaction_user_info[] infos = new zim_message_reaction_user_info[users_length];
                ZIMUtil.GetStructListByPtr<zim_message_reaction_user_info>(ref infos, reactions[i].user_list, users_length);
                reaction.userList = ChangeZIMMessageReactionUserInfoStructListToClassList(infos);

                reactionList.Add(reaction);
            }

            return reactionList;
        }

        public static zim_message_reaction ChangeZIMMessageReactionClassToStruct(ZIMMessageReaction reaction)
        {
            zim_message_reaction message_reaction = new zim_message_reaction();
            message_reaction.conversation_id = reaction.conversationID;
            message_reaction.conversation_type = (int)reaction.conversationType;
            message_reaction.message_id = reaction.messageID;
            message_reaction.reaction_type = reaction.reactionType;
            message_reaction.self_included = reaction.isSelfIncluded;
            message_reaction.total_count = reaction.totalCount;
            message_reaction.user_info_length = (uint)reaction.userList.Count;
            message_reaction.user_list = AllocChangeZIMMessageReactionUserInfoListToPtrList(reaction.userList);
            return message_reaction;
        }

        public static IntPtr AllocChangeZIMMessageReactionListToPtrList(List<ZIMMessageReaction> reactionList)
        {
            var size = reactionList.Count;

            zim_message_reaction[] reaction_list = new zim_message_reaction[size];

            for (int i = 0; i < size; i++)
            {
                reaction_list[i] = ChangeZIMMessageReactionClassToStruct(reactionList[i]);
            }

            IntPtr ptr = ZIMUtil.AllocGetIntPtrFromStructList<zim_message_reaction>(reaction_list);

            return ptr;
        }

        public static IntPtr AllocChangeZIMMessageReactionUserInfoListToPtrList(List<ZIMMessageReactionUserInfo> userList)
        {
            var size = userList.Count;

            zim_message_reaction_user_info[] reaction_user_info = new zim_message_reaction_user_info[size];

            for (int i = 0; i < size; i++)
            {
                reaction_user_info[i].user_id = userList[i].userID;
            }

            IntPtr ptr = ZIMUtil.AllocGetIntPtrFromStructList<zim_message_reaction_user_info>(reaction_user_info);

            return ptr;
        }

        public static List<ZIMMessageReactionUserInfo> ChangeZIMMessageReactionUserInfoStructListToClassList(zim_message_reaction_user_info[] infos)
        {
            int array_length = infos.Length;
            List<ZIMMessageReactionUserInfo> userList = new List<ZIMMessageReactionUserInfo>();

            for (int i = 0; i < array_length; i++)
            {
                var user = new ZIMMessageReactionUserInfo();
                user.userID = infos[i].user_id;
                userList.Add(user);
            }

            return userList;
        }

        public static List<ZIMGroupSearchInfo> ChangeZIMGroupSearchInfoStructListToClassList(zim_group_search_info[] infos)
        {
            int array_length = infos.Length;
            List<ZIMGroupSearchInfo> groupSearchInfoList = new List<ZIMGroupSearchInfo>();

            for (int i = 0; i < array_length; i++)
            {
                ZIMGroupSearchInfo groupSearchInfo = new ZIMGroupSearchInfo();
                groupSearchInfo.groupInfo = ChangeZIMGroupBaseInfoStructToClass(infos[i].group_info);

                uint userListLength = infos[i].user_list_length;
                zim_group_member_info[] group_member_infos = new zim_group_member_info[userListLength];
                ZIMUtil.GetStructListByPtr<zim_group_member_info>(ref group_member_infos, infos[i].user_list, userListLength);
                groupSearchInfo.userList = ChangeZIMGroupMemberInfoStructListToClassList(group_member_infos);

                groupSearchInfoList.Add(groupSearchInfo);
            }

            return groupSearchInfoList;
        }

        public static zim_log_config ChangeZIMLogConfigClassToStruct(ZIMLogConfig config)
        {
            zim_log_config zimLogConfig = new zim_log_config();
            if (config == null)
            {
                return zimLogConfig;
            }
            zimLogConfig.log_path = config.logPath;
            zimLogConfig.log_size = config.logSize;
            return zimLogConfig;
        }

        public static zim_cache_config ChangeZIMCacheConfigClassToStruct(ZIMCacheConfig config)
        {
            zim_cache_config sConfig = new zim_cache_config();
            if (config == null) return sConfig;
            sConfig.cache_path = config.cachePath;
            return sConfig;
        }

        public static zim_room_member_query_config ChangeZIMRoomMemberQueryConfigClassToStruct(ZIMRoomMemberQueryConfig config)
        {
            zim_room_member_query_config sConfig = new zim_room_member_query_config();
            sConfig.next_flag = "";
            if (config == null) return sConfig;
            sConfig.next_flag = config.nextFlag == null ? "" : config.nextFlag;
            sConfig.count = config.count;
            return sConfig;
        }

        public static zim_call_invite_config ChangeZIMCallInviteConfigClassToStruct(ZIMCallInviteConfig config)
        {
            zim_call_invite_config sConfig = new zim_call_invite_config();
            if (config == null) return sConfig;
            sConfig.timeout = config.timeout;
            sConfig.mode = (int)config.mode;
            sConfig.extended_data = config.extendedData;
            bool isEnableOfflinePush = config.pushConfig != null ? true : false;
            sConfig.enable_offline_push = isEnableOfflinePush;
            if (isEnableOfflinePush)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.pushConfig);
            }
            return sConfig;
        }

        public static zim_calling_invite_config ChangeZIMCallingInviteConfigClassToStruct(ZIMCallingInviteConfig config)
        {
            zim_calling_invite_config sConfig = new zim_calling_invite_config();
            if (config == null) return sConfig;
            bool isEnableOfflinePush = config.pushConfig != null ? true : false;
            sConfig.enable_offline_push = isEnableOfflinePush;
            if (isEnableOfflinePush)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.pushConfig);
            }
            return sConfig;
        }

        public static zim_call_cancel_config ChangeZIMCallCancelConfigClassToStruct(ZIMCallCancelConfig config)
        {
            zim_call_cancel_config sConfig = new zim_call_cancel_config();
            if (config == null) return sConfig;
            sConfig.extended_data = config.extendedData;
            return sConfig;
        }

        public static zim_call_accept_config ChangeZIMCallAcceptConfigClassToStruct(ZIMCallAcceptConfig config)
        {
            zim_call_accept_config sConfig = new zim_call_accept_config();
            if (config == null) return sConfig;
            sConfig.extended_data = config.extendedData;
            return sConfig;
        }

        public static ZIMCallQuitSentInfo ChangeZIMCallQuitSentInfoStructToClass(zim_call_quit_sent_info call_quit_sent_info)
        {
            ZIMCallQuitSentInfo callQuitSentInfo = new ZIMCallQuitSentInfo();
            callQuitSentInfo.acceptTime = call_quit_sent_info.accept_time;
            callQuitSentInfo.createTime = call_quit_sent_info.create_time;
            callQuitSentInfo.quitTime = call_quit_sent_info.quit_time;

            return callQuitSentInfo;
        }

        public static ZIMCallEndedSentInfo ChangeZIMCallEndedSentInfoStructToClass(zim_call_end_sent_info call_end_sent_info)
        {
            ZIMCallEndedSentInfo callEndedSentInfo = new ZIMCallEndedSentInfo();
            callEndedSentInfo.acceptTime = call_end_sent_info.accept_time;
            callEndedSentInfo.createTime = call_end_sent_info.create_time;
            callEndedSentInfo.endTime = call_end_sent_info.end_time;

            return callEndedSentInfo;
        }

        public static List<ZIMCallInfo> ChangeZIMCallInfoStructListToClassList(zim_call_info[] call_infos)
        {
            List<ZIMCallInfo> callInfoList = new List<ZIMCallInfo>();
            foreach(zim_call_info info in call_infos)
            {
                ZIMCallInfo callInfo = new ZIMCallInfo();
                callInfo.caller = info.caller;
                callInfo.callID = info.call_id;
                callInfo.createTime = info.create_time;
                callInfo.endTime = info.end_time;
                callInfo.extendedData = info.extended_data;
                callInfo.mode = (ZIMCallInvitationMode)info.mode;
                callInfo.state = (ZIMCallState)info.state;
                zim_call_user_info[] call_user_infos = new zim_call_user_info[info.call_user_list_length];
                ZIMUtil.GetStructListByPtr<zim_call_user_info>(ref call_user_infos, info.call_user_list, info.call_user_list_length);
                callInfo.callUserList = ChangeZIMCallUserInfoStructListToClassList(call_user_infos);
                callInfoList.Add(callInfo);
            }

            return callInfoList;
        }

        public static List<ZIMCallUserInfo> ChangeZIMCallUserInfoStructListToClassList(zim_call_user_info[] call_user_infos)
        {

            List<ZIMCallUserInfo> callUserInfoList = new List<ZIMCallUserInfo>();
            foreach (zim_call_user_info info in call_user_infos)
            {
                ZIMCallUserInfo callUserInfo = new ZIMCallUserInfo();
                callUserInfo.userID = info.user_id;
                callUserInfo.userState = (ZIMCallUserState)info.user_state;
                callUserInfo.state = (ZIMCallUserState)info.user_state;
                callUserInfo.extendedData = info.extended_data;
                callUserInfoList.Add(callUserInfo);
            }

            return callUserInfoList;
        }

        public static zim_call_reject_config ChangeZIMCallRejectConfigClassToStruct(ZIMCallRejectConfig config)
        {
            zim_call_reject_config sConfig = new zim_call_reject_config();
            if (config == null) return sConfig;
            sConfig.extended_data = config.extendedData;
            return sConfig;
        }

        public static zim_call_quit_config ChangeZIMCallQuitConfigClassToStruct(ZIMCallQuitConfig config)
        {
            zim_call_quit_config sConfig = new zim_call_quit_config();
            if (config == null) return sConfig;
            sConfig.extended_data = config.extendedData;
            bool isEnableOfflinePush = config.pushConfig != null ? true : false;
            sConfig.enable_offline_push = isEnableOfflinePush;
            if (isEnableOfflinePush)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.pushConfig);
            }
            return sConfig;
        }

        public static zim_call_end_config ChangeZIMCallEndConfigClassToStruct(ZIMCallEndConfig config)
        {
            zim_call_end_config sConfig = new zim_call_end_config();
            if (config == null) return sConfig;
            sConfig.extended_data = config.extendedData;
            bool isEnableOfflinePush = config.pushConfig != null ? true : false;
            sConfig.enable_offline_push = isEnableOfflinePush;
            if (isEnableOfflinePush)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.pushConfig);
            }
            return sConfig;
        }

        public static zim_query_call_list_config ChangeZIMCallInvitationQueryConfigClassToStruct(ZIMCallInvitationQueryConfig config)
        {
            zim_query_call_list_config sConfig = new zim_query_call_list_config();
            if (config == null) return sConfig;
            sConfig.count = config.count;
            sConfig.next_flag = config.nextFlag;
            return sConfig;
        }

        public static zim_message_send_config ChangeZIMMessageSendConfigClassToStruct(ZIMMessageSendConfig config)
        {
            zim_message_send_config sConfig = new zim_message_send_config();
            if (config == null) return sConfig;
            bool isEnableOfflinePush = config.pushConfig != null ? true : false;
            sConfig.priority = (System.Int32)config.priority;
            sConfig.has_receipt = config.hasReceipt;
            sConfig.enable_offline_push = isEnableOfflinePush;

            if (isEnableOfflinePush)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.pushConfig);
            }
            return sConfig;
        }

        private static zim_push_config ChangeZIMPushConfigClassToStruct(ZIMPushConfig config)
        {
            zim_push_config push_config = new zim_push_config();
            if (config == null) return push_config;
            push_config.title = config.title;
            push_config.content = config.content;
            push_config.payload = config.payload;
            push_config.resources_id = config.resourcesID;
            return push_config;
        }

        public static zim_message_query_config ChangeZIMMessageQueryConfigClassToStruct(ZIMMessageQueryConfig config)
        {
            zim_message_query_config sConfig = new zim_message_query_config();
            if (config == null) return sConfig;
            sConfig.count = config.count;
            sConfig.reverse = config.reverse;
            if (config.nextMessage != null)
            {
                sConfig.next_message = AllocChangeZIMMessageClassToStructPtr(config.nextMessage);
            }
            return sConfig;
        }

        public static zim_message_delete_config ChangeZIMMessageDeleteConfigClassToStruct(ZIMMessageDeleteConfig config)
        {
            zim_message_delete_config sConfig = new zim_message_delete_config();
            sConfig.is_also_delete_server_message = config.isAlsoDeleteServerMessage;
            return sConfig;
        }

        public static zim_message_revoke_config ChangeZIMMessageRevokeConfigClassToStruct(ZIMMessageRevokeConfig config)
        {
            zim_message_revoke_config sConfig = new zim_message_revoke_config();
            if (config == null) return sConfig;

            sConfig.enable_offline_push = config.config != null ? true : false;
            sConfig.revoke_extended_data = config.revokeExtendedData;

            if (sConfig.enable_offline_push)
            {
                sConfig.push_config = ChangeZIMPushConfigClassToStruct(config.config);
            }
            return sConfig;
        }
        public static zim_user_info ChangeZIMUserInfoClassToStruct(ZIMUserInfo userInfo)
        {
            zim_user_info user_info = new zim_user_info();
            if (userInfo == null) return user_info;
            user_info.user_id = userInfo.userID;
            user_info.user_name = userInfo.userName;
            return user_info;
        }

        public static List<ZIMUserInfo> ChangeZIMUserInfoStructListToClassList(zim_user_info[] user_infos_list)
        {
            List<ZIMUserInfo> userInfosList = new List<ZIMUserInfo>();

            foreach(var info in user_infos_list)
            {
                var userInfo = new ZIMUserInfo();
                userInfo.userID = info.user_id;
                userInfo.userName = info.user_name;
                userInfosList.Add(userInfo);
            }
            return userInfosList;
        }

        public static List<ZIMRoomMemberInfo> ChangeZIMRoomMemberInfoStructListToClassList(zim_room_member_info[] user_infos_list)
        {
            List<ZIMRoomMemberInfo> userInfosList = new List<ZIMRoomMemberInfo>();

            foreach(var info in user_infos_list)
            {
                var userInfo = new ZIMRoomMemberInfo();
                userInfo.userID = info.user_info.user_id;
                userInfo.userName = info.user_info.user_name;
                userInfosList.Add(userInfo);
            }
            return userInfosList;
        }

        public static zim_users_info_query_config ChangeZIMUsersInfoQueryConfigToStruct(ZIMUsersInfoQueryConfig config)
        {
            zim_users_info_query_config users_info_query_config = new zim_users_info_query_config();
            if (config == null) return users_info_query_config;
            users_info_query_config.is_query_from_server = config.isQueryFromServer;
            return users_info_query_config;
        }

        public static ZIMUserFullInfo ChangeZIMUserFullInfoStructToClass(zim_user_full_info user_info)
        {
            var base_user_info = user_info.base_info;
            ZIMUserFullInfo userInfo = new ZIMUserFullInfo();
            userInfo.baseInfo.userID = base_user_info.user_id;
            userInfo.baseInfo.userName = base_user_info.user_name;
            userInfo.userAvatarUrl = user_info.user_avatar_url;
            userInfo.extendedData = user_info.user_extended_data;
            return userInfo;
        }

        public static List<ZIMMessageReceiptInfo> ChangeZIMMessageReceiptInfoStructListToClassList(zim_message_receipt_info[] read_receipt_list)
        {
            List<ZIMMessageReceiptInfo> messageReceiptInfoList = new List<ZIMMessageReceiptInfo>();
            foreach(var info in read_receipt_list)
            {
                ZIMMessageReceiptInfo messageReceiptInfo = new ZIMMessageReceiptInfo();
                messageReceiptInfo.conversationID = info.conversation_id;
                messageReceiptInfo.conversationType = (ZIMConversationType)info.conversation_type;
                messageReceiptInfo.status = (ZIMMessageReceiptStatus)info.status;
                messageReceiptInfo.messageID = info.message_id;
                messageReceiptInfo.readMemberCount = info.read_member_count;
                messageReceiptInfo.unreadMemberCount = info.unread_member_count;
                messageReceiptInfoList.Add(messageReceiptInfo);
            }
            return messageReceiptInfoList;
        }

        public static List<ZIMConversation> ChangeZIMConversationStructListToClassList(zim_conversation[] zim_conversations_list)
        {
            List<ZIMConversation> conversationList = new List<ZIMConversation>();
            foreach(var conv in zim_conversations_list)
            {
                ZIMConversation conversation = new ZIMConversation();
                conversation.conversationID = conv.conversation_id;
                conversation.conversationName = conv.conversation_name;
                conversation.conversationAvatarUrl = conv.conversation_avatar_url;
                conversation.type = (ZIMConversationType)conv.type;
                conversation.unreadMessageCount = conv.unread_message_count;
                conversation.orderKey = conv.order_key;
                conversation.lastMessage = ChangeZIMMessageStructToClass(conv.last_message);
                conversation.notificationStatus =
                    (ZIMConversationNotificationStatus)conv.notification_status;
                conversation.isPinned = conv.is_pinned;

                conversationList.Add(conversation);
            }

            return conversationList;
        }

        public static ZIMConversation ChangeZIMConversationStructToClass(zim_conversation zim_conversation)
        {
            ZIMConversation conversation = new ZIMConversation();

            conversation.conversationID = zim_conversation.conversation_id;
            conversation.conversationName = zim_conversation.conversation_name;
            conversation.conversationAvatarUrl = zim_conversation.conversation_avatar_url;
            conversation.type = (ZIMConversationType)zim_conversation.type;
            conversation.unreadMessageCount = zim_conversation.unread_message_count;
            conversation.orderKey = zim_conversation.order_key;
            conversation.lastMessage = ChangeZIMMessageStructToClass(zim_conversation.last_message);
            conversation.notificationStatus =
                (ZIMConversationNotificationStatus)zim_conversation.notification_status;
            conversation.isPinned = zim_conversation.is_pinned;

            return conversation;
        }

        public static ZIMMessageReaction ChangeZIMMessageReactionStructToClass(zim_message_reaction message_reaction)
        {
            ZIMMessageReaction reaction = new ZIMMessageReaction();

            reaction.conversationID = message_reaction.conversation_id;
            reaction.conversationType = (ZIMConversationType)message_reaction.conversation_type;
            reaction.isSelfIncluded = message_reaction.self_included;
            reaction.messageID = message_reaction.message_id;
            reaction.reactionType = message_reaction.reaction_type;
            reaction.totalCount = message_reaction.total_count;
            zim_message_reaction_user_info[] reaction_user_infos = new zim_message_reaction_user_info[message_reaction.user_info_length];
            ZIMUtil.GetStructListByPtr<zim_message_reaction_user_info>(ref reaction_user_infos, message_reaction.user_list, message_reaction.user_info_length);
            reaction.userList = ChangeZIMMessageReactionUserInfoStructListToClassList(reaction_user_infos);

            return reaction;
        }

        public static ZIMConversationSearchInfo ChangeZIMConversationSearchInfoStructToClass(zim_conversation_search_info conversation_search_info)
        {
            ZIMConversationSearchInfo conversationSearchInfo = new ZIMConversationSearchInfo();
            conversationSearchInfo.conversationID = conversation_search_info.conversation_id;
            conversationSearchInfo.conversationType = (ZIMConversationType)conversation_search_info.conversation_type;

            uint length = conversation_search_info.message_list_length;
            zim_message[] messages = new zim_message[length];
            ZIMUtil.GetStructListByPtr<zim_message>(ref messages, conversation_search_info.message_list, length);
            for (int i = 0; i < length; i++)
            {
                conversationSearchInfo.messageList[i] = ZIMConverter.ChangeZIMMessageStructToClass(messages[i]);
            }

            conversationSearchInfo.totalMessageCount = conversation_search_info.total_message_count;

            return conversationSearchInfo;
        }

        public static List<ZIMConversationChangeInfo> ChangeZIMConversationChangeInfoStructListToClassList(zim_conversation_change_info[] zim_conversation_change_infos_list)
        {
            List<ZIMConversationChangeInfo> conversationChangeInfoList = new List<ZIMConversationChangeInfo>();
            foreach(var info in zim_conversation_change_infos_list)
            {
                var conversationChangeInfo = new ZIMConversationChangeInfo();
                conversationChangeInfo.conversationEvent = (ZIMConversationEvent)info.conversation_event;
                conversationChangeInfo.conversation = ChangeZIMConversationStructToClass(info.conversation);
                conversationChangeInfoList.Add(conversationChangeInfo);
            }
            return conversationChangeInfoList;
        }

        public static List<ZIMMessageSentStatusChangeInfo> ChangeZIMMessageSentStatusChangeInfoStructListToClassList(zim_message_sent_status_change_info[] message_sent_status_change_infos_list)
        {
            List<ZIMMessageSentStatusChangeInfo> messageSentStatusChangeInfosList = new List<ZIMMessageSentStatusChangeInfo>();

            foreach(var info in message_sent_status_change_infos_list)
            {
                ZIMMessageSentStatusChangeInfo cInfo = new ZIMMessageSentStatusChangeInfo();
                cInfo.status = (ZIMMessageSentStatus)info.status;
                cInfo.message = ChangeZIMMessageStructToClass(info.message);
                messageSentStatusChangeInfosList.Add(cInfo);
            }

            return messageSentStatusChangeInfosList;
        }

        public static ZIMRoomInfo ChangeZIMRoomInfoStructToClass(zim_room_info sInfo)
        {
            ZIMRoomInfo info = new ZIMRoomInfo();
            info.roomID = sInfo.room_id;
            info.roomName = sInfo.room_name;
            return info;
        }
        public static zim_room_info ChangeZIMRoomInfoClassToStruct(ZIMRoomInfo roomInfo)
        {
            zim_room_info room_info = new zim_room_info();
            room_info.room_id = roomInfo.roomID;
            room_info.room_name = roomInfo.roomName;
            return room_info;
        }

        public static ZIMRoomFullInfo ChangeZIMRoomFullInfoStructToClass(zim_room_full_info room_full_info)
        {
            ZIMRoomFullInfo roomFullInfo = new ZIMRoomFullInfo();
            zim_room_info base_info = room_full_info.base_info;
            roomFullInfo.baseInfo.roomID = base_info.room_id;
            roomFullInfo.baseInfo.roomName = base_info.room_name;
            return roomFullInfo;
        }

        public static IntPtr AllocChangeZIMRoomAttributesDictionaryToPtr(Dictionary<string, string> attributes)
        {
            int roomAttributesCount = attributes.Count;
            zim_room_attribute[] room_attributes = new zim_room_attribute[roomAttributesCount];
            int count = 0;
            foreach (var item in attributes)
            {
                zim_room_attribute room_attribute = new zim_room_attribute();
                room_attribute.key = item.Key;
                room_attribute.value = item.Value;
                room_attributes[count] = room_attribute;
                count++;
            }
            IntPtr ptr = ZIMUtil.AllocGetIntPtrFromStructList<zim_room_attribute>(room_attributes);
            return ptr;
        }

        public static IntPtr AllocChangeZIMGroupAttributesDictionaryToPtr(Dictionary<string, string> attributes)
        {
            int groupAttributesCount = attributes.Count;
            zim_group_attribute[] group_attributes = new zim_group_attribute[groupAttributesCount];
            int count = 0;
            foreach (var item in attributes)
            {
                zim_group_attribute group_attribute = new zim_group_attribute();
                group_attribute.key = item.Key;
                group_attribute.value = item.Value;
                group_attributes[count] = group_attribute;
                count++;
            }
            IntPtr ptr = ZIMUtil.AllocGetIntPtrFromStructList<zim_group_attribute>(group_attributes);

            return ptr;
        }

        public static zim_room_member_attribute[] ChangeZIMRoomMemberAttributesDictionaryToStructList(Dictionary<string, string> attributes)
        {
            int roomMemberAttributesCount = attributes.Count;
            zim_room_member_attribute[] room_member_attributes = new zim_room_member_attribute[roomMemberAttributesCount];
            int count = 0;
            foreach (var item in attributes)
            {
                zim_room_member_attribute room_member_attribute = new zim_room_member_attribute();
                room_member_attribute.key = item.Key;
                room_member_attribute.value = item.Value;
                room_member_attributes[count] = room_member_attribute;
                count++;
            }

            return room_member_attributes;
        }

        public static zim_room_advanced_config ChangeZIMRoomAdvancedConfigClassToStruct(ZIMRoomAdvancedConfig roomAdvancedConfig)
        {
            zim_room_advanced_config room_advanced_config = new zim_room_advanced_config();
            var roomAttributesCount = roomAdvancedConfig.roomAttributes.Count;
            if (roomAttributesCount != 0)
            {
                var ptr = AllocChangeZIMRoomAttributesDictionaryToPtr(roomAdvancedConfig.roomAttributes);
                room_advanced_config.room_attributes = ptr;
            }
            room_advanced_config.room_attributes_length = (uint)roomAttributesCount;
            room_advanced_config.room_destroy_delay_time = roomAdvancedConfig.roomDestroyDelayTime;

            return room_advanced_config;
        }

        public static int[] ChangeZIMMessageTypeListToIntArray(List<ZIMMessageType> typeList)
        {
            int count = typeList.Count;
            int[] messageTypesArray = new int[count];
            for (int i = 0; i < count; i++)
            {
                messageTypesArray[i] = (int)typeList[i];
            }
            return messageTypesArray;
        }

        public static int[] ChangeUIntListToIntArray(List<uint> typeList)
        {
            int count = typeList.Count;
            int[] IntArray = new int[count];
            for (int i = 0; i < count; i++)
            {
                IntArray[i] = (int)typeList[i];
            }
            return IntArray;
        }

        public static zim_message_search_config AllocChangeZIMMessageSearchConfigClassToStruct(ZIMMessageSearchConfig config)
        {

            zim_message_search_config message_search_config = new zim_message_search_config();
            message_search_config.next_message = AllocChangeZIMMessageClassToStructPtr(config.nextMessage);
            message_search_config.count = config.count;
            message_search_config.order = (int)config.order;
            var (keywordsIntptr, keywordHandle) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            message_search_config.keywords = keywordsIntptr;
            message_search_config.keywords_length = (uint)config.keywords.Count();
            message_search_config.end_time = config.endTime;
            int[] messageTypes = ChangeZIMMessageTypeListToIntArray(config.messageTypes);
            message_search_config.message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(messageTypes);
            message_search_config.message_types_length = (uint)config.messageTypes.Count();
            var (senderUserIDsIntptr, senderUserIDsHandle) = ZIMUtil.AllocGetIntPtrFromStringList(config.senderUserIDs);
            message_search_config.sender_user_ids = senderUserIDsIntptr;
            message_search_config.sender_user_ids_length = (uint)config.senderUserIDs.Count();
            message_search_config.start_time = config.startTime;
            int[] subMessageTypes = ChangeUIntListToIntArray(config.subMessageTypes);
            message_search_config.sub_message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(subMessageTypes);
            message_search_config.sub_message_types_length = (uint)config.subMessageTypes.Count();

            return message_search_config;
        }

        public static zim_group_search_config AllocChangeZIMGroupSearchConfigClassToStruct(ZIMGroupSearchConfig config)
        {
            zim_group_search_config group_search_config = new zim_group_search_config();
            var (keywordsIntptr, keywordHandle) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            group_search_config.keywords = keywordsIntptr;
            group_search_config.keywords_length = (uint)config.keywords.Count();
            group_search_config.is_also_match_group_member_nick_name = config.isAlsoMatchGroupMemberNickname;
            group_search_config.is_also_match_group_member_user_name = config.isAlsoMatchGroupMemberUserName;
            group_search_config.next_flag = config.nextFlag;
            group_search_config.count = config.count;
            return group_search_config;
        }

        public static zim_group_member_search_config AllocChangeZIMGroupSearchConfigClassToStruct(ZIMGroupMemberSearchConfig config)
        {
            zim_group_member_search_config group_member_search_config = new zim_group_member_search_config();
            var (keywordsIntptr, keywordHandle) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            group_member_search_config.keywords = keywordsIntptr;
            group_member_search_config.keywords_length = (uint)config.keywords.Count();
            group_member_search_config.is_also_match_group_member_nick_name = config.isAlsoMatchGroupMemberNickname;
            group_member_search_config.next_flag = config.nextFlag;
            group_member_search_config.count = config.count;
            return group_member_search_config;
        }
        public static zim_message_reaction_users_query_config ChangeZIMMessageReactionUserQueryConfigClassToStruct(ZIMMessageReactionUserQueryConfig messageReactionUserQueryConfig)
        {
            zim_message_reaction_users_query_config message_reaction_users_query_config = new zim_message_reaction_users_query_config
            {
                count = messageReactionUserQueryConfig.count,
                next_flag = messageReactionUserQueryConfig.nextFlag,
                reaction_type = messageReactionUserQueryConfig.reactionType
            };

            return message_reaction_users_query_config;
        }

        public static Dictionary<string, string> ChangeZIMRoomAttributesPtrToDictionary(IntPtr room_attributes_ptr, uint room_attributes_length)
        {
            Dictionary<string, string> roomAttributes = new Dictionary<string, string>();
            zim_room_attribute[] room_attributes = new zim_room_attribute[room_attributes_length];
            ZIMUtil.GetStructListByPtr<zim_room_attribute>(ref room_attributes, room_attributes_ptr, room_attributes_length);
            for (int i = 0; i < room_attributes_length; i++)
            {
                var key = room_attributes[i].key;
                var value = room_attributes[i].value;
                roomAttributes.Add(key, value);
            }
            return roomAttributes;
        }

        public static Dictionary<string, string> ChangeZIMGroupAttributesStructListPtrToDictionary(System.IntPtr group_attributes_list_ptr, uint size)
        {

            zim_group_attribute[] group_attributes_list = new zim_group_attribute[size];
            ZIMUtil.GetStructListByPtr<zim_group_attribute>(ref group_attributes_list, group_attributes_list_ptr, size);
            Dictionary<string, string> groupAttributes = new Dictionary<string, string>();
            for (int i = 0; i < size; i++)
            {
                var key = group_attributes_list[i].key;
                var value = group_attributes_list[i].value;
                groupAttributes.Add(key, value);
            }
            return groupAttributes;
        }

        public static List<ZIMRoomMemberAttributesOperatedInfo> ChangeZIMRoomMemberAttributesOperatedInfoStructListToClassList(zim_room_member_attributes_operated_info[] room_member_attributes_operated_infos_list)
        {
            List<ZIMRoomMemberAttributesOperatedInfo> roomMemberAttributesOperatedInfoList = new List<ZIMRoomMemberAttributesOperatedInfo>();
            foreach(var info in room_member_attributes_operated_infos_list)
            {
                var error_keys_length = info.error_keys_length;
                var errorKeyList = new string[error_keys_length];

                IntPtr[] ptrArray = new IntPtr[error_keys_length];
                Marshal.Copy(info.error_keys, ptrArray, 0, (int)error_keys_length);
                for (int j = 0; j < error_keys_length; j++)
                {
                    errorKeyList[j] = ZIMUtil.PtrToStringUTF8(ptrArray[j]);
                }
                var roomMemberAttributesOperatedInfo = new ZIMRoomMemberAttributesOperatedInfo();
                roomMemberAttributesOperatedInfo.errorKeys = new List<string>(errorKeyList);
                var room_member_attributes_info = info.attributes_info;
                roomMemberAttributesOperatedInfo.attributesInfo.userID = room_member_attributes_info.user_id;
                roomMemberAttributesOperatedInfo.attributesInfo.attributes = ChangeZIMRoomMemberAttributesPtrToDictionary(room_member_attributes_info
                    .attributes, room_member_attributes_info.attributes_length);

                roomMemberAttributesOperatedInfoList.Add(roomMemberAttributesOperatedInfo);
            }

            return roomMemberAttributesOperatedInfoList;
        }

        public static Dictionary<string, string> ChangeZIMRoomMemberAttributesPtrToDictionary(IntPtr room_member_attributes_ptr, uint room_member_attributes_length)
        {
            Dictionary<string, string> roomMemberAttributes = new Dictionary<string, string>();
            zim_room_member_attribute[] room_member_attributes = new zim_room_member_attribute[room_member_attributes_length];
            ZIMUtil.GetStructListByPtr<zim_room_member_attribute>(ref room_member_attributes, room_member_attributes_ptr, room_member_attributes_length);
            for (int i = 0; i < room_member_attributes_length; i++)
            {
                var key = room_member_attributes[i].key;
                var value = room_member_attributes[i].value;
                roomMemberAttributes.Add(key, value);
            }
            return roomMemberAttributes;
        }

        public static ZIMGroupInfo ChangeZIMGroupBaseInfoStructToClass(zim_group_info group_info)
        {
            ZIMGroupInfo info = new ZIMGroupInfo();
            info.groupID = group_info.group_id;
            info.groupName = group_info.group_name;
            info.groupAvatarUrl = group_info.group_avatar_url;
            return info;
        }

        public static zim_group_info ChangeZIMGroupBaseInfoClassToStruct(ZIMGroupInfo groupInfo)
        {
            zim_group_info group_info = new zim_group_info();
            group_info.group_id = groupInfo.groupID;
            group_info.group_name = groupInfo.groupName;
            group_info.group_avatar_url = groupInfo.groupAvatarUrl;

            return group_info;
        }

        public static ZIMGroupAdvancedConfig ChangeZIMGroupAdvancedConfigStructToClass(zim_group_advanced_config group_advanced_config)
        {
            ZIMGroupAdvancedConfig groupAdvancedConfig = new ZIMGroupAdvancedConfig();
            groupAdvancedConfig.groupNotice = group_advanced_config.group_notice;

            var groupAttributesLength = group_advanced_config.group_attributes_length;
            zim_group_attribute[] groupAttributes = new zim_group_attribute[groupAttributesLength];

            ZIMUtil.GetStructListByPtr<zim_group_attribute>(ref groupAttributes, group_advanced_config.group_attributes, groupAttributesLength);

            for (uint i = 0; i < groupAttributesLength; i++)
            {
                string key = groupAttributes[i].key;
                string value = groupAttributes[i].value;
                groupAdvancedConfig.groupAttributes.Add(key, value);
            }
            return groupAdvancedConfig;
        }

        public static zim_group_advanced_config ChangeZIMGroupAdvancedConfigClassToStruct(ZIMGroupAdvancedConfig groupAdvancedConfig)
        {
            zim_group_advanced_config group_advanced_config = new zim_group_advanced_config();
            group_advanced_config.group_notice = groupAdvancedConfig.groupNotice;
            group_advanced_config.group_attributes_length = (uint)groupAdvancedConfig.groupAttributes.Count;

            var groupAttributesListPtr = AllocChangeZIMGroupAttributesDictionaryToPtr(groupAdvancedConfig.groupAttributes);

            group_advanced_config.group_attributes = groupAttributesListPtr;
            return group_advanced_config;
        }

        public static List<ZIMGroup> ChangeZIMGroupStructListToClassList(zim_group[] groups_list)
        {
            List<ZIMGroup> groupList = new List<ZIMGroup>();
            foreach(var group in groups_list)
            {
                var cGroup = new ZIMGroup();
                var base_info = group.base_info;
                cGroup.baseInfo.groupID = base_info.group_id;
                cGroup.baseInfo.groupName = base_info.group_name;
                cGroup.baseInfo.groupAvatarUrl = base_info.group_avatar_url;
                cGroup.notificationStatus = (ZIMGroupMessageNotificationStatus)group.notification_status;
                groupList.Add(cGroup);
            }
            return groupList;
        }

        public static ZIMGroupFullInfo ChangeZIMGroupFullInfoStructToClass(zim_group_full_info group_full_info)
        {
            ZIMGroupFullInfo groupFullInfo = new ZIMGroupFullInfo();
            groupFullInfo.baseInfo = ChangeZIMGroupBaseInfoStructToClass(group_full_info.base_info);
            groupFullInfo.groupNotice = group_full_info.group_notice;
            groupFullInfo.groupAttributes = ChangeZIMGroupAttributesStructListPtrToDictionary(group_full_info.group_attributes, group_full_info.group_attributes_length);
            groupFullInfo.notificationStatus = (ZIMGroupMessageNotificationStatus)group_full_info.notification_status;
            return groupFullInfo;
        }

        public static List<ZIMGroupMemberInfo> ChangeZIMGroupMemberInfoStructListToClassList(zim_group_member_info[] group_member_infos_list)
        {
            List<ZIMGroupMemberInfo> groupMemberInfosList = new List<ZIMGroupMemberInfo>();
            foreach(var info in group_member_infos_list)
            {
                var groupMemberInfo = new ZIMGroupMemberInfo();
                groupMemberInfo.userID = info.user_id;
                groupMemberInfo.userName = info.user_name;
                groupMemberInfo.memberAvatarUrl = info.member_avatar_url;
                groupMemberInfo.memberNickname = info.member_nick_name;
                groupMemberInfo.memberRole = info.member_role;
                groupMemberInfosList.Add(groupMemberInfo);
            }
            return groupMemberInfosList;
        }

        public static ZIMGroupMemberInfo ChangeZIMGroupMemberInfoStructToClass(zim_group_member_info group_member_info)
        {
            ZIMGroupMemberInfo groupMemberInfo = new ZIMGroupMemberInfo();

            groupMemberInfo.userID = group_member_info.user_id;
            groupMemberInfo.userName = group_member_info.user_name;
            groupMemberInfo.memberAvatarUrl = group_member_info.member_avatar_url;
            groupMemberInfo.memberNickname = group_member_info.member_nick_name;
            groupMemberInfo.memberRole = group_member_info.member_role;

            return groupMemberInfo;
        }

        public static List<ZIMErrorUserInfo> ChangeZIMErrorUserInfoStructListToClassList(zim_error_user_info[] zim_error_user_infos_list)
        {
            List<ZIMErrorUserInfo> errorUserInfoList = new List<ZIMErrorUserInfo>();
            foreach(var info in zim_error_user_infos_list)
            {
                var errorUserInfo = new ZIMErrorUserInfo();
                errorUserInfo.userID = info.user_id;
                errorUserInfo.reason = info.reason;
                errorUserInfoList.Add(errorUserInfo);
            }
            return errorUserInfoList;
        }

        public static ZIMGroupOperatedInfo ChangeZIMGroupOperatedInfoStructToClass(zim_group_operated_info sInfo)
        {
            ZIMGroupOperatedInfo info = new ZIMGroupOperatedInfo();

            info.userID = sInfo.user_id;
            info.userName = sInfo.user_name;
            info.memberNickname = sInfo.member_nick_name;
            info.memberRole = sInfo.member_role;
            return info;
        }

        public static ZIMCallInvitationSentInfo ZIMCallInvitationSentInfoStructToClass(zim_call_invitation_sent_info sInfo)
        {
            ZIMCallInvitationSentInfo info = new ZIMCallInvitationSentInfo();
            info.timeout = sInfo.timeout;
            info.errorInvitees = new List<ZIMCallUserInfo>();
            info.errorUserList = new List<ZIMCallUserInfo>();
            zim_call_user_info[] call_user_infos = new zim_call_user_info[sInfo.old_error_invitees_length];
            ZIMUtil.GetStructListByPtr<zim_call_user_info>(ref call_user_infos, sInfo.old_error_invitees, sInfo.old_error_invitees_length);
            for (uint i = 0; i < sInfo.old_error_invitees_length; i++)
            {
                ZIMCallUserInfo userInfo = new ZIMCallUserInfo();
                userInfo.userID = call_user_infos[i].user_id;
                userInfo.userState = (ZIMCallUserState)call_user_infos[i].user_state;
                userInfo.state = (ZIMCallUserState)call_user_infos[i].user_state;
                userInfo.extendedData = call_user_infos[i].extended_data;
                info.errorInvitees.Add(userInfo);
                info.errorUserList.Add(userInfo);
            }

            return info;
        }

        public static ZIMCallInvitationReceivedInfo ChangeZIMCallInvitationReceivedInfoStructToClass(zim_call_invitation_received_info sInfo)
        {
            ZIMCallInvitationReceivedInfo info = new ZIMCallInvitationReceivedInfo();
            info.timeout = sInfo.timeout;
            info.inviter = sInfo.inviter;
            info.extendedData = sInfo.extended_data;
            return info;
        }

        public static ZIMCallInvitationCancelledInfo ChangeZIMCallInvitationCancelledInfoStructToClass(zim_call_invitation_cancelled_info sInfo)
        {
            ZIMCallInvitationCancelledInfo info = new ZIMCallInvitationCancelledInfo();
            info.inviter = sInfo.inviter;
            info.extendedData = sInfo.extended_data;
            return info;
        }

        public static ZIMCallInvitationEndedInfo ChangeZIMCallInvitationEndedInfoStructToClass(zim_call_invitation_ended_info sInfo)
        {
            ZIMCallInvitationEndedInfo info = new ZIMCallInvitationEndedInfo();
            info.caller = sInfo.caller;
            info.endTime = sInfo.end_time;
            info.extendedData = sInfo.extended_data;
            info.mode = (ZIMCallInvitationMode)sInfo.mode;
            info.operatedUserID = sInfo.operated_user_id;

            return info;
        }

        public static ZIMCallUserStateChangeInfo ChangeZIMCallUserStateChangeInfoStructToClass(zim_event_call_user_state_changed_info sInfo)
        {
            ZIMCallUserStateChangeInfo info = new ZIMCallUserStateChangeInfo();
            zim_call_user_info[] call_user_infos = new zim_call_user_info[sInfo.call_user_list_length];
            ZIMUtil.GetStructListByPtr<zim_call_user_info>(ref call_user_infos, sInfo.call_user_list, sInfo.call_user_list_length);

            info.callUserList = ChangeZIMCallUserInfoStructListToClassList(call_user_infos);

            return info;
        }

        public static ZIMCallInvitationAcceptedInfo ChangeZIMCallInvitationAcceptedInfoStructToClass(zim_call_invitation_accepted_info sInfo)
        {
            ZIMCallInvitationAcceptedInfo info = new ZIMCallInvitationAcceptedInfo();
            info.invitee = sInfo.invitee;
            info.extendedData = sInfo.extended_data;
            return info;
        }

        public static ZIMCallInvitationRejectedInfo ChangeZIMCallInvitationRejectedInfoStructToClass(zim_call_invitation_rejected_info sInfo)
        {
            ZIMCallInvitationRejectedInfo info = new ZIMCallInvitationRejectedInfo();
            info.invitee = sInfo.invitee;
            info.extendedData = sInfo.extended_data;
            return info;
        }

        public static ZIMCallInvitationTimeoutInfo ChangeZIMCallInvitationTimeoutInfoStructToClass(zim_call_invitation_timeout_info sInfo)
        {
            ZIMCallInvitationTimeoutInfo info = new ZIMCallInvitationTimeoutInfo();
            info.mode = (ZIMCallInvitationMode)sInfo.mode;
            return info;
        }

        public static ZIMError ChangeZIMErrorStructToClass(zim_error sError)
        {
            ZIMError error = new ZIMError();
            error.code = (ZIMErrorCode)sError.code;
            error.message = sError.message;
            return error;
        }

    }
#pragma warning restore CS0618
}
