using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Threading;
using System.Collections.Concurrent;
using System.Reflection;
using UnityEngine;
using static ZEGO.IZIMInternal;
using static ZEGO.ZIMConverter;
using System.Linq;

namespace ZEGO
{
    public class ZIMImpl : ZIM
    {
        private static readonly object zimHandleMapLock = new object();
        private static Dictionary<System.IntPtr, ZIMImpl> zimHandleMap = new Dictionary<System.IntPtr, ZIMImpl>();

        public static GameObject zimGameObj = null;
        public static ZIMUnloadObject zimUploadObject = null;

        public class ZIMUnloadObject : MonoBehaviour
        {

            //void OnApplicationQuit()
            //{
            //    if (ZIM.GetInstance() != null)
            //        ZIM.GetInstance().Destroy();
            //}

            void OnDestroy()
            { 
                if (ZIM.GetInstance() != null)
                    ZIM.GetInstance().Destroy();
            }

        }

        private static void InitUnloadObject()
        {
            zimGameObj = new GameObject("ZIMUnloadObject");
            zimGameObj.AddComponent<ZIMUnloadObject>();
            GameObject.DontDestroyOnLoad(zimGameObj);
            zimGameObj.hideFlags = HideFlags.HideInHierarchy;
            zimUploadObject = zimGameObj.GetComponent<ZIMUnloadObject>();
        }

        public static void UnInitUnloadObject()
        {
            GameObject obj = GameObject.Find("ZIMUnloadObject");
            if (obj is object)
            {
                GameObject.Destroy(zimGameObj);
                zimGameObj = null;
                zimUploadObject = null;
            }
        }

        public static new ZIM GetInstance()
        {
            System.IntPtr handle = System.IntPtr.Zero;
            zim_get_instance(ref handle);
            if (handle == System.IntPtr.Zero)
            {
                return null;
            }
            else
            {
                if (zimHandleMap.ContainsKey(handle))
                {
                    var instance = zimHandleMap[handle];
                    return instance;
                }
                else
                {
                    return null;
                }
            }
        }

        public static ZIMImpl GetZIMImplInstance(IntPtr handle)
        {
            if (zimHandleMap.ContainsKey(handle))
            {
                return zimHandleMap[handle];
            }
            else
            {
                return null;
            }
        }

        public static new string GetVersion()
        {
            return ZIMUtil.PtrToString(zim_get_version());
        }

        public static new ZIM Create(ZIMAppConfig config)
        {
            lock (zimHandleMapLock)
            {
                if (zimHandleMap.Count != 0)
                {
                    return null;
                }
            }

#if UNITY_ANDROID && !UNITY_EDITOR
            jvm = GetJVM();
            AndroidJavaObject currentActivity = new AndroidJavaClass("com.unity3d.player.UnityPlayer").GetStatic<AndroidJavaObject>("currentActivity");
            context = currentActivity.Call<AndroidJavaObject>("getApplicationContext").GetRawObject();
            zim_set_android_env(jvm, context);
#endif

            System.IntPtr handle = new IntPtr(0);
            zim_create(ref handle, config.appID, config.appSign);
            if (handle == System.IntPtr.Zero)
            {
                return null;
            }
            InitUnloadObject();
            var instance = new ZIMImpl
            {
                handle = handle
            };

            lock (zimHandleMapLock)
            {
                zimHandleMap[handle] = instance;
            }

            instance.ZIMRegisterCallback();

            return instance;
        }

        public static new void SetLogConfig(ZIMLogConfig config)
        {
            zim_log_config logConfig =
            ZIMConverter.ChangeZIMLogConfigClassToStruct(config);
            zim_set_log_config(logConfig);
        }

        public static new void SetAdvancedConfig(string key, string value)
        {
            zim_set_advanced_config(key, value);
        }

        public static new void SetCacheConfig(ZIMCacheConfig config)
        {
            zim_cache_config cacheConfig =
                ZIMConverter.ChangeZIMCacheConfigClassToStruct(config);
            zim_set_cache_config(cacheConfig);
        }

        public override void Destroy()
        {
            lock (zimHandleMapLock)
            {
                zimHandleMap.Remove(handle);
                IZIMInternal.zim_destroy(ref handle);
                UnInitUnloadObject();
            }
        }

        public override void UploadLog(ZIMLogUploadedCallback callback)
        {
            int sequence = 0;
            zim_upload_log(handle, ref sequence);
            lock (logUploadedCallbackLock)
            {
                logUploadedCallback[sequence] = callback;
            }
        }

        public override void Login(ZIMUserInfo userInfo, ZIMLoggedInCallback callback)
        {
            Login(userInfo, "", callback);
        }

        public override void Login(ZIMUserInfo userInfo, string token, ZIMLoggedInCallback callback)
        {
            var user_Info = ChangeZIMUserInfoClassToStruct(userInfo);
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (loggedInCallbackLock)
            {
                loggedInCallback[sequence] = callback;
            }
            zim_login(this.handle, user_Info, token, ref sequence);
        }

        public override void Logout()
        {
            zim_logout(handle);
        }

        public override void RenewToken(string token, ZIMTokenRenewedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (tokenRenewedCallbackLock)
            {
                tokenRenewedCallback[sequence] = callback;
            }
            zim_renew_token(handle, token, ref sequence);
        }

        public override void UpdateUserName(string userName, ZIMUserNameUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (userNameUpdatedCallbackLock)
            {
                userNameUpdatedCallback[sequence] = callback;
            }
            zim_update_user_name(handle, userName, ref sequence);
        }

        public override void UpdateUserAvatarUrl(string userAvatarUrl, ZIMUserAvatarUrlUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (userAvatarUrlUpdatedCallbackLock)
            {
                userAvatarUrlUpdatedCallback[sequence] = callback;
            }
            zim_update_user_avatar_url(handle, userAvatarUrl, ref sequence);
        }

        public override void UpdateUserExtendedData(string extendedData, ZIMUserExtendedDataUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (userExtendedDataUpdatedCallbackLock)
            {
                userExtendedDataUpdatedCallback[sequence] = callback;
            }
            zim_update_user_extended_data(handle, extendedData, ref sequence);
        }

        public override void QueryUsersInfo(List<string> userIDs, ZIMUsersInfoQueryConfig config, ZIMUsersInfoQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            zim_users_info_query_config user_info_config = new zim_users_info_query_config();

            if (config != null)
            {
                user_info_config.is_query_from_server = config.isQueryFromServer;
            }

            lock (usersInfoQueriedCallbackLock)
            {
                usersInfoQueriedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            IntPtr[] userIDsPtrList = new IntPtr[size];
            for (int j = 0; j < size; j++)
            {
                userIDsPtrList[j] = Marshal.StringToHGlobalAnsi(userIDArray[j]);
            }

            var users_config = ChangeZIMUsersInfoQueryConfigToStruct(config);
            zim_query_users_info(handle, userIDArray, (uint)size, users_config, ref sequence);
        }

        public override void SendMessage(ZIMMessage message, string toConversationID, ZIMConversationType conversationType, ZIMMessageSendConfig config, ZIMMessageSendNotification notification, ZIMMessageSentCallback callback)
        {
            var sMessage = ChangeZIMMessageClassToStruct(message);
            int sequence = Interlocked.Increment(ref this.sequence);
            var send_config = ChangeZIMMessageSendConfigClassToStruct(config);
            lock (messageSendNotificationCallbackLock)
            {
                messageSendNotificationCallback[sequence] = notification;
            }
            lock (messageSentCallbackLock)
            {
                messageSentCallback[sequence] = callback;
            }
            lock (messageObjMapLock)
            {
                messageObjMap[sequence] = message;
            }

            zim_send_message(handle, sMessage, toConversationID,
            (System.Int32)conversationType, send_config, ref sequence);

        }

        public override void SendMediaMessage(ZIMMediaMessage message, string toConversationID, ZIMConversationType conversationType, ZIMMessageSendConfig config, ZIMMediaMessageSendNotification notification, ZIMMessageSentCallback callback)
        {
            var sMessage = ChangeZIMMessageClassToStruct(message);
            int sequence = Interlocked.Increment(ref this.sequence);
            var send_config = ChangeZIMMessageSendConfigClassToStruct(config);
            lock (mediaMessageSendNotificationCallbackLock)
            {
                mediaMessageSendNotificationCallback[sequence] = notification;
            }

            lock (messageSentCallbackLock)
            {
                messageSentCallback[sequence] = callback;
            }

            lock (messageObjMapLock)
            {
                messageObjMap[sequence] = message;
            }
            zim_send_media_message(handle, sMessage, toConversationID,
            (System.Int32)conversationType, send_config, ref sequence);
        }

        public override void DownloadMediaFile(ZIMMediaMessage message, ZIMMediaFileType fileType, ZIMMediaDownloadingProgress progress, ZIMMediaDownloadedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var sMessage = ChangeZIMMessageClassToStruct(message);
            lock (mediaDownloadingProgressCallbackLock)
            {
                mediaDownloadingProgressCallback[sequence] = progress;
            }

            lock (mediaDownloadedCallbackLock)
            {
                mediaDownloadedCallback[sequence] = callback;
            }

            zim_download_media_file(handle, sMessage, (System.Int32)fileType, ref sequence);
        }

        public override void QueryHistoryMessage(string conversationID, ZIMConversationType conversationType, ZIMMessageQueryConfig config, ZIMMessageQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var query_config = ChangeZIMMessageQueryConfigClassToStruct(config);

            lock (messageQueriedCallbackLock)
            {
                messageQueriedCallback[sequence] = callback;
            }

            zim_query_history_message(handle, conversationID, (System.Int32)conversationType,
                query_config, ref sequence);
            ZIMUtil.ReleaseStructPointer(query_config.next_message);
        }

        public override void DeleteAllMessage(string conversationID, ZIMConversationType conversationType, ZIMMessageDeleteConfig config, ZIMMessageDeletedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var delete_config = ChangeZIMMessageDeleteConfigClassToStruct(config);

            lock (messageDeletedCallbackLock)
            {
                messageDeletedCallback[sequence] = callback;
            }

            zim_delete_all_message(handle, conversationID, (System.Int32)conversationType,
                delete_config, ref sequence);
        }

        public override void DeleteMessages(List<ZIMMessage> messageList, string conversationID, ZIMConversationType conversationType, ZIMMessageDeleteConfig config, ZIMMessageDeletedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var delete_config = ChangeZIMMessageDeleteConfigClassToStruct(config);

            var size = messageList.Count;

            var messageListPtr = AllocChangeZIMMessageListToPtrList(messageList);

            lock (messageDeletedCallbackLock)
            {
                messageDeletedCallback[sequence] = callback;
            }

            zim_delete_messages(handle, messageListPtr, (uint)size,
                conversationID, (System.Int32)conversationType, delete_config, ref sequence);

            ZIMUtil.ReleaseStructPointer(messageListPtr);
        }

        public override void RevokeMessage(ZIMMessage message, ZIMMessageRevokeConfig config, ZIMMessageRevokedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var revoke_config = ChangeZIMMessageRevokeConfigClassToStruct(config);
            var sMessage = ChangeZIMMessageClassToStruct(message);

            lock (messageRevokedCallbackLock)
            {
                messageRevokedCallback[sequence] = callback;
            }

            zim_revoke_message(handle, sMessage, revoke_config, ref sequence);
        }

        public override void InsertMessageToLocalDB(ZIMMessage message, string conversationID, ZIMConversationType conversationType, string senderUserID, ZIMMessageInsertedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var sMessage = ChangeZIMMessageClassToStruct(message);

            lock (messageInsertedCallbackLock)
            {
                messageInsertedCallback[sequence] = callback;
            }
            lock (messageObjMapLock)
            {
                messageObjMap[sequence] = message;
            }

            zim_insert_message_to_local_db(handle, sMessage, conversationID,
                (System.Int32)conversationType, senderUserID, ref sequence);
        }

        public override void SendMessageReceiptsRead(List<ZIMMessage> messageList, string conversationID, ZIMConversationType conversationType, ZIMMessageReceiptsReadSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var size = messageList.Count;

            var messageListPtr = AllocChangeZIMMessageListToPtrList(messageList);

            lock (messageReceiptsReadSentCallbackLock)
            {
                messageReceiptsReadSentCallback[sequence] = callback;
            }

            zim_send_message_receipts_read(handle, messageListPtr, (uint)size, conversationID,
                (System.Int32)conversationType, ref sequence);

            ZIMUtil.ReleaseStructPointer(messageListPtr);
        }

        public override void QueryMessageReceiptsInfo(List<ZIMMessage> messageList, string conversationID, ZIMConversationType conversationType, ZIMMessageReceiptsInfoQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            var size = messageList.Count;

            var messageListPtr = AllocChangeZIMMessageListToPtrList(messageList);

            lock (messageReceiptsInfoQueriedCallbackLock)
            {
                messageReceiptsInfoQueriedCallback[sequence] = callback;
            }

            zim_query_message_receipts_info(handle, messageListPtr, (uint)size,
                conversationID, (System.Int32)conversationType, ref sequence);
            ZIMUtil.ReleaseStructPointer(messageListPtr);
        }

        public override void
        UpdateMessageLocalExtendedData(string localExtendedData,
                           ZIMMessage message,
                           ZIMMessageLocalExtendedDataUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var sMessage = ChangeZIMMessageClassToStruct(message);

            lock (messageLocalExtendedDataUpdatedCallbackLock)
            {
                messageLocalExtendedDataUpdatedCallback[sequence] = callback;
            }

            zim_update_message_local_extended_data(handle, localExtendedData, sMessage,
                                               ref sequence);
        }

        public override void SearchLocalMessages(string conversationID,
                             ZIMConversationType conversationType,
                             ZIMMessageSearchConfig config,
                             ZIMMessagesSearchedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (messagesSearchedCallbackLock)
            {
                messagesSearchedCallback[sequence] = callback;
            }


            zim_message_search_config message_search_config = new zim_message_search_config();
            message_search_config.next_message = AllocChangeZIMMessageClassToStructPtr(config.nextMessage);
            message_search_config.count = config.count;
            message_search_config.order = (int)config.order;
            var (keywordsIntptr, keywordsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            message_search_config.keywords = keywordsIntptr;
            message_search_config.keywords_length = (uint)config.keywords.Count();
            message_search_config.end_time = config.endTime;
            int[] messageTypes = ChangeZIMMessageTypeListToIntArray(config.messageTypes);
            message_search_config.message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(messageTypes);
            message_search_config.message_types_length = (uint)config.messageTypes.Count();
            var (senderUserIDsIntptr, senderUserIDsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.senderUserIDs);
            message_search_config.sender_user_ids = senderUserIDsIntptr;
            message_search_config.sender_user_ids_length = (uint)config.senderUserIDs.Count();
            message_search_config.start_time = config.startTime;
            int[] subMessageTypes = ChangeUIntListToIntArray(config.subMessageTypes);
            message_search_config.sub_message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(subMessageTypes);
            message_search_config.sub_message_types_length = (uint)config.subMessageTypes.Count();


            zim_search_local_messages(handle, conversationID,
                                  (System.Int32)conversationType, message_search_config, ref sequence);

            ZIMUtil.ReleaseStructPointerArray(keywordsPointerArray);
            ZIMUtil.ReleaseStructPointerArray(senderUserIDsPointerArray);
            ZIMUtil.ReleaseStructPointer(message_search_config.next_message);
            ZIMUtil.ReleaseStructPointer(message_search_config.keywords);
            ZIMUtil.ReleaseCoTaskMemPointer(message_search_config.message_types);
            ZIMUtil.ReleaseStructPointer(message_search_config.sender_user_ids);
            ZIMUtil.ReleaseCoTaskMemPointer(message_search_config.sub_message_types);
        }
        public override void SearchGlobalLocalMessages(ZIMMessageSearchConfig config,
                                   ZIMMessagesGlobalSearchedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (messagesGlobalSearchedCallbackLock)
            {
                messagesGlobalSearchedCallback[sequence] = callback;
            }

            zim_message_search_config message_search_config = new zim_message_search_config();
            message_search_config.next_message = AllocChangeZIMMessageClassToStructPtr(config.nextMessage);
            message_search_config.count = config.count;
            message_search_config.order = (int)config.order;
            var (keywordsIntptr, keywordsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            message_search_config.keywords = keywordsIntptr;
            message_search_config.keywords_length = (uint)config.keywords.Count();
            message_search_config.end_time = config.endTime;
            int[] messageTypes = ChangeZIMMessageTypeListToIntArray(config.messageTypes);
            message_search_config.message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(messageTypes);
            message_search_config.message_types_length = (uint)config.messageTypes.Count();
            var (senderUserIDsIntptr, senderUserIDsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.senderUserIDs);
            message_search_config.sender_user_ids = senderUserIDsIntptr;
            message_search_config.sender_user_ids_length = (uint)config.senderUserIDs.Count();
            message_search_config.start_time = config.startTime;
            int[] subMessageTypes = ChangeUIntListToIntArray(config.subMessageTypes);
            message_search_config.sub_message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(subMessageTypes);
            message_search_config.sub_message_types_length = (uint)config.subMessageTypes.Count();

            zim_search_global_local_messages(handle, message_search_config, ref sequence);

            ZIMUtil.ReleaseStructPointerArray(keywordsPointerArray);
            ZIMUtil.ReleaseStructPointerArray(senderUserIDsPointerArray);
            ZIMUtil.ReleaseStructPointer(message_search_config.next_message);
            ZIMUtil.ReleaseStructPointer(message_search_config.keywords);
            ZIMUtil.ReleaseCoTaskMemPointer(message_search_config.message_types);
            ZIMUtil.ReleaseStructPointer(message_search_config.sender_user_ids);
            ZIMUtil.ReleaseCoTaskMemPointer(message_search_config.sub_message_types);
        }
        public override void AddMessageReaction(string reactionType,
                                    ZIMMessage message,
                                    ZIMMessageReactionAddedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (messageReactionAddedCallbackLock)
            {
                messageReactionAddedCallback[sequence] = callback;
            }
            var sMessage = ChangeZIMMessageClassToStruct(message);

            zim_add_message_reaction(handle, reactionType, sMessage, ref sequence);
        }

        public override void DeleteMessageReaction(string reactionType,
                                       ZIMMessage message,
                                       ZIMMessageReactionDeletedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (messageReactionDeletedCallbackLock)
            {
                messageReactionDeletedCallback[sequence] = callback;
            }
            var sMessage = ChangeZIMMessageClassToStruct(message);

            zim_delete_message_reaction(handle, reactionType, sMessage, ref sequence);
        }

        public override void
        QueryMessageReactionUserList(ZIMMessage message,
                                 ZIMMessageReactionUserQueryConfig config,
                                 ZIMMessageReactionUserListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (messageReactionUserListQueriedCallbackLock)
            {
                messageReactionUserListQueriedCallback[sequence] = callback;
            }
            var sMessage = ChangeZIMMessageClassToStruct(message);

            var c_config = ChangeZIMMessageReactionUserQueryConfigClassToStruct(config);

            zim_query_message_reaction_user_list(handle, sMessage, c_config, ref sequence);
        }

        public override void QueryConversation(string conversationID,
                                   ZIMConversationType conversationType,
                                   ZIMConversationQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationDeletedCallbackLock)
            {
                conversationQueriedCallback[sequence] = callback;
            }

            zim_query_conversation(handle, conversationID, (System.Int32)conversationType,
                ref sequence);
        }

        public override void QueryConversationList(ZIMConversationQueryConfig config, ZIMConversationListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_conversation_query_config query_config = new zim_conversation_query_config();
            query_config.next_order_key =
                config.nextConversation != null ? config.nextConversation.orderKey : -1;
            query_config.count = config.count;

            lock (conversationListQueriedCallbackLock)
            {
                conversationListQueriedCallback[sequence] = callback;
            }

            zim_query_conversation_list(handle, query_config, ref sequence);
        }

        public override void QueryConversationPinnedList(ZIMConversationQueryConfig config,
                                             ZIMConversationPinnedListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationPinnedListQueriedCallbackLock)
            {
                conversationPinnedListQueriedCallback[sequence] = callback;
            }

            zim_conversation_query_config query_config = new zim_conversation_query_config();
            query_config.next_order_key =
                config.nextConversation != null ? config.nextConversation.orderKey : -1;
            query_config.count = config.count;

            zim_query_conversation_pinned_list(handle, query_config, ref sequence);
        }

        public override void SearchLocalConversations(ZIMConversationSearchConfig config,
                                  ZIMConversationsSearchedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationsSearchedCallbackLock)
            {
                conversationsSearchedCallback[sequence] = callback;
            }

            zim_conversation_search_config conversation_search_config = new zim_conversation_search_config();
            var (keywordsIntptr, keywordsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            conversation_search_config.keywords = keywordsIntptr;
            conversation_search_config.keywords_length = (uint)config.keywords.Count();
            conversation_search_config.end_time = config.endTime;
            int[] messageTypes = ChangeZIMMessageTypeListToIntArray(config.messageTypes);
            conversation_search_config.message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(messageTypes);
            conversation_search_config.message_types_length = (uint)config.messageTypes.Count();
            var (senderUserIDsIntptr, senderUserIDsPointerArray) = ZIMUtil.AllocGetIntPtrFromStringList(config.senderUserIDs);
            conversation_search_config.sender_user_ids = senderUserIDsIntptr;
            conversation_search_config.sender_user_ids_length = (uint)config.senderUserIDs.Count();
            conversation_search_config.start_time = config.startTime;
            int[] subMessageTypes = ChangeUIntListToIntArray(config.subMessageTypes);
            conversation_search_config.sub_message_types = ZIMUtil.AllocCoTaskMemGetIntPtrFromIntArray(subMessageTypes);
            conversation_search_config.sub_message_types_length = (uint)config.subMessageTypes.Count();
            conversation_search_config.next_flag = config.nextFlag;
            conversation_search_config.total_conversation_count = config.totalConversationCount;
            conversation_search_config.conversation_message_count = config.conversationMessageCount;

            zim_search_local_conversations(handle, conversation_search_config, ref sequence);

            ZIMUtil.ReleaseStructPointerArray(senderUserIDsPointerArray);
            ZIMUtil.ReleaseStructPointerArray(keywordsPointerArray);
            ZIMUtil.ReleaseCoTaskMemPointer(conversation_search_config.keywords);
            ZIMUtil.ReleaseCoTaskMemPointer(conversation_search_config.message_types);
            ZIMUtil.ReleaseCoTaskMemPointer(conversation_search_config.sender_user_ids);
            ZIMUtil.ReleaseCoTaskMemPointer(conversation_search_config.sub_message_types);
        }

        public override void ClearConversationUnreadMessageCount(string conversationID, ZIMConversationType conversationType, ZIMConversationUnreadMessageCountClearedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationUnreadMessageCountClearedCallbackLock)
            {
                conversationUnreadMessageCountClearedCallback[sequence] = callback;
            }

            zim_clear_conversation_unread_message_count(handle, conversationID, (System.Int32)conversationType,
                ref sequence);
        }

        public override void UpdateConversationPinnedState(bool isPinned, string conversationID,
                                  ZIMConversationType conversationType,
                                  ZIMConversationPinnedStateUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationPinnedStateUpdatedCallbackLock)
            {
                conversationPinnedStateUpdatedCallback[sequence] = callback;
            }

            zim_update_conversation_pinned_state(handle, isPinned, conversationID, (System.Int32)conversationType,
                ref sequence);
        }

        public override void DeleteConversation(string conversationID, ZIMConversationType conversationType, ZIMConversationDeleteConfig config, ZIMConversationDeletedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_conversation_delete_config delete_config = new zim_conversation_delete_config();
            delete_config.is_also_delete_server_conversation = config.isAlsoDeleteServerConversation;

            lock (conversationDeletedCallbackLock)
            {
                conversationDeletedCallback[sequence] = callback;
            }

            zim_delete_conversation(handle, conversationID, (System.Int32)conversationType,
                delete_config, ref sequence);
        }

        public override void SetConversationNotificationStatus(ZIMConversationNotificationStatus status, string conversationID, ZIMConversationType conversationType, ZIMConversationNotificationStatusSetCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationNotificationStatusSetCallbackLock)
            {
                conversationNotificationStatusSetCallback[sequence] = callback;
            }

            zim_set_conversation_notification_status(handle, (System.Int32)status,
                conversationID, (System.Int32)conversationType, ref sequence);
        }

        public override void SendConversationMessageReceiptRead(string conversationID, ZIMConversationType conversationType, ZIMConversationMessageReceiptReadSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (conversationMessageReceiptReadSentCallbackLock)
            {
                conversationMessageReceiptReadSentCallback[sequence] = callback;
            }

            zim_send_conversation_message_receipt_read(handle, conversationID,
                (System.Int32)conversationType, ref sequence);
        }

        public override void CreateRoom(ZIMRoomInfo roomInfo, ZIMRoomCreatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_room_info room_info = ChangeZIMRoomInfoClassToStruct(roomInfo);

            lock (roomCreatedCallbackLock)
            {
                roomCreatedCallback[sequence] = callback;
            }
            zim_room_advanced_config config = new zim_room_advanced_config();
            zim_create_room(handle, room_info, ref config, ref sequence);
        }

        public override void CreateRoom(ZIMRoomInfo roomInfo, ZIMRoomAdvancedConfig config, ZIMRoomCreatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_room_info room_info = ChangeZIMRoomInfoClassToStruct(roomInfo);
            var room_advanced_config = ChangeZIMRoomAdvancedConfigClassToStruct(config);
            lock (roomCreatedCallbackLock)
            {
                roomCreatedCallback[sequence] = callback;
            }
            zim_create_room(handle, room_info, ref room_advanced_config, ref sequence);
            ZIMUtil.ReleaseStructPointer(room_advanced_config.room_attributes);
        }

        public override void JoinRoom(string roomID, ZIMRoomJoinedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (roomJoinedCallbackLock)
            {
                roomJoinedCallback[sequence] = callback;
            }

            zim_join_room(handle, roomID, ref sequence);
        }

        public override void EnterRoom(ZIMRoomInfo roomInfo, ZIMRoomAdvancedConfig config, ZIMRoomEnteredCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_room_info room_info = ChangeZIMRoomInfoClassToStruct(roomInfo);
            var room_advanced_config = ChangeZIMRoomAdvancedConfigClassToStruct(config);
            lock (roomEnteredCallbackLock)
            {
                roomEnteredCallback[sequence] = callback;
            }

            zim_enter_room(handle, room_info, ref room_advanced_config, ref sequence);
            ZIMUtil.ReleaseStructPointer(room_advanced_config.room_attributes);
        }

        public override void LeaveRoom(string roomID, ZIMRoomLeftCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomLeftCallbackLock)
            {
                roomLeftCallback[sequence] = callback;
            }

            zim_leave_room(handle, roomID, ref sequence);
        }

        public override void QueryRoomMemberList(string roomID, ZIMRoomMemberQueryConfig config, ZIMRoomMemberQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var room_member_query_config = ChangeZIMRoomMemberQueryConfigClassToStruct(config);
            lock (roomMemberQueriedCallbackLock)
            {
                roomMemberQueriedCallback[sequence] = callback;
            }

            zim_query_room_member_list(handle, roomID, room_member_query_config, ref sequence);
        }

        public override void QueryRoomMembers(List<string> userIDs, string room_id,
                                  ZIMRoomMembersQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomMembersQueriedCallbackLock)
            {
                roomMembersQueriedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_query_room_members(handle, userIDArray, (uint)size, room_id, ref sequence);
        }

        public override void QueryRoomOnlineMemberCount(string roomID, ZIMRoomOnlineMemberCountQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomOnlineMemberCountQueriedCallbackLock)
            {
                roomOnlineMemberCountQueriedCallback[sequence] = callback;
            }

            zim_query_room_online_member_count(handle, roomID, ref sequence);
        }

        public override void SetRoomAttributes(Dictionary<string, string> roomAttributes, string roomID, ZIMRoomAttributesSetConfig config, ZIMRoomAttributesOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var roomAttributesCount = roomAttributes.Count;
            var roomAttributesListPtr = AllocChangeZIMRoomAttributesDictionaryToPtr(roomAttributes);

            zim_room_attributes_set_config room_attributes_set_config = new zim_room_attributes_set_config();
            room_attributes_set_config.is_force = config.isForce;
            room_attributes_set_config.is_delete_after_owner_left = config.isDeleteAfterOwnerLeft;
            room_attributes_set_config.is_update_owner = config.isUpdateOwner;

            lock (roomAttributesOperatedCallbackLock)
            {
                roomAttributesOperatedCallback[sequence] = callback;
            }

            zim_set_room_attributes(handle, roomAttributesListPtr, (uint)roomAttributesCount, roomID,
                room_attributes_set_config, ref sequence);
            ZIMUtil.ReleaseStructPointer(roomAttributesListPtr);

        }

        public override void DeleteRoomAttributes(List<string> keys, string roomID, ZIMRoomAttributesDeleteConfig config, ZIMRoomAttributesOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_room_attributes_delete_config room_attributes_delete_config = new zim_room_attributes_delete_config();
            room_attributes_delete_config.is_force = config.isForce;

            lock (roomAttributesOperatedCallbackLock)
            {
                roomAttributesOperatedCallback[sequence] = callback;
            }

            int size = keys.Count;
            string[] keyArray = new string[size];
            int i = 0;
            foreach (string key in keys)
            {
                keyArray[i] = key;
                i++;
            }

            zim_delete_room_attributes(handle, keyArray, (uint)size, roomID, room_attributes_delete_config, ref sequence);
        }

        public override void BeginRoomAttributesBatchOperation(string roomID, ZIMRoomAttributesBatchOperationConfig config)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_room_attributes_batch_operation_config room_attributes_batch_operation_config = new zim_room_attributes_batch_operation_config();
            room_attributes_batch_operation_config.is_force = config.isForce;
            room_attributes_batch_operation_config.is_delete_after_owner_left = config.isDeleteAfterOwnerLeft;
            room_attributes_batch_operation_config.is_update_owner = config.isUpdateOwner;

            zim_begin_room_attributes_batch_operation(handle, roomID, room_attributes_batch_operation_config, ref sequence);
        }

        public override void EndRoomAttributesBatchOperation(string roomID, ZIMRoomAttributesBatchOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomAttributesBatchOperatedCallbackLock)
            {
                roomAttributesBatchOperatedCallback[sequence] = callback;
            }

            zim_end_room_attributes_batch_operation(handle, roomID, ref sequence);
        }

        public override void QueryRoomAllAttributes(string roomID, ZIMRoomAttributesQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomAttributesQueriedCallbackLock)
            {
                roomAttributesQueriedCallback[sequence] = callback;
            }

            zim_query_room_all_attributes(handle, roomID, ref sequence);
        }

        public override void SetRoomMembersAttributes(Dictionary<string, string> attributes, List<string> userIDs, string roomID, ZIMRoomMemberAttributesSetConfig config, ZIMRoomMembersAttributesOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomMembersAttributesOperatedCallbackLock)
            {
                roomMembersAttributesOperatedCallback[sequence] = callback;
            }

            var room_member_attributes = ChangeZIMRoomMemberAttributesDictionaryToStructList(attributes);

            var structListPtr = ZIMUtil.AllocGetIntPtrFromStructList<zim_room_member_attribute>(room_member_attributes);

            zim_room_member_attributes_set_config room_member_attributes_set_config = new zim_room_member_attributes_set_config();
            room_member_attributes_set_config.is_delete_after_owner_left =
                config.isDeleteAfterOwnerLeft;

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_set_room_members_attributes(handle, structListPtr, (uint)attributes.Count, userIDArray,
                (uint)size,
                roomID, room_member_attributes_set_config, ref sequence);

            ZIMUtil.ReleaseStructPointer(structListPtr);
        }

        public override void QueryRoomMembersAttributes(List<string> userIDs, string roomID, ZIMRoomMembersAttributesQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomMembersAttributesQueriedCallbackLock)
            {
                roomMembersAttributesQueriedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_query_room_members_attributes(handle, userIDArray, (uint)size, roomID, ref sequence);
        }

        public override void QueryRoomMemberAttributesList(string roomID, ZIMRoomMemberAttributesQueryConfig config, ZIMRoomMemberAttributesListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (roomMemberAttributesListQueriedCallbackLock)
            {
                roomMemberAttributesListQueriedCallback[sequence] = callback;
            }

            zim_room_member_attributes_query_config room_member_attributes_query_config = new zim_room_member_attributes_query_config();
            room_member_attributes_query_config.next_flag = config.nextFlag == null ? "" : config.nextFlag;
            room_member_attributes_query_config.count = config.count;

            zim_query_room_member_attributes_list(handle, roomID, room_member_attributes_query_config,
                ref sequence);
        }

        public override void CreateGroup(ZIMGroupInfo groupInfo, List<string> userIDs, ZIMGroupCreatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var group_info = ChangeZIMGroupBaseInfoClassToStruct(groupInfo);

            lock (groupCreatedCallbackLock)
            {
                groupCreatedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_group_advanced_config group_advanced_config = new zim_group_advanced_config();
            zim_create_group(handle, group_info, userIDArray, (uint)size, ref group_advanced_config, ref sequence);
        }

        public override void CreateGroup(ZIMGroupInfo groupInfo, List<string> userIDs, ZIMGroupAdvancedConfig config, ZIMGroupCreatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var group_info = ChangeZIMGroupBaseInfoClassToStruct(groupInfo);
            var group_advanced_config = ChangeZIMGroupAdvancedConfigClassToStruct(config);
            lock (groupCreatedCallbackLock)
            {
                groupCreatedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_create_group(handle, group_info, userIDArray, (uint)size, ref group_advanced_config, ref sequence);

            ZIMUtil.ReleaseStructPointer(group_advanced_config.group_attributes);
        }

        public override void SearchLocalGroups(ZIMGroupSearchConfig config, ZIMGroupsSearchedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (groupsSearchedCallbackLock)
            {
                groupsSearchedCallback[sequence] = callback;
            }

            zim_group_search_config group_search_config = new zim_group_search_config();
            var (keywordsIntptr, keywordArrayPointer) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            group_search_config.keywords = keywordsIntptr;
            group_search_config.keywords_length = (uint)config.keywords.Count();
            group_search_config.is_also_match_group_member_nick_name = config.isAlsoMatchGroupMemberNickname;
            group_search_config.is_also_match_group_member_user_name = config.isAlsoMatchGroupMemberUserName;
            group_search_config.next_flag = config.nextFlag;
            group_search_config.count = config.count;

            zim_search_local_groups(handle, group_search_config, ref sequence);

            ZIMUtil.ReleaseStructPointerArray(keywordArrayPointer);
            ZIMUtil.ReleaseStructPointer(group_search_config.keywords);
        }

        public override void SearchLocalGroupMembers(string groupID, ZIMGroupMemberSearchConfig config, ZIMGroupMembersSearchedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            lock (groupMembersSearchedCallbackLock)
            {
                groupMembersSearchedCallback[sequence] = callback;
            }

            zim_group_member_search_config group_member_search_config = new zim_group_member_search_config();
            var (keywordsIntptr, keywordArrayPointer) = ZIMUtil.AllocGetIntPtrFromStringList(config.keywords);
            group_member_search_config.keywords = keywordsIntptr;
            group_member_search_config.keywords_length = (uint)config.keywords.Count();
            group_member_search_config.is_also_match_group_member_nick_name = config.isAlsoMatchGroupMemberNickname;
            group_member_search_config.next_flag = config.nextFlag;
            group_member_search_config.count = config.count;

            zim_search_local_group_members(handle, groupID, group_member_search_config, ref sequence);

            ZIMUtil.ReleaseStructPointerArray(keywordArrayPointer);
            ZIMUtil.ReleaseStructPointer(group_member_search_config.keywords);
        }

        public override void DismissGroup(string groupID, ZIMGroupDismissedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupDismissedCallbackLock)
            {
                groupDismissedCallback[sequence] = callback;
            }

            zim_dismiss_group(handle, groupID, ref sequence);
        }

        public override void JoinGroup(string groupID, ZIMGroupJoinedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupJoinedCallbackLock)
            {
                groupJoinedCallback[sequence] = callback;
            }

            zim_join_group(handle, groupID, ref sequence);
        }

        public override void LeaveGroup(string groupID, ZIMGroupLeftCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupLeftCallbackLock)
            {
                groupLeftCallback[sequence] = callback;
            }

            zim_leave_group(handle, groupID, ref sequence);
        }

        public override void InviteUsersIntoGroup(List<string> userIDs, string groupID, ZIMGroupUsersInvitedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupUsersInvitedCallbackLock)
            {
                groupUsersInvitedCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }

            zim_invite_users_into_group(handle, userIDArray, (uint)size, groupID, ref sequence);
        }

        public override void KickGroupMembers(List<string> userIDs, string groupID, ZIMGroupMemberKickedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupMemberKickedoutCallbackLock)
            {
                groupMemberKickedoutCallback[sequence] = callback;
            }

            int size = userIDs.Count;
            string[] userIDArray = new string[size];
            int i = 0;
            foreach (string userID in userIDs)
            {
                userIDArray[i] = userID;
                i++;
            }
            zim_kick_group_members(handle, userIDArray, (uint)size, groupID, ref sequence);
        }

        public override void TransferGroupOwner(string toUserID, string groupID, ZIMGroupOwnerTransferredCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupOwnerTransferredCallbackLock)
            {
                groupOwnerTransferredCallback[sequence] = callback;
            }

            zim_transfer_group_owner(handle, toUserID, groupID, ref sequence);
        }

        public override void UpdateGroupName(string groupName, string groupID, ZIMGroupNameUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupNameUpdatedCallbackLock)
            {
                groupNameUpdatedCallback[sequence] = callback;
            }

            zim_update_group_name(handle, groupID, groupName, ref sequence);
        }

        public override void UpdateGroupAvatarUrl(string groupAvatarUrl, string groupID, ZIMGroupAvatarUrlUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupAvatarUrlUpdatedCallbackLock)
            {
                groupAvatarUrlUpdatedCallback[sequence] = callback;
            }

            zim_update_group_avatar_url(handle, groupAvatarUrl, groupID, ref sequence);
        }

        public override void UpdateGroupNotice(string groupNotice, string groupID, ZIMGroupNoticeUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupNoticeUpdatedCallbackLock)
            {
                groupNoticeUpdatedCallback[sequence] = callback;
            }

            zim_update_group_notice(handle, groupNotice, groupID, ref sequence);
        }

        public override void QueryGroupInfo(string groupID, ZIMGroupInfoQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupInfoQueriedCallbackLock)
            {
                groupInfoQueriedCallback[sequence] = callback;
            }

            zim_query_group_info(handle, groupID, ref sequence);
        }

        public override void SetGroupAttributes(Dictionary<string, string> groupAttributes, string groupID, ZIMGroupAttributesOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupAttributesOperatedCallbackLock)
            {
                groupAttributesOperatedCallback[sequence] = callback;
            }

            var groupAttributesCount = groupAttributes.Count;
            var groupAttributesListPtr = AllocChangeZIMGroupAttributesDictionaryToPtr(groupAttributes);

            zim_set_group_attributes(handle, groupAttributesListPtr, (uint)groupAttributesCount, groupID, ref sequence);

            ZIMUtil.ReleaseStructPointer(groupAttributesListPtr);
        }

        public override void DeleteGroupAttributes(List<string> keys, string groupID, ZIMGroupAttributesOperatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupAttributesOperatedCallbackLock)
            {
                groupAttributesOperatedCallback[sequence] = callback;
            }

            int size = keys.Count;
            string[] keyArray = new string[size];
            int i = 0;
            foreach (string key in keys)
            {
                keyArray[i] = key;
                i++;
            }

            zim_delete_group_attributes(handle, keyArray, (uint)size, groupID, ref sequence);
        }

        public override void QueryGroupAttributes(List<string> keys, string groupID, ZIMGroupAttributesQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupAttributesQueriedCallbackLock)
            {
                groupAttributesQueriedCallback[sequence] = callback;
            }

            int size = keys.Count;
            string[] keyArray = new string[size];
            int i = 0;
            foreach (string key in keys)
            {
                keyArray[i] = key;
                i++;
            }

            zim_query_group_attributes(handle, keyArray, (uint)size, groupID, ref sequence);
        }

        public override void QueryGroupAllAttributes(string groupID, ZIMGroupAttributesQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupAttributesQueriedCallbackLock)
            {
                groupAttributesQueriedCallback[sequence] = callback;
            }

            zim_query_group_all_attributes(handle, groupID, ref sequence);
        }

        public override void SetGroupMemberRole(int role, string forUserID, string groupID, ZIMGroupMemberRoleUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupMemberRoleUpdatedCallbackLock)
            {
                groupMemberRoleUpdatedCallback[sequence] = callback;
            }

            zim_set_group_member_role(handle, role, forUserID, groupID, ref sequence);
        }

        public override void SetGroupMemberNickname(string nickname, string forUserID, string groupID, ZIMGroupMemberNicknameUpdatedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupMemberNicknameUpdatedCallbackLock)
            {
                groupMemberNicknameUpdatedCallback[sequence] = callback;
            }

            zim_set_group_member_nickname(handle, nickname, forUserID, groupID, ref sequence);
        }

        public override void QueryGroupMemberInfo(string userID, string groupID, ZIMGroupMemberInfoQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupMemberInfoQueriedCallbackLock)
            {
                groupMemberInfoQueriedCallback[sequence] = callback;
            }

            zim_query_group_member_info(handle, userID, groupID, ref sequence);
        }

        public override void QueryGroupList(ZIMGroupListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupListQueriedCallbackLock)
            {
                groupListQueriedCallback[sequence] = callback;
            }

            zim_query_group_list(handle, ref sequence);
        }

        public override void QueryGroupMemberList(string groupID, ZIMGroupMemberQueryConfig config, ZIMGroupMemberListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            zim_group_member_query_config group_member_query_config = new zim_group_member_query_config();
            group_member_query_config.count = config.count;
            group_member_query_config.next_flag = config.nextFlag;

            lock (groupMemberListQueriedCallbackLock)
            {
                groupMemberListQueriedCallback[sequence] = callback;
            }

            zim_query_group_member_list(handle, groupID, group_member_query_config, ref sequence);
        }

        public override void QueryGroupMemberCount(string groupID, ZIMGroupMemberCountQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (groupMemberCountQueriedCallbackLock)
            {
                groupMemberCountQueriedCallback[sequence] = callback;
            }

            zim_query_group_member_count(handle, groupID, ref sequence);
        }

        public override void QueryGroupMessageReceiptReadMemberList(ZIMMessage message, string groupID, ZIMGroupMessageReceiptMemberQueryConfig config, ZIMGroupMessageReceiptMemberListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var sMessage = ChangeZIMMessageClassToStruct(message);
            zim_group_message_receipt_member_query_config group_message_receipt_member_query_config = new zim_group_message_receipt_member_query_config();
            group_message_receipt_member_query_config.count = config.count;
            group_message_receipt_member_query_config.next_flag = config.nextFlag;

            lock (groupMessageReceiptMemberListQueriedCallbackLock)
            {
                groupMessageReceiptMemberListQueriedCallback[sequence] = callback;
            }

            zim_query_group_message_receipt_read_member_list(handle, sMessage, groupID, group_message_receipt_member_query_config,
                ref sequence);
        }

        public override void QueryGroupMessageReceiptUnreadMemberList(ZIMMessage message, string groupID, ZIMGroupMessageReceiptMemberQueryConfig config, ZIMGroupMessageReceiptMemberListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);
            var sMessage = ChangeZIMMessageClassToStruct(message);
            zim_group_message_receipt_member_query_config group_message_receipt_member_query_config = new zim_group_message_receipt_member_query_config();
            group_message_receipt_member_query_config.count = config.count;
            group_message_receipt_member_query_config.next_flag = config.nextFlag;
            lock (groupMessageReceiptMemberListQueriedCallbackLock)
            {
                groupMessageReceiptMemberListQueriedCallback[sequence] = callback;
            }

            zim_query_group_message_receipt_unread_member_list(handle, sMessage, groupID,
                group_message_receipt_member_query_config, ref sequence);
        }

        public override void CallInvite(List<string> invitees, ZIMCallInviteConfig config, ZIMCallInvitationSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callInvitationSentCallbackLock)
            {
                callInvitationSentCallback[sequence] = callback;
            }

            zim_call_invite_config call_invite_config = ChangeZIMCallInviteConfigClassToStruct(config);


            int size = invitees.Count;
            string[] inviteeArray = new string[size];
            int i = 0;
            foreach (string invitee in invitees)
            {
                inviteeArray[i] = invitee;
                i++;
            }

            zim_call_invite(handle, inviteeArray, (uint)size, call_invite_config, ref sequence);
        }

        public override void CallingInvite(List<string> invitees, string callID,
                               ZIMCallingInviteConfig config,
                               ZIMCallingInvitationSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callingInvitationSentCallbackLock)
            {
                callingInvitationSentCallback[sequence] = callback;
            }

            zim_calling_invite_config call_invite_config = ChangeZIMCallingInviteConfigClassToStruct(config);

            int size = invitees.Count;
            string[] inviteeArray = new string[size];
            int i = 0;
            foreach (string invitee in invitees)
            {
                inviteeArray[i] = invitee;
                i++;
            }

            zim_calling_invite(handle, callID, inviteeArray,
                           (uint)size, call_invite_config, ref sequence);
        }

        public override void CallCancel(List<string> invitees, string callID, ZIMCallCancelConfig config, ZIMCallCancelSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callCancelSentCallbackLock)
            {
                callCancelSentCallback[sequence] = callback;
            }

            zim_call_cancel_config call_cancel_config = ChangeZIMCallCancelConfigClassToStruct(config);

            int size = invitees.Count;
            string[] inviteeArray = new string[size];
            int i = 0;
            foreach (string invitee in invitees)
            {
                inviteeArray[i] = invitee;
                i++;
            }

            zim_call_cancel(handle, callID, inviteeArray, (uint)size, call_cancel_config, ref sequence);
        }

        public override void CallAccept(string callID, ZIMCallAcceptConfig config, ZIMCallAcceptanceSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callAcceptanceSentCallbackLock)
            {
                callAcceptanceSentCallback[sequence] = callback;
            }

            zim_call_accept_config call_accept_config = ChangeZIMCallAcceptConfigClassToStruct(config);

            zim_call_accept(handle, callID, call_accept_config, ref sequence);
        }

        public override void CallReject(string callID, ZIMCallRejectConfig config, ZIMCallRejectionSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callRejectionSentCallbackLock)
            {
                callRejectionSentCallback[sequence] = callback;
            }

            zim_call_reject_config call_reject_config = ChangeZIMCallRejectConfigClassToStruct(config);

            zim_call_reject(handle, callID, call_reject_config, ref sequence);
        }

        public override void CallQuit(string callID, ZIMCallQuitConfig config,
                  ZIMCallQuitSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callQuitSentCallbackLock)
            {
                callQuitSentCallback[sequence] = callback;
            }
            zim_call_quit_config call_quit_config = ChangeZIMCallQuitConfigClassToStruct(config);

            zim_call_quit(handle, callID, call_quit_config, ref sequence);
        }

        public override void CallEnd(string callID, ZIMCallEndConfig config,
                         ZIMCallEndSentCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callEndSentCallbackLock)
            {
                callEndSentCallback[sequence] = callback;
            }

            zim_call_end_config call_end_config = ChangeZIMCallEndConfigClassToStruct(config);

            zim_call_end(handle, callID, call_end_config, ref sequence);
        }

        public override void QueryCallInvitationList(ZIMCallInvitationQueryConfig config,
                                         ZIMCallInvitationListQueriedCallback callback)
        {
            int sequence = Interlocked.Increment(ref this.sequence);

            lock (callInvitationListQueriedCallbackLock)
            {
                callInvitationListQueriedCallback[sequence] = callback;
            }

            zim_query_call_list_config query_call_list_config = ChangeZIMCallInvitationQueryConfigClassToStruct(config);

            zim_query_call_list(handle, query_call_list_config, ref sequence);
        }

        private void ZIMRegisterCallback()
        {
            // main callback
            zim_register_log_uploaded_callback(handle, ZIMBridge.zim_on_log_uploaded_callback);
            zim_register_logged_in_callback(handle, ZIMBridge.zim_on_logged_in_callback);
            zim_register_token_renewed_callback(handle, ZIMBridge.zim_on_token_renewed_callback);
            zim_register_user_name_updated_callback(handle, ZIMBridge.zim_on_user_name_updated_callback);
            zim_register_user_avatar_url_updated_callback(handle, ZIMBridge.zim_on_user_avatar_url_updated_callback);
            zim_register_user_extended_data_updated_callback(handle,
                                                             ZIMBridge.zim_on_user_extended_data_updated_callback);
            zim_register_users_info_queried_callback(handle, ZIMBridge.zim_on_users_info_queried_callback);

            // message callback
            zim_register_message_sent_callback(handle, ZIMBridge.zim_on_message_sent_callback);
            zim_register_message_queried_callback(handle, ZIMBridge.zim_on_message_queried_callback);
            zim_register_message_deleted_callback(handle, ZIMBridge.zim_on_message_deleted_callback);
            zim_register_message_inserted_callback(handle, ZIMBridge.zim_on_message_inserted_callback);
            zim_register_message_revoked_callback(handle, ZIMBridge.zim_on_message_revoked_callback);
            zim_register_message_attached_callback(handle, ZIMBridge.zim_on_message_attached_callback);
            zim_register_media_uploading_progress_callback(handle, ZIMBridge.zim_on_media_uploading_progress_callback);
            zim_register_media_downloaded_callback(handle, ZIMBridge.zim_on_media_downloaded_callback);
            zim_register_media_downloading_progress_callback(handle,
                                                             ZIMBridge.zim_on_media_downloading_progress_callback);
            zim_register_message_receipts_read_sent_callback(handle, ZIMBridge.zim_on_message_receipts_read_sent_callback);
            zim_register_message_receipts_info_queried_callback(handle,
                                                                ZIMBridge.zim_on_message_receipts_info_queried_callback);
            
            zim_register_message_local_extended_data_updated_callback(
            handle, ZIMBridge.zim_on_message_local_extended_data_updated_callback);
            zim_register_messages_searched_callback(handle, ZIMBridge.zim_on_messages_searched_callback);
            zim_register_messages_global_searched_callback(handle, ZIMBridge.zim_on_messages_global_searched_callback);
            zim_register_conversations_searched_callback(handle, ZIMBridge.zim_on_conversations_searched_callback);
            zim_register_message_added_reaction_callback(handle, ZIMBridge.on_message_reaction_added_callback);
            zim_register_message_deleted_reaction_callback(handle, ZIMBridge.on_message_reaction_deleted_callback);
            zim_register_message_queried_reaction_user_list_callback(
                handle, ZIMBridge.on_message_reaction_users_queried_callback);

            // conversation callback
            zim_register_conversation_pinned_list_queried_callback(handle,
            ZIMBridge.zim_on_conversation_pinned_list_queried_callback);
            zim_register_conversation_pinned_state_updated_callback(handle,
            ZIMBridge.zim_on_conversation_pinned_state_updated_callback);
            zim_register_conversation_queried_callback(handle,
            ZIMBridge.zim_on_conversation_queried_callback);
            zim_register_conversation_list_queried_callback(handle,
                                                            ZIMBridge.zim_on_conversation_list_queried_callback);
            zim_register_conversation_deleted_callback(handle, ZIMBridge.zim_on_conversation_deleted_callback);
            zim_register_conversation_unread_message_count_cleared_callback(
                handle, ZIMBridge.zim_on_conversation_unread_message_count_cleared_callback);
            zim_register_conversation_message_receipt_read_sent_callback(
                handle, ZIMBridge.zim_on_conversation_message_receipt_read_sent_callback);
            zim_register_conversation_notification_status_set_callback(
                handle, ZIMBridge.zim_on_conversation_notification_status_set_callback);

            // room callback
            zim_register_room_created_callback(handle, ZIMBridge.zim_on_room_created_callback);
            zim_register_room_joined_callback(handle, ZIMBridge.zim_on_room_joined_callback);
            zim_register_room_entered_callback(handle, ZIMBridge.zim_on_room_entered_callback);
            zim_register_room_left_callback(handle, ZIMBridge.zim_on_room_left_callback);
            zim_register_room_member_queried_callback(handle, ZIMBridge.zim_on_room_member_queried_callback);
            zim_register_room_members_queried_callback(handle,
            ZIMBridge.zim_on_room_members_queried);
            zim_register_room_online_member_count_queried_callback(
                handle, ZIMBridge.zim_on_room_online_member_count_queried_callback);
            zim_register_room_attributes_operated_callback(handle, ZIMBridge.zim_on_room_attributes_operated_callback);
            zim_register_room_attributes_queried_callback(handle, ZIMBridge.zim_on_room_attributes_queried_callback);
            zim_register_room_attributes_batch_operated_callback(
                handle, ZIMBridge.zim_on_room_attributes_batch_operated_callback);
            zim_register_room_members_attributes_operated_callback(
                handle, ZIMBridge.zim_on_room_members_attributes_operated_callback);
            zim_register_room_members_attributes_queried_callback(
                handle, ZIMBridge.zim_on_room_members_attributes_queried_callback);
            zim_register_room_member_attributes_list_queried_callback(
                handle, ZIMBridge.zim_on_room_member_attributes_list_queried_callback);

            // group callback
            zim_register_group_created_callback(handle, ZIMBridge.zim_on_group_created_callback);
            zim_register_group_dismissed_callback(handle, ZIMBridge.zim_on_group_dismissed_callback);
            zim_register_group_joined_callback(handle, ZIMBridge.zim_on_group_joined_callback);
            zim_register_group_left_callback(handle, ZIMBridge.zim_on_group_left_callback);
            zim_register_group_users_invited_callback(handle, ZIMBridge.zim_on_group_users_invited_callback);
            zim_register_group_member_kicked_callback(handle, ZIMBridge.zim_on_group_member_kicked_callback);
            zim_register_group_owner_transferred_callback(handle, ZIMBridge.zim_on_group_owner_transferred_callback);
            zim_register_group_name_updated_callback(handle, ZIMBridge.zim_on_group_name_updated_callback);
            zim_register_group_avatar_url_updated_callback(handle, ZIMBridge.zim_on_group_avatar_url_updated_callback);
            zim_register_group_notice_updated_callback(handle, ZIMBridge.zim_on_group_notice_updated_callback);
            zim_register_group_info_queried_callback(handle, ZIMBridge.zim_on_group_info_queried_callback);
            zim_register_group_attributes_operated_callback(handle,
                                                            ZIMBridge.zim_on_group_attributes_operated_callback);
            zim_register_group_attributes_queried_callback(handle, ZIMBridge.zim_on_group_attributes_queried_callback);
            zim_register_group_member_nickname_updated_callback(
                handle, ZIMBridge.zim_on_group_member_nickname_updated_callback);
            zim_register_group_member_role_updated_callback(handle,
                                                            ZIMBridge.zim_on_group_member_role_updated_callback);
            zim_register_group_member_info_queried_callback(
                handle, ZIMBridge.zim_on_group_member_info_queried_callback);
            zim_register_group_list_queried_callback(handle, ZIMBridge.zim_on_group_list_queried_callback);
            zim_register_group_member_list_queried_callback(handle,
                                                            ZIMBridge.zim_on_group_member_list_queried_callback);
            zim_register_group_member_count_queried_callback(handle,
                                                             ZIMBridge.zim_on_group_member_count_queried_callback);
            zim_register_group_message_receipt_member_list_queried_callback(
                handle, ZIMBridge.zim_on_group_message_receipt_member_list_queried_callback);
            zim_register_groups_searched_callback(handle, ZIMBridge.zim_on_groups_searched_callback);
            zim_register_group_members_searched_callback(handle, ZIMBridge.zim_on_group_members_searched_callback);

            // call callback
            zim_register_call_invitation_sent_callback(handle, ZIMBridge.zim_on_call_invitation_sent_callback);
            zim_register_call_cancel_sent_callback(handle, ZIMBridge.zim_on_call_cancel_sent_callback);
            zim_register_call_acceptance_sent_callback(handle, ZIMBridge.zim_on_call_acceptance_sent_callback);
            zim_register_call_rejection_sent_callback(handle, ZIMBridge.zim_on_call_rejection_sent_callback);
            zim_register_calling_invitation_sent_callback(handle,
                                              ZIMBridge.zim_on_calling_invitation_sent_callback);
            zim_register_call_quit_sent_callback(handle, ZIMBridge.zim_on_call_quit_sent_callback);
            zim_register_call_end_sent_callback(handle, ZIMBridge.zim_on_call_end_sent_callback);
            zim_register_call_list_queried_callback(handle, ZIMBridge.zim_on_call_list_queried_callback);

            // main event
            zim_register_error_event(handle, ZIMBridge.zim_on_error_event);
            zim_register_connection_state_changed_event(handle, ZIMBridge.zim_on_connection_state_changed_event);
            zim_register_token_will_expire_event(handle, ZIMBridge.zim_on_token_will_expire_event);

            // message event
            zim_register_receive_peer_message_event(handle, ZIMBridge.zim_on_receive_peer_message_event);
            zim_register_receive_room_message_event(handle, ZIMBridge.zim_on_receive_room_message_event);
            zim_register_receive_group_message_event(handle, ZIMBridge.zim_on_receive_group_message_event);
            zim_register_message_receipt_changed_event(handle, ZIMBridge.zim_on_message_receipt_changed_event);
            zim_register_message_revoke_received_event(handle, ZIMBridge.zim_on_message_revoke_received_event);
            zim_register_message_sent_status_changed_event(handle,
                                                           ZIMBridge.zim_on_message_sent_status_changed_event);
            zim_register_broadcast_message_received_event(handle, ZIMBridge.zim_on_broadcast_message_received_event);
            zim_register_message_reactions_changed_event(handle, ZIMBridge.zim_on_message_reactions_changed_event);

            // conversation event
            zim_register_conversation_changed_event(handle, ZIMBridge.zim_on_conversation_changed_event);
            zim_register_conversation_total_unread_message_count_updated_event(
                handle, ZIMBridge.zim_on_conversation_total_unread_message_count_updated_event);

            // room event
            zim_register_room_state_changed_event(handle, ZIMBridge.zim_on_room_state_changed_event);
            zim_register_room_member_joined_event(handle, ZIMBridge.zim_on_room_member_joined_event);
            zim_register_room_member_left_event(handle, ZIMBridge.zim_on_room_member_left_event);
            zim_register_room_attributes_updated_event(handle, ZIMBridge.zim_on_room_attributes_updated_event);
            zim_register_room_attributes_batch_updated_event(handle,
                                                             ZIMBridge.zim_on_room_attributes_batch_updated_event);
            zim_register_room_member_attributes_updated_event(
                handle, ZIMBridge.zim_on_room_member_attributes_updated_event);

            // group event
            zim_register_group_state_changed_event(handle, ZIMBridge.zim_on_group_state_changed_event);
            zim_register_group_member_state_changed_event(handle, ZIMBridge.zim_on_group_member_state_changed_event);
            zim_register_group_name_updated_event(handle, ZIMBridge.zim_on_group_name_updated_event);
            zim_register_group_avatar_url_updated_event(handle, ZIMBridge.zim_on_group_avatar_url_updated_event);
            zim_register_group_notice_updated_event(handle, ZIMBridge.zim_on_group_notice_updated_event);
            zim_register_group_attributes_updated_event(handle, ZIMBridge.zim_on_group_attributes_updated_event);
            zim_register_group_member_info_updated_event(handle, ZIMBridge.zim_on_group_member_info_updated_event);

            // call event
            zim_register_call_invitation_received_event(handle, ZIMBridge.zim_on_call_invitation_received_event);
            zim_register_call_invitation_cancelled_event(handle, ZIMBridge.zim_on_call_invitation_cancelled_event);
            zim_register_call_invitation_accepted_event(handle, ZIMBridge.zim_on_call_invitation_accepted_event);
            zim_register_call_invitation_rejected_event(handle, ZIMBridge.zim_on_call_invitation_rejected_event);
            zim_register_call_invitation_timeout_event(handle, ZIMBridge.zim_on_call_invitation_timeout_event);
            zim_register_call_invitees_answered_timeout_event(handle,
                                                              ZIMBridge.zim_on_call_invitees_answered_timeout_event);
            zim_register_call_user_state_changed_event(handle, ZIMBridge.zim_on_call_user_state_changed_event);
            zim_register_call_invitation_ended_event(handle, ZIMBridge.zim_on_call_invitation_ended_event);
        }

        public static ZIMImpl zim = null;

        static IntPtr jvm = IntPtr.Zero;

        static IntPtr context = IntPtr.Zero;

        public System.IntPtr handle = System.IntPtr.Zero;

        public static object loggedInCallbackLock = new object();
        public Dictionary<int, ZIMLoggedInCallback> loggedInCallback = new Dictionary<int, ZIMLoggedInCallback>();

        public static object tokenRenewedCallbackLock = new object();
        public Dictionary<int, ZIMTokenRenewedCallback> tokenRenewedCallback = new Dictionary<int, ZIMTokenRenewedCallback>();

        public static object userNameUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMUserNameUpdatedCallback> userNameUpdatedCallback = new Dictionary<int, ZIMUserNameUpdatedCallback>();

        public static object userAvatarUrlUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMUserAvatarUrlUpdatedCallback> userAvatarUrlUpdatedCallback = new Dictionary<int, ZIMUserAvatarUrlUpdatedCallback>();

        public static object userExtendedDataUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMUserExtendedDataUpdatedCallback> userExtendedDataUpdatedCallback = new Dictionary<int, ZIMUserExtendedDataUpdatedCallback>();

        public static object usersInfoQueriedCallbackLock = new object();
        public Dictionary<int, ZIMUsersInfoQueriedCallback> usersInfoQueriedCallback = new Dictionary<int, ZIMUsersInfoQueriedCallback>();

        public static object roomMemberQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomMemberQueriedCallback> roomMemberQueriedCallback = new Dictionary<int, ZIMRoomMemberQueriedCallback>();

        public static object roomMembersQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomMembersQueriedCallback> roomMembersQueriedCallback = new Dictionary<int, ZIMRoomMembersQueriedCallback>();

        public static object messageQueriedCallbackLock = new object();
        public Dictionary<int, ZIMMessageQueriedCallback> messageQueriedCallback = new Dictionary<int, ZIMMessageQueriedCallback>();

        public static object messageSentCallbackLock = new object();
        public Dictionary<int, ZIMMessageSentCallback> messageSentCallback = new Dictionary<int, ZIMMessageSentCallback>();

        public static object messageInsertedCallbackLock = new object();
        public Dictionary<int, ZIMMessageInsertedCallback> messageInsertedCallback = new Dictionary<int, ZIMMessageInsertedCallback>();

        public static object messageReceiptsReadSentCallbackLock = new object();
        public Dictionary<int, ZIMMessageReceiptsReadSentCallback> messageReceiptsReadSentCallback = new Dictionary<int, ZIMMessageReceiptsReadSentCallback>();

        public static object messageLocalExtendedDataUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMMessageLocalExtendedDataUpdatedCallback> messageLocalExtendedDataUpdatedCallback = new Dictionary<int, ZIMMessageLocalExtendedDataUpdatedCallback>();

        public static object messagesSearchedCallbackLock = new object();
        public Dictionary<int, ZIMMessagesSearchedCallback> messagesSearchedCallback = new Dictionary<int, ZIMMessagesSearchedCallback>();

        public static object messagesGlobalSearchedCallbackLock = new object();
        public Dictionary<int, ZIMMessagesGlobalSearchedCallback> messagesGlobalSearchedCallback = new Dictionary<int, ZIMMessagesGlobalSearchedCallback>();

        public static object messageReactionAddedCallbackLock = new object();
        public Dictionary<int, ZIMMessageReactionAddedCallback> messageReactionAddedCallback = new Dictionary<int, ZIMMessageReactionAddedCallback>();

        public static object messageReactionDeletedCallbackLock = new object();
        public Dictionary<int, ZIMMessageReactionDeletedCallback> messageReactionDeletedCallback = new Dictionary<int, ZIMMessageReactionDeletedCallback>();

        public static object messageReactionUserListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMMessageReactionUserListQueriedCallback> messageReactionUserListQueriedCallback = new Dictionary<int, ZIMMessageReactionUserListQueriedCallback>();

        public static object messageReceiptsInfoQueriedCallbackLock = new object();
        public Dictionary<int, ZIMMessageReceiptsInfoQueriedCallback> messageReceiptsInfoQueriedCallback = new Dictionary<int, ZIMMessageReceiptsInfoQueriedCallback>();

        public static object conversationsSearchedCallbackLock = new object();
        public Dictionary<int, ZIMConversationsSearchedCallback> conversationsSearchedCallback = new Dictionary<int, ZIMConversationsSearchedCallback>();

        public static object groupsSearchedCallbackLock = new object();
        public Dictionary<int, ZIMGroupsSearchedCallback> groupsSearchedCallback = new Dictionary<int, ZIMGroupsSearchedCallback>();

        public static object groupMembersSearchedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMembersSearchedCallback> groupMembersSearchedCallback = new Dictionary<int, ZIMGroupMembersSearchedCallback>();

        public static object mediaDownloadedCallbackLock = new object();
        public Dictionary<int, ZIMMediaDownloadedCallback> mediaDownloadedCallback = new Dictionary<int, ZIMMediaDownloadedCallback>();

        public static object callingInvitationSentCallbackLock = new object();
        public Dictionary<int, ZIMCallingInvitationSentCallback> callingInvitationSentCallback = new Dictionary<int, ZIMCallingInvitationSentCallback>();

        public static object callEndSentCallbackLock = new object();
        public Dictionary<int, ZIMCallEndSentCallback> callEndSentCallback = new Dictionary<int, ZIMCallEndSentCallback>();

        public static object callInvitationListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMCallInvitationListQueriedCallback> callInvitationListQueriedCallback = new Dictionary<int, ZIMCallInvitationListQueriedCallback>();

        public static object callQuitSentCallbackLock = new object();
        public Dictionary<int, ZIMCallQuitSentCallback> callQuitSentCallback = new Dictionary<int, ZIMCallQuitSentCallback>();

        public static object messageSendNotificationCallbackLock = new object();
        public Dictionary<int, ZIMMessageSendNotification> messageSendNotificationCallback = new Dictionary<int, ZIMMessageSendNotification>();

        public static object mediaMessageSendNotificationCallbackLock = new object();
        public Dictionary<int, ZIMMediaMessageSendNotification> mediaMessageSendNotificationCallback = new Dictionary<int, ZIMMediaMessageSendNotification>();

        public static object mediaDownloadingProgressCallbackLock = new object();
        public Dictionary<int, ZIMMediaDownloadingProgress> mediaDownloadingProgressCallback = new Dictionary<int, ZIMMediaDownloadingProgress>();

        public static object messageDeletedCallbackLock = new object();
        public Dictionary<int, ZIMMessageDeletedCallback> messageDeletedCallback = new Dictionary<int, ZIMMessageDeletedCallback>();

        public static object messageRevokedCallbackLock = new object();
        public Dictionary<int, ZIMMessageRevokedCallback> messageRevokedCallback = new Dictionary<int, ZIMMessageRevokedCallback>();

        public static object conversationListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMConversationListQueriedCallback> conversationListQueriedCallback = new Dictionary<int, ZIMConversationListQueriedCallback>();

        public static object conversationPinnedListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMConversationPinnedListQueriedCallback> conversationPinnedListQueriedCallback = new Dictionary<int, ZIMConversationPinnedListQueriedCallback>();

        public static object conversationUnreadMessageCountClearedCallbackLock = new object();
        public Dictionary<int, ZIMConversationUnreadMessageCountClearedCallback> conversationUnreadMessageCountClearedCallback = new Dictionary<int, ZIMConversationUnreadMessageCountClearedCallback>();

        public static object conversationDeletedCallbackLock = new object();
        public Dictionary<int, ZIMConversationDeletedCallback> conversationDeletedCallback = new Dictionary<int, ZIMConversationDeletedCallback>();

        public static object conversationPinnedStateUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMConversationPinnedStateUpdatedCallback> conversationPinnedStateUpdatedCallback = new Dictionary<int, ZIMConversationPinnedStateUpdatedCallback>();

        public static object conversationQueriedCallbackLock = new object();
        public Dictionary<int, ZIMConversationQueriedCallback> conversationQueriedCallback = new Dictionary<int, ZIMConversationQueriedCallback>();

        public static object conversationMessageReceiptReadSentCallbackLock = new object();
        public Dictionary<int, ZIMConversationMessageReceiptReadSentCallback> conversationMessageReceiptReadSentCallback = new Dictionary<int, ZIMConversationMessageReceiptReadSentCallback>();

        public static object roomCreatedCallbackLock = new object();
        public Dictionary<int, ZIMRoomCreatedCallback> roomCreatedCallback = new Dictionary<int, ZIMRoomCreatedCallback>();

        public static object roomJoinedCallbackLock = new object();
        public Dictionary<int, ZIMRoomJoinedCallback> roomJoinedCallback = new Dictionary<int, ZIMRoomJoinedCallback>();

        public static object roomEnteredCallbackLock = new object();
        public Dictionary<int, ZIMRoomEnteredCallback> roomEnteredCallback = new Dictionary<int, ZIMRoomEnteredCallback>();

        public static object roomLeftCallbackLock = new object();
        public Dictionary<int, ZIMRoomLeftCallback> roomLeftCallback = new Dictionary<int, ZIMRoomLeftCallback>();

        public static object roomOnlineMemberCountQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomOnlineMemberCountQueriedCallback> roomOnlineMemberCountQueriedCallback = new Dictionary<int, ZIMRoomOnlineMemberCountQueriedCallback>();

        public static object logUploadedCallbackLock = new object();
        public Dictionary<int, ZIMLogUploadedCallback> logUploadedCallback = new Dictionary<int, ZIMLogUploadedCallback>();

        public static object roomAttributesOperatedCallbackLock = new object();
        public Dictionary<int, ZIMRoomAttributesOperatedCallback> roomAttributesOperatedCallback = new Dictionary<int, ZIMRoomAttributesOperatedCallback>();

        public static object roomAttributesQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomAttributesQueriedCallback> roomAttributesQueriedCallback = new Dictionary<int, ZIMRoomAttributesQueriedCallback>();

        public static object roomAttributesBatchOperatedCallbackLock = new object();
        public Dictionary<int, ZIMRoomAttributesBatchOperatedCallback> roomAttributesBatchOperatedCallback = new Dictionary<int, ZIMRoomAttributesBatchOperatedCallback>();

        public static object roomMembersAttributesOperatedCallbackLock = new object();
        public Dictionary<int, ZIMRoomMembersAttributesOperatedCallback> roomMembersAttributesOperatedCallback = new Dictionary<int, ZIMRoomMembersAttributesOperatedCallback>();

        public static object roomMembersAttributesQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomMembersAttributesQueriedCallback> roomMembersAttributesQueriedCallback = new Dictionary<int, ZIMRoomMembersAttributesQueriedCallback>();

        public static object roomMemberAttributesListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMRoomMemberAttributesListQueriedCallback> roomMemberAttributesListQueriedCallback = new Dictionary<int, ZIMRoomMemberAttributesListQueriedCallback>();

        public static object groupCreatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupCreatedCallback> groupCreatedCallback = new Dictionary<int, ZIMGroupCreatedCallback>();

        public static object groupDismissedCallbackLock = new object();
        public Dictionary<int, ZIMGroupDismissedCallback> groupDismissedCallback = new Dictionary<int, ZIMGroupDismissedCallback>();

        public static object groupJoinedCallbackLock = new object();
        public Dictionary<int, ZIMGroupJoinedCallback> groupJoinedCallback = new Dictionary<int, ZIMGroupJoinedCallback>();

        public static object groupLeftCallbackLock = new object();
        public Dictionary<int, ZIMGroupLeftCallback> groupLeftCallback = new Dictionary<int, ZIMGroupLeftCallback>();

        public static object groupUsersInvitedCallbackLock = new object();
        public Dictionary<int, ZIMGroupUsersInvitedCallback> groupUsersInvitedCallback = new Dictionary<int, ZIMGroupUsersInvitedCallback>();

        public static object groupMemberKickedoutCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberKickedCallback> groupMemberKickedoutCallback = new Dictionary<int, ZIMGroupMemberKickedCallback>();

        public static object groupOwnerTransferredCallbackLock = new object();
        public Dictionary<int, ZIMGroupOwnerTransferredCallback> groupOwnerTransferredCallback = new Dictionary<int, ZIMGroupOwnerTransferredCallback>();

        public static object groupNameUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupNameUpdatedCallback> groupNameUpdatedCallback = new Dictionary<int, ZIMGroupNameUpdatedCallback>();

        public static object groupAvatarUrlUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupAvatarUrlUpdatedCallback> groupAvatarUrlUpdatedCallback = new Dictionary<int, ZIMGroupAvatarUrlUpdatedCallback>();

        public static object groupNoticeUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupNoticeUpdatedCallback> groupNoticeUpdatedCallback = new Dictionary<int, ZIMGroupNoticeUpdatedCallback>();

        public static object groupInfoQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupInfoQueriedCallback> groupInfoQueriedCallback = new Dictionary<int, ZIMGroupInfoQueriedCallback>();

        public static object groupAttributesOperatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupAttributesOperatedCallback> groupAttributesOperatedCallback = new Dictionary<int, ZIMGroupAttributesOperatedCallback>();

        public static object groupAttributesQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupAttributesQueriedCallback> groupAttributesQueriedCallback = new Dictionary<int, ZIMGroupAttributesQueriedCallback>();

        public static object groupMemberRoleUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberRoleUpdatedCallback> groupMemberRoleUpdatedCallback = new Dictionary<int, ZIMGroupMemberRoleUpdatedCallback>();

        public static object groupMemberNicknameUpdatedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberNicknameUpdatedCallback> groupMemberNicknameUpdatedCallback = new Dictionary<int, ZIMGroupMemberNicknameUpdatedCallback>();

        public static object groupMemberInfoQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberInfoQueriedCallback> groupMemberInfoQueriedCallback = new Dictionary<int, ZIMGroupMemberInfoQueriedCallback>();

        public static object groupListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupListQueriedCallback> groupListQueriedCallback = new Dictionary<int, ZIMGroupListQueriedCallback>();

        public static object groupMemberListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberListQueriedCallback> groupMemberListQueriedCallback = new Dictionary<int, ZIMGroupMemberListQueriedCallback>();

        public static object groupMemberCountQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMemberCountQueriedCallback> groupMemberCountQueriedCallback = new Dictionary<int, ZIMGroupMemberCountQueriedCallback>();

        public static object groupMessageReceiptMemberListQueriedCallbackLock = new object();
        public Dictionary<int, ZIMGroupMessageReceiptMemberListQueriedCallback> groupMessageReceiptMemberListQueriedCallback = new Dictionary<int, ZIMGroupMessageReceiptMemberListQueriedCallback>();

        public static object conversationNotificationStatusSetCallbackLock = new object();
        public Dictionary<int, ZIMConversationNotificationStatusSetCallback> conversationNotificationStatusSetCallback = new Dictionary<int, ZIMConversationNotificationStatusSetCallback>();

        public static object callInvitationSentCallbackLock = new object();
        public Dictionary<int, ZIMCallInvitationSentCallback> callInvitationSentCallback = new Dictionary<int, ZIMCallInvitationSentCallback>();

        public static object callCancelSentCallbackLock = new object();
        public Dictionary<int, ZIMCallCancelSentCallback> callCancelSentCallback = new Dictionary<int, ZIMCallCancelSentCallback>();

        public static object callAcceptanceSentCallbackLock = new object();
        public Dictionary<int, ZIMCallAcceptanceSentCallback> callAcceptanceSentCallback = new Dictionary<int, ZIMCallAcceptanceSentCallback>();

        public static object callRejectionSentCallbackLock = new object();
        public Dictionary<int, ZIMCallRejectionSentCallback> callRejectionSentCallback = new Dictionary<int, ZIMCallRejectionSentCallback>();

        public static object messageObjMapLock = new object();
        public Dictionary<int, ZIMMessage> messageObjMap = new Dictionary<int, ZIMMessage>();

        private int sequence;
    }
}
