﻿using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using UnityEngine;


namespace ZEGO
{

    // MARK: - CallEnum

    // MARK: - Struct

    // MARK: - MainStruct

    /// Error structure
    ///
    /// Description: A structure composed of error code enumeration values and detailed description of the corresponding error information.
    ///
    public struct zim_error
    {
        public zim_error_code code;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string message;
    };

    /// Log configuration
    ///
    /// Description: Configure the storage path of log files and the maximum log file size.
    ///
    public struct zim_log_config
    {

        /// The storage path of the log files. Refer to the official website document for the default path.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string log_path;

        /// The maximum log file size (Bytes). The default maximum size is 5MB (5 * 1024 * 1024 Bytes)
        public ulong log_size;
    };

    /// Cache configuration
    ///
    /// Description: Configure the storage path of cache files (such as chat records).
    ///
    public struct zim_cache_config
    {

        /// The storage path of the cache files. Refer to the official website document for the default path.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string cache_path;
    };

    /// User information object.
    ///
    /// Description: Identifies a unique user.
    ///
    /// Caution: Note that the userID must be unique under the same appID, otherwise mutual kicks out will occur.
    /// It is strongly recommended that userID corresponds to the user ID of the business APP,
    /// that is, a userID and a real user are fixed and unique, and should not be passed to the SDK in a random userID.
    /// Because the unique and fixed userID allows ZEGO technicians to quickly locate online problems.
    ///
    public struct zim_user_info
    {

        /// User ID, a string with a maximum length of 32 bytes or less. Only support numbers, English characters and '~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+', '=', '-', '`', ';', '’', ',', '.', '<', '>', '/', '\'.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;

        /// User name, a string with a maximum length of 64 bytes or less
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_name;
    };

    public struct zim_room_member_info
    {
        public zim_user_info user_info;
    };

    public struct zim_user_full_info
    {
        /// User ID, a string with a maximum length of 32 bytes or less. Only support numbers, English characters and '~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+', '=', '-', '`', ';', '’', ',', '.', '<', '>', '/', '\'.
        public zim_user_info base_info;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_avatar_url;

        /// User extra, a string with a maximum length of 2K bytes.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_extended_data;
    };

    public struct zim_error_user_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
        public uint reason;
    };

    public struct zim_users_info_query_config
    {
        public bool is_query_from_server;
    };

    // MARK: - MessageStruct

    /// Base class of message object
    ///
    /// Description: Identifies the basic parameters of a message.
    ///
    /// Caution: Some of the parameters, such as Message ID, only have values ​​during the callback.
    /// Developers do not need to assign values ​​to these parameters when they actively create this object for sending messages.
    ///
    [System.Runtime.InteropServices.StructLayoutAttribute(System.Runtime.InteropServices.LayoutKind.Sequential)]
    public struct zim_message
    {

        // BaseMessage

        /// Identifies the type of this message.
        public System.Int32 type;

        /// Description: The unique ID that identifies this message.
        ///
        /// Use cases: Can be used to index other messages.
        ///
        /// Caution: When the developer actively creates a message, there is no need to modify this parameter.
        /// This parameter only has a value during callback.
        public System.Int64 message_id;

        public System.Int64 local_message_id;
        /// Description: Identifies which User sent this message.
        ///
        /// Use cases: Used to show who sent a message.
        ///
        /// Caution: When the developer actively creates a message, there is no need to modify this parameter.
        /// This parameter only has a value during callback.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string sender_user_id;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_id;

        public System.Int32 direction;

        public System.Int32 sent_status;

        public System.Int32 conversation_type;

        public System.Int32 receipt_status;

        public System.Int64 conversation_seq;

        /// Description: Identifies the sending time of a message
        ///
        /// Use cases: Used to present the sending time of a message, and can be used for message sorting.
        ///
        /// Caution: This is a standard UNIX timestamp, in milliseconds.
        public System.UInt64 timestamp;

        /// Developers can use this field to sort the index
        public System.Int64 order_key;

        [MarshalAs(UnmanagedType.U1)]
        public bool is_user_inserted;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;

        // TextMessage/BarrageMessage/SystemMessage Only

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string local_extended_data;

        [MarshalAs(UnmanagedType.U1)]
        public bool is_broadcast_message;

        /// The content of the message supports arbitrary data. If the message type is text, it should be a standard UTF-8 string.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string message;

        // CommandMessage Only
        public System.IntPtr command_message;

        /// The length of the message. If the message type is text, it should be the length of the string.
        public System.UInt32 command_message_length;

        // MediaMessage/FileMessage Only
        // 文件本地地址，该地址为发送富媒体消息时，用户唯一需要填入的字段
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string file_local_path;

        // 文件下载地址
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string file_download_url;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string file_uid;

        // 文件名，仅在接收文件消息时使用，供 SDK 返回给上层以展示文件名，其余时候为 nullptr
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string file_name;

        // 文件大小，接收图片消息时为原图大小，接收文件信息时为文件大小，接收音视频消息时为音视频大小，
        // 由发送端 SDK 获取文件后，自动填入，并在发送消息时写入消息体内，接收端自行解析，以便接收端展示文件大小
        public System.UInt64 file_size;

        // ImageMessage Only

        // 大图下载地址，仅在接收图片消息时使用，供 SDK 返回给上层以给到上层使用文件下载（大图下载），其余时候为 nullptr
        //todo
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string large_image_download_url;

        // 缩略图下载地址，仅在接收图片/视频消息时使用，供 SDK 返回给上层以给到上层使用文件下载（缩略图/视频首帧缩略图下载），其余时候为 nullptr
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string thumbnail_download_url;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string large_image_local_path;

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string thumbnail_local_path;

        public System.UInt32 original_image_width;
        public System.UInt32 original_image_height;

        public System.UInt32 large_image_width;
        public System.UInt32 large_image_height;

        public System.UInt32 thumbnail_width;
        public System.UInt32 thumbnail_height;

        // AudioMessage/VideoMessage

        // 音视频时长，由发送端 SDK 获取文件后，自动填入，并在发送消息时写入消息体内，接收端自行解析，以便接收端展示时长
        public System.UInt64 media_duration;

        // VideoMessage Only

        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string video_first_frame_local_path;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string video_first_frame_download_url;

        public System.UInt32 video_first_frame_width;
        public System.UInt32 video_first_frame_height;
        // RevokeMessage Only
        public System.Int32 revoke_status;
        public System.Int32 revoke_type;
        public System.UInt64 revoke_timestamp;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string operated_userid;
        public System.Int32 original_message_type;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string original_text_message;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string revoke_extended_data;
        public System.UInt32 sub_type;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string searched_content;
        public System.IntPtr reactions;
        public System.UInt32 reaction_length;
    };

    public struct zim_push_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string title;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string content;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string payload;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string resources_id;
    };

    public struct zim_message_send_config
    {
        public System.Int32 priority;
        public bool has_receipt;
        public bool enable_offline_push;
        public zim_push_config push_config;
    };

    /// Configuration for querying message.
    ///
    /// Description: When querying message, you need to configure this object.
    ///
    public struct zim_message_query_config
    {
        /// Description: The unique ID that identifies a message.
        public System.IntPtr next_message;

        /// Description: How many messages are retrieved in one query.
        ///
        /// Caution: To obtain messages in pages to reduce overhead, it is recommended to obtain within 100 messages at a time.
        public uint count;

        /// Description: Whether to query message from old to new forwardly or from new to old reversely.
        public bool reverse;
    };

    public struct zim_message_delete_config
    {
        public bool is_also_delete_server_message;
    };

    public struct zim_message_sent_status_change_info
    {
        public zim_message message;
        public System.Int32 status;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string reason;
    };

    public struct zim_message_receipt_info
    {
        public System.Int32 status;
        public long message_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_id;
        public System.Int32 conversation_type;
        public uint read_member_count;
        public uint unread_member_count;
    };

    public struct zim_group_message_receipt_member_query_config
    {
        public uint next_flag;
        public uint count;
        //public bool query_read_list;
    };

    public struct zim_message_search_config
    {
        public System.IntPtr next_message;
        public System.UInt32 count;
        public System.Int32 order;

        public System.IntPtr keywords;
        public System.UInt32 keywords_length;
        public System.IntPtr message_types;
        public System.UInt32 message_types_length;

        public System.IntPtr sub_message_types;
        public System.UInt32 sub_message_types_length;

        public System.IntPtr sender_user_ids;
        public System.UInt32 sender_user_ids_length;
        public System.Int64 start_time;
        public System.Int64 end_time;
    };

    public struct zim_conversation_search_config
    {
        public System.UInt32 next_flag;
        public System.UInt32 total_conversation_count;
        public System.UInt32 conversation_message_count;
        public System.IntPtr keywords;
        public System.UInt32 keywords_length;
        public System.IntPtr message_types;
        public System.UInt32 message_types_length;
        public System.IntPtr sub_message_types;
        public System.UInt32 sub_message_types_length;
        public System.IntPtr sender_user_ids;
        public System.UInt32 sender_user_ids_length;
        public System.Int64 start_time;
        public System.Int64 end_time;
    };

    public struct zim_conversation_search_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_id;
        public System.Int32 conversation_type;
        public System.UInt32 total_message_count;
        public System.IntPtr message_list;
        public System.UInt32 message_list_length;
    };

    public struct zim_group_search_config
    {
        public System.UInt32 next_flag;
        public System.UInt32 count;
        public System.IntPtr keywords;
        public System.UInt32 keywords_length;
        public bool is_also_match_group_member_user_name;
        public bool is_also_match_group_member_nick_name;
    };

    public struct zim_group_search_info
    {
        public zim_group_info group_info;
        public System.IntPtr user_list;
        public System.UInt32 user_list_length;
    };

    public struct zim_group_member_search_config
    {
        public System.UInt32 next_flag;
        public System.UInt32 count;
        public System.IntPtr keywords;
        public System.UInt32 keywords_length;
        public bool is_also_match_group_member_nick_name;
    };

    public struct zim_message_reaction_user_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
    };

    public struct zim_message_reaction
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_id;
        public System.Int32 conversation_type;
        public System.Int64 message_id;

        public System.IntPtr user_list;
        public System.UInt32 user_info_length;

        public System.UInt32 total_count;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string reaction_type;
        public bool self_included;
    };

    public struct zim_message_reaction_users_query_config
    {
        public System.UInt64 next_flag;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string reaction_type;
        public System.UInt32 count;
    };

    public struct zim_message_revoke_config
    {
        public bool enable_offline_push;
        public zim_push_config push_config;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string revoke_extended_data;
    };

    // MARK: - ConversationStruct

    public struct zim_conversation
    {
        // 会话ID
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_id;
        // 会话名
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_name;
        // 会话头像
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string conversation_avatar_url;
        // 会话类型
        public System.Int32 type;
        // 未读消息数
        public System.UInt32 unread_message_count;
        // 最后一条消息对象
        public zim_message last_message;
        // 用于排序的索引
        public System.Int64 order_key;
        // 会话的免打扰状态
        public System.Int32 notification_status;
        [MarshalAs(UnmanagedType.U1)]
        public bool is_pinned;
    };

    public struct zim_conversation_change_info
    {
        public System.Int32 conversation_event;
        public zim_conversation conversation;
    };

    public struct zim_conversation_query_config
    {
        //public struct zim_conversation *next_conversation;
        public long next_order_key;
        // 一次分页拉取的会话数量
        public int count;
    };

    public struct zim_conversation_delete_config
    {
        // 是否同时删除服务端会话
        public bool is_also_delete_server_conversation;
    };

    // MARK: - GroupStruct

    public struct zim_group_attribute
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string key;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string value;
    };

    public struct zim_group_member_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_name;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string member_nick_name;
        public int member_role;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string member_avatar_url;
    };

    public struct zim_group_operated_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_name;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string member_nick_name;
        public int member_role;
    };

    public struct zim_group_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string group_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string group_name;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string group_avatar_url;
    };

    public struct zim_group_advanced_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string group_notice;
        public System.IntPtr group_attributes;
        public uint group_attributes_length;
    };

    public struct zim_group_full_info
    {
        public zim_group_info base_info;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string group_notice;
        public System.IntPtr group_attributes;
        public uint group_attributes_length;
        public System.Int32 notification_status;
    };

    public struct zim_group
    {
        public zim_group_info base_info;
        public System.Int32 notification_status;
    };

    public struct zim_group_member_query_config
    {
        public uint next_flag;
        public uint count;
    };

    public struct zim_group_attributes_update_info
    {
        public System.Int32 action;

        public System.IntPtr group_attributes;

        public uint group_attributes_length;
    };

    // MARK: - RoomStruct

    /// Room information object.
    ///
    /// Description: Identifies a unique room.
    ///
    public struct zim_room_info
    {

        /// Room ID, a string with a maximum length of 32 bytes or less. Only support numbers, English characters and '~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '_', '+', '=', '-', '`', ';', '’', ',', '.', '<', '>', '/', '\'.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string room_id;

        /// Room name, a string with a maximum length of 64 bytes or less
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string room_name;
    };

    /// Room attribute object.
    ///
    /// Description: Room attribute structure.
    ///
    public struct zim_room_attribute
    {

        /// The sequence of the room attribute
        //public uint sequence;

        /// Attribute key, visible characters, no more than 16
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string key;

        /// The attribute value, which can be any character, does not exceed 1k in a single entry, and does not exceed 5k in total
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string value;
    };

    public struct zim_room_advanced_config
    {
        public System.IntPtr room_attributes;

        public uint room_attributes_length;

        public uint room_destroy_delay_time;
    };

    /// Room attribute updated information (usually obtained by server push)
    ///
    /// Description: Room attributes updated information structure.
    ///

    public struct zim_room_attributes_update_info
    {

        /// Room attribute change event
        public System.Int32 room_attributes_action;

        /// 房间属性信息
        public System.IntPtr room_attributes;

        public uint room_attributes_length;
    };

    public struct zim_room_attributes_set_config
    {
        public bool is_force;
        public bool is_delete_after_owner_left;
        public bool is_update_owner;
    };

    public struct zim_room_attributes_delete_config
    {
        public bool is_force;
    };

    public struct zim_room_attributes_batch_operation_config
    {
        public bool is_force;
        public bool is_delete_after_owner_left;
        public bool is_update_owner;
    };

    public struct zim_room_member_attribute
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string key;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string value;
    };

    public struct zim_room_member_attributes_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
        public System.IntPtr attributes;
        public uint attributes_length;
    };

    public struct zim_room_member_attributes_update_info
    {
        public zim_room_member_attributes_info attributes_info;
    };

    public struct zim_room_member_attributes_operated_info
    {
        public zim_room_member_attributes_info attributes_info;
        public System.IntPtr error_keys;
        public uint error_keys_length;
    };

    public struct zim_room_member_attributes_set_config
    {
        public bool is_delete_after_owner_left;
    };

    public struct zim_room_member_attributes_query_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string next_flag;
        public uint count;
    };

    /// The room details object.
    ///
    /// Description: Identifies the basic parameters of a message.
    ///
    /// Caution: Identifies the detailed information of a room.
    ///
    public struct zim_room_full_info
    {

        /// The base object of the room.
        public zim_room_info base_info;
    };

    /// Configuration for querying member.
    ///
    /// Description: When querying member, you need to configure this object.
    ///
    public struct zim_room_member_query_config
    {

        /// Description: The flag of the paging query. For the first query, set this field to an empty string. If the "nextFlag" field of the callback is not an empty string, it needs to be set here to continue the query on the next page.
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string next_flag;

        /// Description: How many messages are retrieved in one query.
        ///
        /// Caution: To obtain messages in pages to reduce overhead, it is recommended to obtain within 100 messages at a time.
        public uint count;
    };

    public struct zim_room_operated_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
    };

    // MARK: - CallStruct

    public struct zim_call_user_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string user_id;
        public System.Int32 user_state;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invite_config
    {
        public int timeout;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public bool enable_offline_push;
        public System.Int32 mode;
        public zim_push_config push_config;
    };
    public struct zim_calling_invite_config
    {
        public bool enable_offline_push;
        public zim_push_config push_config;
    };
    public struct zim_call_cancel_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public bool enable_offline_push;
        public zim_push_config push_config;
    };

    public struct zim_call_accept_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_reject_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invitation_sent_info
    {
        public int timeout;
        public System.IntPtr old_error_invitees;
        public System.IntPtr error_invitees;
        public uint error_invitees_length;
        public uint old_error_invitees_length;
    };

    public struct zim_calling_invitation_sent_info
    {
        public System.IntPtr error_invitees;
        public uint error_invitees_length;
    };

    public struct zim_call_quit_sent_info
    {
        public System.Int64 create_time;
        public System.Int64 accept_time;
        public System.Int64 quit_time;
    };

    public struct zim_call_end_sent_info
    {
        public System.Int64 create_time;
        public System.Int64 accept_time;
        public System.Int64 end_time;
    };

    public struct zim_call_invitation_received_info
    {
        public System.Int32 mode;
        public System.Int64 create_time;
        public System.Int32 timeout;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string caller;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string inviter;
        public System.IntPtr call_user_list;
        public System.UInt32 old_error_invitees_length;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invitation_cancelled_info
    {
        public System.Int32 mode;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string inviter;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invitation_accepted_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string invitee;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invitation_rejected_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string invitee;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
    };

    public struct zim_call_invitation_timeout_info
    {
        public System.Int32 mode;
        public bool also_invoke_discard_method;
    };

    public struct zim_call_invitation_ended_info
    {
        public System.Int32 mode;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string caller;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string operated_user_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public System.Int64 end_time;
    };

    public struct zim_event_call_user_state_changed_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string call_id;
        public System.IntPtr call_user_list;
        public System.UInt32 call_user_list_length;
    };

    public struct zim_call_quit_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public bool enable_offline_push;
        public zim_push_config push_config;
    };

    public struct zim_call_end_config
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public bool enable_offline_push;
        public zim_push_config push_config;
    };

    public struct zim_query_call_list_config
    {
        public System.UInt32 count;

        public System.Int64 next_flag;
    };

    public struct zim_call_info
    {
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string call_id;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string caller;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string inviter;
        public System.Int64 create_time;
        public System.Int64 end_time;
        public System.Int32 state;
        public System.Int32 mode;
        [System.Runtime.InteropServices.MarshalAsAttribute(System.Runtime.InteropServices.UnmanagedType.LPStr)]
        public string extended_data;
        public System.IntPtr call_user_list;
        public System.UInt32 call_user_list_length;
        // int timeout;
        //    long long call_duration;
        //    long long user_duration;
    };

    public struct zim_event_call_state_changed_info
    {
        public zim_call_info call_info;
    };

}